// Source file for the pdb viewer program



// Include files 

#include "PDB/PDB.h"
#include "R2Shapes/R2Shapes.h"
#include <GL/glut.h>
#include "GL/glui.h"



// GLUI commands

enum {
  NULL_COMMAND,
  QUIT_COMMAND
};



// Program variables

static char *chd_name = NULL;
static RNScalar background_color[3] = { 1, 1, 1 };
static int make_images = 0;
static int print_verbose = 0;
static int print_debug = 0;



// GLUI variables 

static GLUI *glui = NULL;
static int GLUIwindow = 0;
static int GLUIwindow_height = 1024;
static int GLUIwindow_width = 1024;
static int GLUImouse[2] = { 0, 0 };
static int GLUIbutton[3] = { 0, 0, 0 };
static int GLUImodifiers = 0;



// Chord histogram data

static int max_bin = 0;
static RNScalar min_distance = 0;
static RNScalar max_distance = 0;
static RNScalar nearby_threshold = 0;
static RNScalar bin_spacing = 0;
static int nelements = 0;
static const int max_distance_bins = 1024;
static const int max_parameter_bins = max_distance_bins;
static const int max_element_bins = 6;
static RNScalar bins[max_distance_bins][max_parameter_bins][max_element_bins] = { 0 };
static RNScalar observations[max_distance_bins] = { 0 };



// Display variables

static int distance_index = 0;
static int element_index = 0;



static int
ReadChordFile(const char *chd_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(chd_name, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open chord histogram file: %s\n", chd_name);
    return 0;
  }

  // Read parameters into global variables
  if (fscanf(fp, "%d%lf%lf%lf%lf%d", &max_bin, &min_distance, &max_distance, &bin_spacing, &nearby_threshold, &nelements) != 6) {
    fprintf(stderr, "Error reading chord file header\n");
    return 0;
  }

  // Read histogram into global variables
  int bin, nbins = 0;
  RNScalar sum = 0;
  RNScalar max_value = 0;
  assert(max_bin < max_distance_bins);
  assert(max_bin == (int) (max_distance / bin_spacing));
  for (int i = 0; i <= max_bin; i++) {
    // Read per-distance information
    if (fscanf(fp, "%d%lf", &bin, &(observations[i])) != 2) {
      fprintf(stderr, "Error reading chord file leader for %d\n", i);
      return 0;
    }

    // Check bin number
    if (bin != i) {
      fprintf(stderr, "Bin number does not match for %d in %s\n", i, chd_name);
      return 0;
    }

    // Read ":" 
    char buffer[64];
    if ((fscanf(fp, "%s", buffer) != 1) || (strcmp(buffer, ":"))) {
      fprintf(stderr, "Error reading chord file colon for %d\n", i);
      return 0;
    }


    // Read bins
    for (int j = 0; j <= i; j++) {
      for (int k = 0; k < nelements; k++) {
        // Read bin
        if (fscanf(fp, "%lf", &(bins[i][j][k])) != 1) {
          fprintf(stderr, "Error reading chord file bin (%d %d %d)\n", i, j, k);
          return 0;
        }

        // Gather statistics
        if (bins[i][j][k] > max_value) max_value = bins[i][j][k];
        sum += bins[i][j][k];
        nbins++;
      }
    }
  }

  // Close file
  fclose(fp);

  // Set default indices
  distance_index = (max_bin / bin_spacing < 10) ? max_bin : 10 / bin_spacing;
  element_index = (nelements == 1) ? 0 : 1;

  // Print statistics
  if (print_verbose) {
    printf("Read chord histogram ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bins = %d\n", nbins);
    printf("  # Elements = %d\n", nelements);
    printf("  Max bin = %d\n", max_bin);
    printf("  Min distance = %g\n", min_distance);
    printf("  Max distance = %g\n", max_distance);
    printf("  Bin spacing = %g\n", bin_spacing);
    printf("  Nearby threshold = %g\n", nearby_threshold);
    printf("  Max value = %g\n", max_value);
    printf("  Sum = %g\n", sum);
    fflush(stdout);
  }

  // Return success
  return 1;
}



void GLUIStop(void)
{
  // Destroy window 
  glutDestroyWindow(GLUIwindow);

  // Exit
  exit(0);
}



void GLUIRedraw(void)
{
  // Clear window 
  glClearColor(background_color[0], background_color[1], background_color[2], 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glDisable(GL_LIGHTING);

  // Determine maximum value
  RNScalar sum = 0;
  RNScalar max_value = RN_EPSILON;
  for (int i = 0; i <= max_bin; i++) {
    RNScalar value = bins[distance_index][i][element_index];
    if (value > max_value) max_value = value;
    sum += value;
  }

  // Set transformation for pixels
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity(); 
  gluOrtho2D(0, GLUIwindow_width-1, 0, GLUIwindow_height-1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Draw status text
  char buffer[128];
  sprintf(buffer, "%d  ,  %s    :    %g     %g", distance_index, PDBelements[element_index].Name(), observations[distance_index], sum);
  glColor3f(1 - background_color[0], 1 - background_color[1], 1 - background_color[2]);
  R2DrawText((RNScalar) (GLUIwindow_width - 300), (RNScalar) (GLUIwindow_height - 50), buffer);

  // Set transformation for plot
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity(); 
  gluOrtho2D(-0.1 * max_bin, 1.1 * max_bin, -0.1 * max_value, 1.1 * max_value);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Draw axes
  glLineWidth(5);
  glColor3f(1 - background_color[0], 1 - background_color[1], 1 - background_color[2]);
  R2BeginLine();
  R2LoadPoint(0.0, max_value);
  R2LoadPoint(0.0, 0.0);
  R2LoadPoint(max_bin, 0.0);
  R2EndLine();
  sprintf(buffer, "%g", max_value);
  R2DrawText(0.0, max_value, buffer);

  // Draw tic marks
  glLineWidth(1);
  glColor3f(1 - background_color[0], 1 - background_color[1], 1 - background_color[2]);
  for (int i = 0; i <= distance_index; i++) {
    R2BeginLine();
    R2LoadPoint(i, 0.0);
    R2LoadPoint(i, -0.01 * max_value);
    R2EndLine();
    sprintf(buffer, "%g", i * bin_spacing);
    R2DrawText(i, -0.05 * max_value, buffer);
  }

  // Draw curve
  glLineWidth(5);
  glColor3f(background_color[0], 1 - background_color[1], 1 - background_color[2]);
  R2BeginLine();
  for (int i = 0; i <= distance_index; i++) {
    RNCoord x = i;
    RNScalar y = bins[distance_index][i][element_index];
    R2LoadPoint(x, y);
  }
  R2EndLine();

  // Capture image and exit
  if (make_images) {
    if (sum != 0) {
      char image_name[256];
      sprintf(image_name, "%s_%d_%s.jpg", chd_name, distance_index, PDBelements[element_index].Name());
      R2Image image(GLUIwindow_width, GLUIwindow_height, 3);
      image.Capture();
      image.Write(image_name);
    }

    // Update frame index and element_index 
    element_index++;
    if (element_index >= nelements) {
      distance_index++;
      element_index = 0;
    }
    if (distance_index > max_bin) {
      distance_index = 0;
      GLUIStop();
    }
  }

  // Swap buffers 
  glutSwapBuffers();
}    



void GLUIResize(int a, int b)
{
  // Resize window
  int x, y, w, h;
  GLUI_Master.get_viewport_area( &x, &y, &w, &h );
  glViewport( x, y, w, h );

  // Remember window size 
  GLUIwindow_width = w;
  GLUIwindow_height = h;

  // Redraw
  glutPostRedisplay();
}



void GLUIMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Compute mouse movement
  int dx = x - GLUImouse[0];
  int dy = y - GLUImouse[1];
  
  // Process mouse movement

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;
}



void GLUIMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;
  
  // Process mouse button event

  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUIbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Redraw
  glutPostRedisplay();
}



void GLUISpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Process keyboard button event 
  switch (key) {
  case GLUT_KEY_DOWN:
    break;

  case GLUT_KEY_UP:
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();
}



void GLUIKeyboard(unsigned char key, int x, int y)
{
  // Process keyboard button event 
  switch (key) {
  case ',':
    distance_index--;
    if (distance_index < 0) distance_index = max_bin;
    break;

  case '.':
    distance_index++;
    if (distance_index > max_bin) distance_index = 0;
    break;

  case '<':
    element_index--;
    if (element_index < 0) element_index = nelements - 1;
    break;

  case '>':
    element_index++;
    if (element_index >= nelements) element_index = 0;
    break;

  case 27: // ESCAPE
    GLUIStop();
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = GLUIwindow_height - y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();  
}



void GLUICallback(int control)
{
  switch (control) {
  case QUIT_COMMAND:
    GLUIStop();
    break;
  }
}



void GLUIIdle(void)
{
  // Just checking
  assert(make_images);

  // Set current window
  if ( glutGetWindow() != GLUIwindow ) 
    glutSetWindow(GLUIwindow); 

  // Redraw
  glutPostRedisplay();
}



void GLUISetup(void)
{
  GLUI_Panel *panel;
  GLUI_Spinner *spinner;

  // Initialize GLUI
  glui = GLUI_Master.create_glui("MBE Controls", 0, GLUIwindow_width + 120, 100);
  glui->set_main_gfx_window(GLUIwindow);

  // Add frame/element selections
  panel = glui->add_panel("Chord histogram Selection");
  spinner = glui->add_spinner_to_panel(panel, "Frame", GLUI_SPINNER_INT, &distance_index);
  spinner->set_int_limits(0, max_bin, GLUI_LIMIT_WRAP);
  spinner = glui->add_spinner_to_panel(panel, "Element", GLUI_SPINNER_INT, &element_index);
  spinner->set_int_limits(0, nelements-1, GLUI_LIMIT_WRAP);

  // Add button
  glui->add_button("Quit", QUIT_COMMAND, GLUICallback);
}



void GLUIInit(int *argc, char **argv)
{
  // Open window 
  glutInit(argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUIwindow_width, GLUIwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUIwindow = glutCreateWindow("OpenGL Viewer");

  // Initialize background color 
  glClearColor(200.0/255.0, 200.0/255.0, 200.0/255.0, 1.0);

  // Initialize lights
  static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glEnable(GL_LIGHT0);
  static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);
  glEnable(GL_LIGHTING);

  // Initialize graphics modes 
  glEnable(GL_DEPTH_TEST);

  // Initialize GLUT callback functions 
  glutDisplayFunc(GLUIRedraw);
  GLUI_Master.set_glutReshapeFunc(GLUIResize);
  GLUI_Master.set_glutKeyboardFunc(GLUIKeyboard);
  GLUI_Master.set_glutSpecialFunc(GLUISpecial);
  GLUI_Master.set_glutMouseFunc(GLUIMouse);
  glutMotionFunc(GLUIMotion);

  // Initialize font
#if (RN_OS == RN_WINDOWSNT)
  int font = glGenLists(256);
  wglUseFontBitmaps(wglGetCurrentDC(), 0, 256, font); 
  glListBase(font);
#endif
}



void GLUIMainLoop(void)
{
  // Setup GLUI controls
  GLUISetup();

  // Idle if making images
  if (make_images) {
    distance_index = 0;
    element_index = 0;
    GLUI_Master.set_glutIdleFunc(GLUIIdle);
  }

  // Run main loop -- never returns 
  glutMainLoop();
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: chdview [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) {
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-images")) {
        make_images = 1;
      }
      else if (!strcmp(*argv, "-background")) { 
        argc--; argv++; background_color[0] = atof(*argv); 
        argc--; argv++; background_color[1] = atof(*argv); 
        argc--; argv++; background_color[2] = atof(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!chd_name) chd_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
  }

  // Check filename
  if (!chd_name) {
    fprintf(stderr, "You did not specify a chord histogram file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Initialize GLUI
  GLUIInit(&argc, argv);

  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read chord file
  int status = ReadChordFile(chd_name);
  if (!status) exit(-1);

  // Run GLUT interface
  GLUIMainLoop();

  // Return success 
  return 0;
}



