// Source file for the site analyzer program



// Include files 

#include "R3Shapes/R3Shapes.h"
#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *grid_name = NULL;
static char *pdb_name = NULL;
static char *plot_name = NULL;
static char *ligand_name = NULL;
static char *element_name = NULL;
static RNBoolean largest_component_only = FALSE;
static RNBoolean print_verbose = FALSE;



static R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocated grid");
    return NULL;
  }

  // Read grid 
  if (!grid->ReadFile(grid_name)) {
    RNFail("Unable to read grid file %s", grid_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Dimensions = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Minimum = %g\n", grid->Minimum());
    printf("  Maximum = %g\n", grid->Maximum());
    printf("  Volume = %g\n", grid->Volume());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) {
    RNFail("Unable to read PDB file: %s", filename);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static RNArray<PDBAtom *> *
FindLigandAtoms(PDBFile *file, const char *ligand_name, char *element_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int residue_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Allocate array of atoms
  RNArray<PDBAtom *> *ligand_atoms = new RNArray<PDBAtom *>();
  if (!ligand_atoms) {
    fprintf(stderr, "Unable to allocate array of ligand atoms.\n");
    return NULL;
  }

  // Check ligand name
  if (ligand_name) {
    // Find ligand
    PDBResidue *ligand = file->FindResidue(ligand_name);
    if (!ligand) {
      fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
      return NULL;
    }

    // Add atoms of ligand to array
    residue_count = 1;
    for (int j = 0; j < ligand->NAtoms(); j++) {
      PDBAtom *atom = ligand->Atom(j);
      if (element_name && (!atom->Element()|| strcmp(atom->Element()->Name(), element_name))) continue;
      ligand_atoms->Insert(atom);
    }
  }
  else {
    // Add all hetatoms to array
    for (int i = 0; i < model->NResidues(); i++) {
      PDBResidue *residue = model->Residue(i);
      RNBoolean found = FALSE;
      for (int j = 0; j < residue->NAtoms(); j++) {
        PDBAtom *atom = residue->Atom(j);
        if (!atom->IsHetAtom()) continue;
        if (element_name && (!atom->Element()|| strcmp(atom->Element()->Name(), element_name))) continue;
        ligand_atoms->Insert(atom);
        found = TRUE;
      }
      if (found) residue_count++;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand atoms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Element = %s\n", (element_name) ? element_name : "All");
    printf("  # Ligands = %d\n", residue_count);
    printf("  # Atoms = %d\n", ligand_atoms->NEntries());
    fflush(stdout);
  }

  // Return ligand atoms
  return ligand_atoms;
}



static int 
PrintPlot(R3Grid *grid, PDBFile *file, RNArray<PDBAtom *> *ligand_atoms, RNBoolean largest_component_only, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdout;
  if (filename) {
    fp = fopen(filename, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open site plot file: %s\n", filename);
      return 0;
    }
  }

  // Write first entry
  fprintf(fp, "0 0\n");

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Create grid with ligand rasterized
  R3Grid ligand_grid(grid->XResolution(), grid->YResolution(), grid->ZResolution());
  ligand_grid.SetWorldToGridTransformation(grid->WorldToGridTransformation());
  for (int i = 0; i < ligand_atoms->NEntries(); i++) {
    PDBAtom *atom = ligand_atoms->Kth(i);
    ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
  }

  // Check ligand volume
  int ligand_nvalues = ligand_grid.Cardinality();
  if (ligand_nvalues > 0) {
    // Fill array of ligand values
    int count = 0;
    RNScalar *ligand_values = new RNScalar [ ligand_nvalues ];
    assert(ligand_values);
    for (int i = 0; i < grid->NEntries(); i++) {
      if (ligand_grid.GridValue(i) == 0) continue;
      assert(count < ligand_nvalues);
      RNScalar value = grid->GridValue(i);
      ligand_values[count++] = value;
    }

    // Sort array of values to check in ligand
    qsort(ligand_values, ligand_nvalues, sizeof(RNScalar), RNCompareScalars);

    // Determine if checking only largest component
    if (largest_component_only) {
      // Allocate arrays to store connected component information
      int *component_sizes = new int [ grid->NEntries() ];
      int *component_ids = new int [ grid->NEntries() ];
      assert(component_sizes && component_ids);

      // Print plot
      for (int ligand_index = ligand_nvalues-1; ligand_index >= 0; ligand_index--) {
        // Get threshold
        RNScalar threshold = ligand_values[ligand_index];
        if (threshold <= 0) break;

        // Find connected components at threshold
        int ncomponents = grid->ConnectedComponents(threshold, grid->NEntries(), NULL, component_sizes, component_ids);
        assert(ncomponents <= grid->NEntries());

        // Find largest connected component 
        int largest_component = -1;
        int largest_component_size = -1;
        for (int j = 0; j < ncomponents; j++) {
          if (component_sizes[j] > largest_component_size) {
            largest_component_size = component_sizes[j];
            largest_component = j;
          }
        }

        // Check if there was a maximum component
        if (largest_component < 0) break;

        // Count the number of cells that are inside the ligand and covered by the largest component
        int inside_nvalues = 0;
        int *component_idsp = component_ids;
        RNScalar *ligand_grid_valuesp = (RNScalar *) ligand_grid.GridValues();
        for (int j = 0; j < grid->NEntries(); j++) {
          if ((*component_idsp == largest_component) && (*ligand_grid_valuesp > 0)) inside_nvalues++;
          component_idsp++;
          ligand_grid_valuesp++;
        }

        // Determine fraction of ligand volume number of grid cells with value greater of equal to ligand value
        RNScalar site_fraction = (RNScalar) largest_component_size / (RNScalar) ligand_nvalues;

        // Convert number of grid cells in ligand and largest component to a fraction
        RNScalar inside_fraction = (RNScalar) inside_nvalues / (RNScalar) ligand_nvalues;

        // Print plot numbers
        fprintf(fp, "%g %g\n", inside_fraction, site_fraction);
      }

      // Delete arrays to store connected component information
      delete [] component_sizes;
      delete [] component_ids;
    }
    else {
      // Make sorted array of grid values
      int grid_nvalues = grid->NEntries();
      RNScalar *grid_values = new RNScalar [ grid_nvalues ];
      assert(grid_values);
      for (int i = 0; i < grid_nvalues; i++) grid_values[i] = grid->GridValue(i);
      qsort(grid_values, grid_nvalues, sizeof(RNScalar), RNCompareScalars);
      RNScalar grid_minimum = grid_values[0];

      // Print plot
      fprintf(fp, "0 0\n");
      int grid_index = grid_nvalues-1;
      for (int ligand_index = ligand_nvalues-1; ligand_index >= 0; ligand_index--) {
        // Get threshold
        RNScalar threshold = ligand_values[ligand_index];
        if (threshold <= grid_minimum) break;

        // Check if did not find any of grid value less or equal to ligand value
        if (grid_index < 0) break;

        // Find next grid value less or equal to ligand value
        for (; grid_index >= 0; grid_index--) {
          if (RNIsLess(grid_values[grid_index], ligand_values[ligand_index], 1.0E-8)) break;
        }

        // Check if did not find any of grid value less or equal to ligand value
        if (grid_index < 0) break;

        // Determine fraction of ligand volume number of grid cells with value greater of equal to ligand value
        RNScalar site_fraction = (RNScalar) (grid_nvalues - grid_index) / (RNScalar) ligand_nvalues;

        // Determine fraction of volume inside ligand covered by grid above threshold
        RNScalar inside_fraction = (RNScalar) (ligand_nvalues - ligand_index) / (RNScalar) ligand_nvalues;

        // Print plot numbers
        fprintf(fp, "%g %g\n", inside_fraction, site_fraction);
      }

      // Deallocate array of grid values
      delete [] grid_values;
    }

    // Delete array of ligand values
    delete [] ligand_values;
  }

  // Close file
  if (filename) fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote plot ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", ligand_nvalues);
    printf("  Ligand volume = %g\n", ligand_nvalues * volume_scale_factor);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: grd2plt gridfile pdbfile plotfile [-ligand ligandname]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = TRUE; 
      else if (!strcmp(*argv, "-ligand")) { argc--; argv++; ligand_name = *argv; }
      else if (!strcmp(*argv, "-largest_component_only")) largest_component_only = TRUE; 
      else if (!strcmp(*argv, "-element")) { argc--; argv++; element_name = *argv; }
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!grid_name) grid_name = *argv;
      else if (!pdb_name) pdb_name = *argv;
      else if (!plot_name) plot_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check plot filename
  if (!plot_name) {
    fprintf(stderr, "You did not specify a plot file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grid
  R3Grid *grid = ReadGrid(grid_name);
  if (!grid) exit(-1);

  // Read pdb
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Find ligand atoms
  RNArray<PDBAtom *> *ligand_atoms = FindLigandAtoms(file, ligand_name, element_name);
  if (!ligand_atoms) exit(-1);

  // Print plot
  int status = PrintPlot(grid, file, ligand_atoms, largest_component_only, plot_name);
  if (!status) exit(-1);

  // Return success 
  return 0;
}



