// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"



// Program variables

static char *grid_name = NULL;
static char *sfd_name = NULL;
static RNScalar gedt_sigma = 0;
static int print_verbose = 0;



static R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate a grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocate grid");
    return NULL;
  }

  // Read grid
  int status = grid->ReadFile(grid_name);
  if (!status) {
    RNFail("Unable to read grid file %s", grid_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static R3SphericalGrid *
CreateSphericalGrid(R3Grid *grid1)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical grid
  R3SphericalGrid *grid = new R3SphericalGrid(grid1->XResolution()/2, grid1->XResolution()/2, grid1->XResolution()/2);
  if (!grid) {
    fprintf(stderr, "Unable to allocate spherical grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(grid1->WorldBox());

  // Check sigma
  if (RNIsPositive(gedt_sigma)) {
    // Fill samples on spherical grid with Gaussian of DT
    const RNScalar sqrt_two_pi = sqrt(RN_TWO_PI);
    RNScalar a = sqrt_two_pi * gedt_sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = -2.0 * gedt_sigma * gedt_sigma;
    grid1->SquaredDistanceTransform();
    for (int k = 0; k < grid->RadiusResolution(); k++) {
      for (int j = 0; j < grid->ThetaResolution(); j++) {
        for (int i = 0; i < grid->PhiResolution(); i++) {
          R3Point world_position = grid->WorldPosition(i, j, k);
          RNScalar value1 = grid1->WorldValue(world_position);
          RNScalar value = fac * exp( value1 / denom );
          grid->SetGridValue(i, j, k, value);
        }
      }
    }
  }
  else {
    // Handle case of gedt_sigma=0 separarately
    for (int k = 0; k < grid->RadiusResolution(); k++) {
      for (int j = 0; j < grid->ThetaResolution(); j++) {
        for (int i = 0; i < grid->PhiResolution(); i++) {
          R3Point world_position = grid->WorldPosition(i, j, k);
          RNScalar value = grid1->WorldValue(world_position);
          grid->SetGridValue(i, j, k, value);
        }
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created spherical grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->PhiResolution(), grid->ThetaResolution(), grid->RadiusResolution());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return spherical grid
  return grid;
}



static R3SphericalHarmonics *
CreateSphericalHarmonics(R3SphericalGrid *grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical harmonics
  R3SphericalHarmonics *harmonics = new R3SphericalHarmonics(grid->ThetaResolution()/2, grid->RadiusResolution());
  if (!harmonics) {
    fprintf(stderr, "Unable to allocate spherical harmonics\n");
    exit(-1);
  }

  // Compute forward harmonic transform
  grid->ForwardSphericalHarmonicTransform(*harmonics);

  // Print statistics
  if (print_verbose) {
    printf("Created spherical harmonics ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Coefficients = %d\n", harmonics->NCoefficients());
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    printf("  L2Norm = %g\n", harmonics->L2Norm());
    fflush(stdout);
  }

  // Return spherical harmonics
  return harmonics;
}



static float *
CreateSphericalFourierDescriptor(R3SphericalHarmonics *harmonics)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate array of floats for descriptor
  float *sfd = new float [harmonics->RadiusResolution() * harmonics->NCoefficients() / 2 ];
  assert(sfd);

  // Compute L2Norms for each radius and frequency
  int nvalues = 0;
  for (int S = 0; S < harmonics->RadiusResolution(); S++) {
    for (int L = 0; L < harmonics->Bandwidth(); L++) {
      for (int M = -L; M <= L; M++) {
        sfd[nvalues++] = harmonics->L2Norm(S, L, M);
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created spherical fourier descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", nvalues);
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    fflush(stdout);
  }

  // Return descriptor
  return sfd;
}



static int
WriteSphericalFourierDescriptor(int nvalues, float *values, const char *sfd_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(sfd_name, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open SFD file %s\n", sfd_name);
    return 0;
  }

  // Write number of values
  if (fwrite(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to write number of values to SFD file %s\n", sfd_name);
    return 0;
  }

  // Write values
  if (fwrite(values, sizeof(float), nvalues, fp) != (unsigned int) nvalues) {
    fprintf(stderr, "Unable to write values to SFD file %s\n", sfd_name);
    return 0;
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", sizeof(int) + nvalues * sizeof(float));
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: grd2sfd gridfile sfdfile [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-sigma")) { argc--; argv++; gedt_sigma = atof(*argv); }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    else {
      if (!grid_name) grid_name = *argv;
      else if (!sfd_name) sfd_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    argv++; argc--;
  }

  // Check grid filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify a grid file.\n");
    return 0;
  }

  // Check sfd filename
  if (!sfd_name) {
    fprintf(stderr, "You did not specify a sfd file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grid file
  R3Grid *grid = ReadGrid(grid_name);
  if (!grid) exit(-1);

  // Create spherical grid
  R3SphericalGrid *sgrid = CreateSphericalGrid(grid);
  if (!sgrid) exit(-1);

  // Create spherical harmonics
  R3SphericalHarmonics *harmonics = CreateSphericalHarmonics(sgrid);
  if (!harmonics) exit(-1);

  // Create spherical harmonic descriptor
  float *sfd = CreateSphericalFourierDescriptor(harmonics);
  if (!sfd) exit(-1);

  // Write spherical harmonic descriptor
  int sfd_size = harmonics->RadiusResolution() * harmonics->NCoefficients() / 2;
  int status = WriteSphericalFourierDescriptor(sfd_size, sfd, sfd_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
