// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"



// Program variables

static RNArray<char *> grid_names;
static char *shd_name = NULL;
static RNScalar gedt_sigma = 0;
static int descriptor_type = 0; // 0=shd, 1=sfd, 2=sgd
static int normalization_type = 0;
static int print_verbose = 0;



static R3Grid *
ReadGrid(char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate a grid
  R3Grid *grid = new R3Grid();
  if (!grid) {
    RNFail("Unable to allocate grid");
    return NULL;
  }

  // Read grid
  int status = grid->ReadFile(grid_name);
  if (!status) {
    RNFail("Unable to read grid file %s", grid_name);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return success
  return grid;
}



static R3SphericalGrid *
CreateSphericalGrid(R3Grid *grid1)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical grid
  R3SphericalGrid *grid = new R3SphericalGrid(grid1->XResolution()/2, grid1->XResolution()/2, grid1->XResolution()/2);
  if (!grid) {
    fprintf(stderr, "Unable to allocate spherical grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(grid1->WorldBox());

  // Check sigma
  if (RNIsPositive(gedt_sigma)) {
    // Fill samples on spherical grid with Gaussian of DT
    const RNScalar sqrt_two_pi = sqrt(RN_TWO_PI);
    RNScalar a = sqrt_two_pi * gedt_sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = -2.0 * gedt_sigma * gedt_sigma;
    grid1->SquaredDistanceTransform();
    for (int k = 0; k < grid->RadiusResolution(); k++) {
      for (int j = 0; j < grid->ThetaResolution(); j++) {
        for (int i = 0; i < grid->PhiResolution(); i++) {
          R3Point world_position = grid->WorldPosition(i, j, k);
          RNScalar value1 = grid1->WorldValue(world_position);
          RNScalar value = fac * exp( value1 / denom );
          grid->SetGridValue(i, j, k, value);
        }
      }
    }
  }
  else {
    // Handle case of gedt_sigma=0 separarately
    for (int k = 0; k < grid->RadiusResolution(); k++) {
      for (int j = 0; j < grid->ThetaResolution(); j++) {
        for (int i = 0; i < grid->PhiResolution(); i++) {
          R3Point world_position = grid->WorldPosition(i, j, k);
          RNScalar value = grid1->WorldValue(world_position);
          grid->SetGridValue(i, j, k, value);
        }
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created spherical grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->PhiResolution(), grid->ThetaResolution(), grid->RadiusResolution());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return spherical grid
  return grid;
}



static R3SphericalHarmonics *
CreateSphericalHarmonics(R3SphericalGrid *grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical harmonics
  R3SphericalHarmonics *harmonics = new R3SphericalHarmonics(grid->ThetaResolution()/2, grid->RadiusResolution());
  if (!harmonics) {
    fprintf(stderr, "Unable to allocate spherical harmonics\n");
    exit(-1);
  }

  // Compute forward harmonic transform
  grid->ForwardSphericalHarmonicTransform(*harmonics);

  // Print statistics
  if (print_verbose) {
    printf("Created spherical harmonics ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Coefficients = %d\n", harmonics->NCoefficients());
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    printf("  L2Norm = %g\n", harmonics->L2Norm());
    fflush(stdout);
  }

  // Return spherical harmonics
  return harmonics;
}



static int
CreateDescriptor(R3SphericalHarmonics *harmonics, int descriptor_type, float *values)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  float sum = 0.0;

  // Initialize counts
  int nvalues = 0;
  int count = 0;

  // Check descriptor type
  if (descriptor_type == 0) {
    // Compute number of values
    nvalues = harmonics->RadiusResolution() * harmonics->Bandwidth();

    // Compute L2Norm for each radius and frequency
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        float value_squared = harmonics->L2NormSquared(S, L);
        float value = sqrt(value_squared);
        assert(count <= nvalues);
        values[count++] = value;
        sum += value_squared;
      }
    }
  }
  else if (descriptor_type == 1) {
    // Compute number of values
    nvalues = harmonics->NCoefficients() / 2;

    // Compute L2Norm for each coefficient
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        for (int M = -L; M <= L; M++) {
          float value_squared = harmonics->L2NormSquared(S, L, M);
          float value = sqrt(value_squared);
          assert(count <= nvalues);
          values[count++] = value;
          sum += value_squared;
        }
      }
    }
  }
  else if (descriptor_type == 2) {
    // Compute number of values
    nvalues = harmonics->NCoefficients();

    // Compute coefficients for each radius and frequency
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        for (int M = -L; M <= L; M++) {
          for (int k = 0; k < 2; k++) {
            float value = harmonics->Shell(S).Coefficient(L, M, k);
            float value_squared = value * value;
            assert(count <= nvalues);
            values[count++] = value;
            sum += value_squared;
          }
        }
      }
    }
  }
  else {
    RNAbort("Invalid descriptor type");
  }

  // Compute l2norm
  assert(nvalues > 0);
  assert(count == nvalues);
  RNScalar l2norm = sqrt(sum);

  // Normalize descriptor
  if ((normalization_type % 2) == 1) {
    if (l2norm > 0) {
      for (int i = 0; i < nvalues; i++) {
        values[i] /= l2norm;
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", nvalues);
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    printf("  L2Norm = %g\n", l2norm);
    fflush(stdout);
  }

  // Return number of values
  return nvalues;
}



static int
NormalizeDescriptor(int nvalues, float *values, int normalization_type)
{
  // Check if normalizing
  if (normalization_type < 2) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute l2norm
  float l2norm = 0;
  for (int i = 0; i < nvalues; i++)
    l2norm += values[i] * values[i];
  l2norm = sqrt(l2norm);

  // Normalize descriptor
  if (l2norm > 0) {
    for (int i = 0; i < nvalues; i++)
      values[i] /= l2norm;
  }

  // Print statistics
  if (print_verbose) {
    printf("Normalized spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    printf("  L2Norm = %g\n", l2norm);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WriteDescriptor(int nvalues, float *values, const char *shd_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(shd_name, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open SHD file %s\n", shd_name);
    return 0;
  }

  // Write number of values
  if (fwrite(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to write number of values to SHD file %s\n", shd_name);
    return 0;
  }

  // Write values
  if (fwrite(values, sizeof(float), nvalues, fp) != (unsigned long) nvalues) {
    fprintf(stderr, "Unable to write values to SHD file %s\n", shd_name);
    return 0;
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", sizeof(int) + nvalues * sizeof(float));
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: grd2shd gridfile shdfile [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-sig")) { argc--; argv++; descriptor_type = atoi(*argv); }
      else if (!strcmp(*argv, "-normalize")) { argc--; argv++; normalization_type = atoi(*argv); }
      else if (!strcmp(*argv, "-sigma")) { argc--; argv++; gedt_sigma = atof(*argv); }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    else {
      grid_names.Insert(*argv);
    }
    argv++; argc--;
  }

  // Check grid filenames
  if (grid_names.IsEmpty()) {
    fprintf(stderr, "You did not specify a map or grid file.\n");
    return 0;
  }

  // Check shd filename
  if (grid_names.NEntries() < 2) {
    fprintf(stderr, "You did not specify a shd file.\n");
    return 0;
  }

  // Get shd filename
  shd_name = grid_names.Tail();
  grid_names.RemoveTail();

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Allocate buffer for descriptor
  const int max_descriptor_size = 16 * 1024 * 1024;
  static float descriptor[max_descriptor_size];
  int descriptor_size = 0;

  // Compute/concatenate SHDs for all grids
  for (int i = 0; i < grid_names.NEntries(); i++) {
    // Get grid name
    char *grid_name = grid_names[i];

    // Read grid file
    R3Grid *grid = ReadGrid(grid_name);
    if (!grid) exit(-1);

    // Create spherical grid
    R3SphericalGrid *sgrid = CreateSphericalGrid(grid);
    if (!sgrid) exit(-1);

    // Create spherical harmonics
    R3SphericalHarmonics *harmonics = CreateSphericalHarmonics(sgrid);
    if (!harmonics) exit(-1);

    // Create spherical harmonic descriptor
    int nvalues = CreateDescriptor(harmonics, descriptor_type, &descriptor[descriptor_size]);
    if (nvalues == 0) exit(-1);

    // Update descriptor size
    descriptor_size += nvalues;
    assert(descriptor_size < max_descriptor_size);
  }

  // Normalize descriptor
  int status1 = NormalizeDescriptor(descriptor_size, descriptor, normalization_type);
  if (!status1) exit(-1);

  // Write descriptor
  int status2 = WriteDescriptor(descriptor_size, descriptor, shd_name);
  if (!status2) exit(-1);

  // Return success
  return 0;
}
