// Source file for the mesh to grid conversion program



// Include files 

#include "R3Grids/R3Grids.h"



// Program variables

static char *kont_name = NULL;
static char *grid_name = NULL;
static int grid_resolution[3] = { 0, 0, 0 };
static R3Box world_box = R3null_box;
static int print_verbose = 0;



static R3Grid *
ReadKont(char *kont_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  char buffer[1024];
  int line_counter = 0;
  
  // Open file
  FILE *fp = fopen(kont_name, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open %s\n", kont_name);
    return NULL;
  }

  // Create grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    return NULL;
  }

  // Set grid transformation
  grid->SetWorldToGridTransformation(world_box);

  // Read grid positions from input file
  for (int k = 0; k < grid->ZResolution(); k++) {
    for (int i = 0; i < grid->XResolution(); i++) {
      for (int j = 0; j < grid->YResolution(); j++) {
        // Read line from file
        line_counter++;
        if (!fgets(buffer, 1023, fp)) {
          fprintf(stderr, "Unable to read grid position at (%d, %d, %d) from %s\n", i, j, k, kont_name);
          return NULL;
        }

        // Parse line
        int counter;
        RNScalar x, y, z;
        if (sscanf(buffer, "%d%lf%lf%lf", &counter, &x, &y, &z) != 4) {
          fprintf(stderr, "Unable to parse grid position at (%d, %d, %d) in %s\n", i, j, k, kont_name);
          return NULL;
        }

        // Confirm position
        R3Point position = grid->WorldPosition(i, j, k);
        if (RNIsNotEqual(position[0], x) || RNIsNotEqual(position[1], y) || RNIsNotEqual(position[2], z)) {
          fprintf(stderr, "Mismatched grid position at (%d, %d, %d) in %s: %g %g %g  vs. %g %g %g\n", i, j, k, kont_name,
            x, y, z, position[0], position[1], position[2]);
          return NULL;
        }
      }
    }
  }

  // Read header line from file
  line_counter++;
  if (!fgets(buffer, 1023, fp)) {
    fprintf(stderr, "Unable to read header line from %s\n", kont_name);
    return NULL;
  }

  // Check header line
  if (!strstr(buffer, "HEADER")) {
    fprintf(stderr, "Unable to parse HEADER (%s) in %s\n", buffer, kont_name);
    return NULL;
  }

  // Read grid values from input file
  for (int k = 0; k < grid->ZResolution(); k++) {
    for (int i = 0; i < grid->XResolution(); i++) {
      for (int j = 0; j < grid->YResolution(); j++) {
        // Read line from file
        line_counter++;
        if (!fgets(buffer, 1023, fp)) {
          fprintf(stderr, "Unable to read grid position at (%d, %d, %d) from %s\n", i, j, k, kont_name);
          return NULL;
        }

        // Assign grid value
        grid->SetGridValue(i, j, k, atof(buffer));
      }
    }
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Read grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    const R3Box bbox = grid->WorldBox();
    printf("  World Box = ( %g %g %g ) ( %g %g %g )\n", bbox[0][0], bbox[0][1], bbox[0][2], bbox[1][0], bbox[1][1], bbox[1][2]);
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



static int 
WriteGrid(R3Grid *grid, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grid
  int status = grid->WriteFile(grid_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", status * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: kont2grd kontfile grdfile -box x1 y1 z1 x2 y2 z2 -resolution xres yres zres [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-box")) { 
        argc--; argv++; world_box[0][0] = atof(*argv); 
        argc--; argv++; world_box[0][1] = atof(*argv); 
        argc--; argv++; world_box[0][2] = atof(*argv); 
        argc--; argv++; world_box[1][0] = atof(*argv); 
        argc--; argv++; world_box[1][1] = atof(*argv); 
        argc--; argv++; world_box[1][2] = atof(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); 
      }
    }
    else {
      if (!kont_name) kont_name = *argv;
      else if (!grid_name) grid_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
    }
    argv++; argc--;
  }

  // Check input filename
  if (!kont_name) {
    fprintf(stderr, "You did not specify an input file.\n");
    return 0;
  }

  // Check output filename
  if (!grid_name) {
    fprintf(stderr, "You did not specify an output file.\n");
    return 0;
  }

  // Check resolution
  if ((grid_resolution[0] <= 0) || (grid_resolution[1] <= 0) || (grid_resolution[2] <= 0)) {
    fprintf(stderr, "You must specify a grid resolution with -resolution x y z\n");
    return 0;
  }

  // Check bounding box
  if (world_box.IsEmpty()) {
    fprintf(stderr, "You must specify a world box with -box x1 y1 z1 x2 y2 z2\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grid 
  R3Grid *grid = ReadKont(kont_name);
  if (!grid) exit(-1);

  // Write grid
  int status = WriteGrid(grid, grid_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
