static int 
ProcessMap(Map *map, RNBoolean aggregate_aminoacid_groups, RNBoolean aggregate_spherical_shells)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Aggregate grids within each amino acid group
  if (aggregate_aminoacid_groups) {
    // Allocate a grid per group
    R3Grid *grids[PDB_NUM_AMINOACID_GROUPS];
    for (int group = 0; group < PDB_NUM_AMINOACID_GROUPS; group++) {
      grids[group] = new R3Grid(map->XResolution(), map->YResolution(), map->ZResolution());
      assert(grids[group]);
      grids[group]->SetWorldToGridTransformation(map->grids[0][0]->WorldToGridTransformation());
    }

    // Aggregate grids per group
    for (int j = 0; j < map->NElements(); j++) {
      // Initialize grid for each group
      for (int group = 0; group < PDB_NUM_AMINOACID_GROUPS; group++) {
        grids[group]->Clear();
      }

      // Add grids within each group
      for (int i = 0; i < map->NAminoAcids(); i++) {
        PDBAminoAcidGroup group = PDBaminoacids[i].Group();
        grids[group]->Add(*(map->grids[i][j]));
      }

      // Replace grids with group's
      for (int i = 0; i < map->NAminoAcids(); i++) {
        PDBAminoAcidGroup group = PDBaminoacids[i].Group();
        *(map->grids[i][j]) = *(grids[group]);
      }
    }

    // Delete grid per group
    for (int group = 0; group < PDB_NUM_AMINOACID_GROUPS; group++) {
      delete grids[group];
    }
  }

  // Aggregate grids within each spherical band
  if (aggregate_spherical_shells) {
    // Allocate histogram
    int nbins = map->XResolution()/2;
    RNScalar *histogram = new RNScalar [nbins];
    assert(histogram);

    // Compute grid centroid
    RNScalar cx = 0.5*(map->XResolution()-1);
    RNScalar cy = 0.5*(map->YResolution()-1);
    RNScalar cz = 0.5*(map->ZResolution()-1);

    // Agreggate grids with each spherical band
    for (int i = 0; i < map->NAminoAcids(); i++) {
      for (int j = 0; j < map->NElements(); j++) {
        // Initialize histogram
        for (int k = 0; k < nbins; k++) {
          histogram[k] = 0;
        }

        // Compute histogram 
        for (int x = 0; x < map->Grid(i, j)->XResolution(); x++) {
          RNScalar dx = x - cx;
          RNScalar dx_squared = dx * dx;
          for (int y = 0; y < map->Grid(i, j)->YResolution(); y++) {
            RNScalar dy = y - cx;
            RNScalar dy_squared = dy * dy;
            for (int z = 0; z < map->Grid(i, j)->ZResolution(); z++) {
              RNScalar dz = z - cx;
              RNScalar dz_squared = dz * dz;
              RNScalar distance_squared = dx_squared + dy_squared + dz_squared;
              RNScalar distance = sqrt(distance_squared);
              int bin = distance / nbins;
              if (bin >= nbins) continue;
              histogram[bin] += map->Grid(i, j)->GridValue(x, y ,z);
            }
          }
        }

        // Replace grid values
        for (int x = 0; x < map->Grid(i, j)->XResolution(); x++) {
          RNScalar dx = x - cx;
          RNScalar dx_squared = dx * dx;
          for (int y = 0; y < map->Grid(i, j)->YResolution(); y++) {
            RNScalar dy = y - cx;
            RNScalar dy_squared = dy * dy;
            for (int z = 0; z < map->Grid(i, j)->ZResolution(); z++) {
              RNScalar dz = z - cx;
              RNScalar dz_squared = dz * dz;
              RNScalar distance_squared = dx_squared + dy_squared + dz_squared;
              RNScalar distance = sqrt(distance_squared);
              int bin = distance / nbins;
              if (bin >= nbins) continue;
              map->grids[i][j]->SetGridValue(x, y ,z, histogram[bin]);
            }
          }
        }
      }
    }

    // Delete histogram
    delete [] histogram;
  }

  // Print statistics
  if (print_verbose) {
    printf("Processed map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Aggregate Groups = %d\n", aggregate_aminoacid_groups);
    printf("  Aggregate Shells = %d\n", aggregate_spherical_shells);
    fflush(stdout);
  }

  // Return success
  return 1;
}



