// Source file for the pdb viewer program



// Include files 

#include "PDB/PDB.h"
#include "R3Grids/R3Grids.h"



// Program variables

static char *pdb_names = NULL;
static char *histogram_name = NULL;
static char *ligand_names = NULL;
static char *asa_names = NULL;
static char *consurf_names = NULL;
static int nelements = 1;
static RNScalar min_distance = 0;
static RNScalar max_distance = 20;
static RNScalar nearby_threshold = 1;
static RNScalar bin_spacing = 1;
static RNBoolean weight_spans_by_surface_area = FALSE;
static RNBoolean weight_spans_by_conservation = FALSE;
static int print_verbose = 0;



// Histogram data

static const int max_distance_hits = 1024;
static const int max_parameter_hits = max_distance_hits;
static const int max_element_hits = 6;
static RNScalar hits[max_distance_hits][max_parameter_hits][max_element_hits] = { 0 };
static RNScalar observations[max_distance_hits] = { 0 };



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Check models
  if (file->NModels() == 0) {
    fprintf(stderr, "PDB File %s has no models\n", filename);
    return 0;
  }

  // Check atoms
  if (file->Model(0)->NAtoms() == 0) {
    fprintf(stderr, "PDB File %s has no atoms\n", filename);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("  Read PDB file %s ...\n", filename);
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("    Model %s ...\n", model->Name());
      printf("    # Chains = %d\n", model->NChains());
      printf("    # Residues = %d\n", model->NResidues());
      printf("    # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static PDBResidue *
FindLigand(PDBFile *file, char *ligand_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Find ligand
  PDBResidue *ligand = file->FindResidue(ligand_name);
  if (!ligand) {
    fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, file->Name()); 
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("  Found ligand %s ...\n", ligand_name);
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Atoms = %d\n", ligand->NAtoms());
    fflush(stdout);
  }

  // Return ligand
  return ligand;
}



static int
ReadASAFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read ASA file
  int natoms = file->ReadASAFile(filename);
  if (natoms == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("  Read accessible surface area file %s ...\n", filename);
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Surface Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadConsurfFiles(PDBFile *file, char *hssp_basename)
{
  // Check number of models
  if (file->NModels() < 1) {
    fprintf(stderr, "File must have at least one model to read conservation scores: %s.\n", hssp_basename);
    return 0;
  }

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read conservation files
  int nresidues = file->ReadConsurfFiles(hssp_basename);

  // Print statistics
  if (print_verbose && (nresidues > 0)) {
    printf("Read consurf files ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Residues = %d\n", nresidues);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
CreateHistogram(RNScalar hits[max_distance_hits][max_parameter_hits][max_element_hits], RNScalar observations[max_distance_hits],
  int nelements, RNScalar max_distance, RNScalar bin_spacing, RNScalar nearby_threshold, 
  char *pdb_names, char *ligand_names, char *asa_names)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int file_count = 0;
  int span_count = 0;
  int sample_count = 0;

  // Open the pdb file list
  FILE *pdb_list = fopen(pdb_names, "r");
  if (!pdb_list) {
    fprintf(stderr, "Unable to open list of pdb files: %s\n", pdb_names);
    return 0;
  }

  // Open the ligand list
  FILE *ligand_list = NULL;
  if (ligand_names) {
    ligand_list = fopen(ligand_names, "r");
    if (!ligand_list) {
      fprintf(stderr, "Unable to open list of ligands: %s\n", ligand_names);
      return 0;
    }
  }

  // Open the asa list
  FILE *asa_list = NULL;
  if (asa_names) {
    asa_list = fopen(asa_names, "r");
    if (!asa_list) {
      fprintf(stderr, "Unable to open list of asa files: %s\n", asa_names);
      return 0;
    }
  }

  // Open the consurf list
  FILE *consurf_list = NULL;
  if (consurf_names) {
    consurf_list = fopen(consurf_names, "r");
    if (!consurf_list) {
      fprintf(stderr, "Unable to open list of consurf files: %s\n", consurf_names);
      return 0;
    }
  }

  // Read pdb files and add them to the map
  char pdb_name[256];
  while (fscanf(pdb_list, "%s", pdb_name) == 1) {
    // Increment file counter
    file_count++;

    // Print message
    if (print_verbose) {
      printf("Processing %d ...\n", file_count);
    }

    // Read PDB file
    PDBFile *file = ReadPDB(pdb_name);
    if (!file) return 0;

    // Find ligand
    PDBResidue *ligand = NULL;
    if (ligand_names) {
      char ligand_name[128];
      if (fscanf(ligand_list, "%s", ligand_name) == 1) {
        ligand = FindLigand(file, ligand_name);
        if (!ligand) return 0;
      }
      else {
        fprintf(stderr, "Not enough entries in ligand list: %s\n", ligand_names);
        return 0;
      }
    }

    // Read accessible surface area file
    if (asa_names) {
      char asa_name[128];
      if (fscanf(asa_list, "%s", asa_name) == 1) {
        int status = ReadASAFile(file, asa_name);
        if (!status) return 0;
      }
      else {
        fprintf(stderr, "Not enough entries in asa file list: %s\n", asa_names);
        return 0;
      }
    }

    // Read conservation file
    if (consurf_names) {
      char consurf_name[128];
      if (fscanf(consurf_list, "%s", consurf_name) == 1) {
        int status = ReadConsurfFiles(file, consurf_name);
        if (!status) return 0;
      }
      else {
        fprintf(stderr, "Not enough entries in consurf file list: %s\n", consurf_names);
        return 0;
      }
    }

    // Start file statistics
    RNTime file_start_time;
    file_start_time.Read();
    int file_span_count = 0;
    int file_sample_count = 0;

    // Get model from PDB file
    if (file->NModels() == 0) return 0;
    PDBModel *model = file->Model(0);

    // Compute mask (1 outside protein, and 0 inside)
    RNScalar grid_spacing = 0.5;
    R3Box world_box = model->BBox();
    int xres = (int) (world_box.XLength() / grid_spacing + 1); if (xres > 256) xres = 256;
    int yres = (int) (world_box.YLength() / grid_spacing + 1); if (yres > 256) yres = 256;
    int zres = (int) (world_box.ZLength() / grid_spacing + 1); if (zres > 256) zres = 256;
    R3Grid mask(xres, yres, zres);
    mask.SetWorldToGridTransformation(world_box);
    for (int i = 0; i < model->NAtoms(); i++) {
      PDBAtom *atom = model->Atom(i);
      if (atom->IsHetAtom()) continue;
      mask.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1);
    }
    mask.Threshold(0.5, 1, 0);

    // Make arrays of atoms in model
    RNArray<PDBAtom *> ligand_atoms;
    RNArray<PDBAtom *> protein_atoms;
    for (int i = 0; i < model->NAtoms(); i++) {
      PDBAtom *atom = model->Atom(i);
      if (atom->IsHetAtom()) {
        if (!ligand || (atom->Residue() == ligand)) {
          ligand_atoms.Insert(atom);
        }
      }
      else {
        if (atom->accessible_surface_area != 0) {
          protein_atoms.Insert(atom);
        }
      }
    }

    // Consider span between all pairs of atoms on protein surface
    for (int i = 0; i < protein_atoms.NEntries(); i++) {
      PDBAtom *atom1 = protein_atoms.Kth(i);
      for (int j = i+1; j < protein_atoms.NEntries(); j++) {
        PDBAtom *atom2 = protein_atoms.Kth(j);

        // Get span and check its length
        R3Span span(atom1->position, atom2->position);
        RNLength distance = span.Length();
        if (distance <= min_distance) continue;
        if (distance >= max_distance) continue;

        // Compute visibility of atom1 and atom2
        RNBoolean visible = FALSE;
        RNScalar t_step = mask.GridToWorldScaleFactor();
        RNScalar t_start = atom1->Radius() + t_step;
        RNScalar t_stop = span.Length() - atom2->Radius() - t_step; 
        for (RNScalar t = t_start; t <= t_stop; t += t_step) {
          R3Point world_position = span.Point(t);
          R3Point grid_position = mask.GridPosition(world_position);
          int x = (int) (grid_position[0] + 0.5);
          int y = (int) (grid_position[1] + 0.5);
          int z = (int) (grid_position[2] + 0.5);
          if (mask.GridValue(x, y, z) == 0) { visible = FALSE; break; }
          else visible = TRUE;
        }

        // Check visibility
        if (!visible) continue;

        // Determine distance bin
        int distance_bin = (int) (distance / bin_spacing + 0.5);
        assert(distance_bin < max_distance_hits);

        // Determine weight of the observation
        RNScalar weight = 1;
        if (weight_spans_by_surface_area) {
          if (atom1->accessible_surface_area != PDB_UNKNOWN) weight *= atom1->accessible_surface_area;
          if (atom2->accessible_surface_area != PDB_UNKNOWN) weight *= atom2->accessible_surface_area;
        }
        if (weight_spans_by_conservation) {
          if (atom1->conservation == PDB_UNKNOWN) weight *= 0.1;
          else weight *= atom1->conservation;
          if (atom2->conservation == PDB_UNKNOWN) weight *= 0.1;
          else weight *= atom2->conservation;
        }

        // Log observation
        observations[distance_bin] += weight;

        // Consider all hetatoms
        for (int k = 0; k < ligand_atoms.NEntries(); k++) {
          PDBAtom *atom3 = ligand_atoms.Kth(k);

          // Check if atom3 is within distance threshold of span
          RNLength offset_distance = R3Distance(atom3->Position(), span.Line());
          if (offset_distance > nearby_threshold) continue;

          // Compute parametric value along span
          RNScalar parameter = span.T(atom3->Position());
          if ((parameter < 0) || (parameter > distance)) continue;

          // Determine element index
          PDBElement *element3 = atom3->Element();
          int element3_bin = (element3) ? element3 - PDBelements : 0;
          assert((element3_bin >= 0) && (element3_bin < PDBnelements));
          if (element3_bin >= nelements) element3_bin = 0;

          // Determine parameter bin for atom1->atom2
          RNScalar parameter_bin = parameter / bin_spacing;
          int parameter_bin1 = (int) parameter_bin;
          int parameter_bin2 = parameter_bin1 + 1;
          RNScalar parameter_remainder = parameter_bin - parameter_bin1;
          assert(parameter_bin2 < max_parameter_hits);

          // Add sample to histogram for atom1->atom2
          hits[distance_bin][parameter_bin1][element3_bin] += weight * (1 - parameter_remainder);
          hits[distance_bin][parameter_bin2][element3_bin] += weight * parameter_remainder;

          // Determine parameter bin for atom2->atom1
          parameter_bin = (distance - parameter) / bin_spacing;
          parameter_bin1 = (int) parameter_bin;
          parameter_bin2 = parameter_bin1 + 1;
          parameter_remainder = parameter_bin - parameter_bin1;
          assert(parameter_bin2 < max_parameter_hits);

          // Add sample to histogram for atom2->atom1
          hits[distance_bin][parameter_bin1][element3_bin] += weight * (1 - parameter_remainder);
          hits[distance_bin][parameter_bin2][element3_bin] += weight * parameter_remainder;

          // Update statistics
          file_sample_count++;
          sample_count++;
        }

        // Update statistics
        file_span_count++;
        span_count++;
      }
    }

    // Print message
    if (print_verbose) {
      printf("  Processed %s ...\n", file->Name());
      printf("    Time = %.2f seconds\n", file_start_time.Elapsed());
      printf("    # File Spans = %d\n", file_span_count);
      printf("    # File Samples = %d\n", file_sample_count);
    }

    // Delete the PDB file
    delete file;
  }

  // Close the files
  fclose(pdb_list);
  if (ligand_names) fclose(ligand_list);
  if (asa_names) fclose(asa_list);
  if (consurf_names) fclose(consurf_list);

  // Print statistics
  if (print_verbose) {
    printf("Created histogram ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Elements = %d\n", nelements);
    printf("  Min Distance = %g\n", min_distance);
    printf("  Max Distance = %g\n", max_distance);
    printf("  Nearby Threshold = %g\n", nearby_threshold);
    printf("  Bin Spacing = %g\n", bin_spacing);
    printf("  # Files = %d\n", file_count);
    printf("  # Spans = %d\n", span_count);
    printf("  # Samples = %d\n", sample_count);
    fflush(stdout);
  }

  // Return success
  return sample_count;
}



static int 
WriteHistogram(RNScalar hits[max_distance_hits][max_parameter_hits][max_element_hits], RNScalar observations[max_distance_hits],
  int nelements, RNScalar max_distance, RNScalar bin_spacing, const char *histogram_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(histogram_name, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open histogram file: %s\n", histogram_name);
    return 0;
  }

  // Write histogram
  int nhits = 0;
  RNScalar max_value = 0;
  RNScalar sum = 0;
  int max_bin = (int) (max_distance / bin_spacing);
  assert(max_bin < max_distance_hits);
  fprintf(fp, "%d %g %g %g %g %d\n", max_bin, min_distance, max_distance, bin_spacing, nearby_threshold, nelements);
  for (int i = 0; i <= max_bin; i++) {
    fprintf(fp, "%d %g : ", i, observations[i]);
    for (int j = 0; j <= i; j++) {
      for (int k = 0; k < nelements; k++) {
        RNScalar probability = (RNIsPositive(observations[i])) ? hits[i][j][k] / observations[i] : 0;
        fprintf(fp, "%g ", probability);
        if (hits[i][j][k] > max_value) max_value = hits[i][j][k];
        sum += hits[i][j][k];
        nhits++;
      }
      fprintf(fp, " ");
    }
    fprintf(fp, "\n");
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote histogram ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Hits = %d\n", nhits);
    printf("  Max value = %g\n", max_value);
    printf("  L1Norm = %g\n", sum);
    fflush(stdout);
  }

  // Return success
  return nhits;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdb2chd listfile plot [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-ligands")) { 
        argc--; argv++; ligand_names = *argv; 
      }
      else if (!strcmp(*argv, "-asas")) { 
        argc--; argv++; asa_names = *argv; 
      }
      else if (!strcmp(*argv, "-consurfs")) { 
        argc--; argv++; consurf_names = *argv; 
      }
      else if (!strcmp(*argv, "-nelements")) { 
        argc--; argv++; nelements = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-min_distance")) { 
        argc--; argv++; min_distance = atof(*argv); 
      }
      else if (!strcmp(*argv, "-max_distance")) { 
        argc--; argv++; max_distance = atof(*argv); 
      }
      else if (!strcmp(*argv, "-bin_spacing")) { 
        argc--; argv++; bin_spacing = atof(*argv); 
      }
      else if (!strcmp(*argv, "-nearby_threshold")) { 
        argc--; argv++; nearby_threshold = atof(*argv); 
      }
      else if (!strcmp(*argv, "-weight_spans_by_surface_area")) { 
        weight_spans_by_surface_area = TRUE;
      }
      else if (!strcmp(*argv, "-weight_spans_by_conservation")) { 
        weight_spans_by_conservation = TRUE;
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_names) pdb_names = *argv;
      else if (!histogram_name) histogram_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check list filename
  if (!pdb_names) {
    fprintf(stderr, "You did not specify a list of pdb files.\n");
    return 0;
  }

  // Check histogram filename
  if (!histogram_name) {
    fprintf(stderr, "You did not specify a map filename.\n");
    return 0;
  }

  // Check number of elements
  if (nelements > max_element_hits) {
    fprintf(stderr, "Invalid number of elements (%d)\n", nelements);
    return 0;
  }

  // Check bin_spacing
  if (max_distance / bin_spacing > max_distance_hits) {
    fprintf(stderr, "Unable to set bin_spacing (%g) so fine for such a large range (%g %g).\n", bin_spacing, min_distance, max_distance);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Create histogram
  int status1 = CreateHistogram(hits, observations, nelements, max_distance, bin_spacing, nearby_threshold, pdb_names, ligand_names, asa_names);
  if (!status1) exit(-1);

  // Write histogram
  int status2 = WriteHistogram(hits, observations, nelements, max_distance, bin_spacing, histogram_name);
  if (!status2) exit(-1);

  // Return success
  return 0;
}
