// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"
#include "KDTree.h"
#include "map.h"



// Program variables

static char *pdb_names = NULL;
static char *map_name = NULL;
static char *ligand_names = NULL;
static char *asa_names = NULL;
static int frame_type = RESIDUE_MAP_FRAME; 
static int atom_type = AMINOACID_MAP_ATOM; 
static int storage_type = POINT_MAP_STORAGE;
static int nelements = 6;
static int resolution = 32;
static RNBoolean bonded_protein_atoms_only = FALSE;
static RNBoolean bonded_ligand_atoms_only = FALSE;
static RNLength world_radius = 10;
static RNLength nearby_radius = 3.5;
static RNLength max_bond_length = 1;
static RNBoolean randomize_atom_positions = FALSE;
static RNLength world_sigma = 0;
static RNBoolean blend = FALSE;
static RNBoolean normalization_type = 0;
static RNBoolean normalize_by_ligand = 0;
static int print_verbose = 0;
static int print_debug = 0;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Check models
  if (file->NModels() == 0) {
    fprintf(stderr, "PDB File %s has no models\n", filename);
    return 0;
  }

  // Check atoms
  if (file->Model(0)->NAtoms() == 0) {
    fprintf(stderr, "PDB File %s has no atoms\n", filename);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("  Read PDB file %s ...\n", filename);
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("    Model %s ...\n", model->Name());
      printf("    # Chains = %d\n", model->NChains());
      printf("    # Residues = %d\n", model->NResidues());
      printf("    # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static PDBResidue *
FindLigand(PDBFile *file, char *ligand_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Find ligand
  PDBResidue *ligand = file->FindResidue(ligand_name);
  if (!ligand) {
    fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, file->Name()); 
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("  Found ligand %s ...\n", ligand_name);
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Atoms = %d\n", ligand->NAtoms());
    fflush(stdout);
  }

  // Return ligand
  return ligand;
}



static int
ReadASAFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read ASA file
  int natoms = file->ReadASAFile(filename);
  if (natoms == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("  Read accessible surface area file %s ...\n", filename);
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Surface Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return success
  return 1;
}



int
AddSiteToMap(Map *map, PDBFile *file, PDBResidue *ligand)
{
  // Check ligand
  if (!ligand) return 0;

  // Start statistics
  RNTime start_time;
  start_time.Read();
  int atom_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Get ligand origin
  R3Point ligand_centroid = ligand->Centroid();

  // Consider each protein residue
  for (int i = 0; i < model->NResidues(); i++) {
    PDBResidue *residue = model->Residue(i);
    if (residue == ligand) continue;
    PDBAminoAcid *aminoacid = residue->AminoAcid();
    if (!aminoacid) continue;
    if (R3Distance(ligand_centroid, residue->BBox()) > world_radius) continue;
    RNLength distance = PDBDistance(residue, ligand);
    if (bonded_ligand_atoms_only && (distance > max_bond_length)) continue;
    if (distance > nearby_radius) continue;

    // Consider each atom in residue
    for (int j = 0; j < residue->NAtoms(); j++) {
      PDBAtom *atom = residue->Atom(j);
      if (atom->IsHetAtom()) continue;
      if ((asa_names) && (atom->accessible_surface_area == 0)) continue;
      if (R3Distance(ligand_centroid, atom->Position()) > world_radius) continue;
      RNLength distance = PDBDistance(atom, ligand);
      if (bonded_protein_atoms_only && (distance > max_bond_length)) continue;
      if (distance > nearby_radius) continue;
      atom_count++;
  
      // Transform atom position into frame
      R3Point atom_position = atom->Position();
      atom_position -= ligand_centroid.Vector();

      // Randomize atom position?
      if (randomize_atom_positions) {
        RNCoord x = 2 * world_radius * (RNRandomScalar() - 0.5);
        RNCoord y = 2 * world_radius * (RNRandomScalar() - 0.5);
        RNCoord z = 2 * world_radius * (RNRandomScalar() - 0.5);
        atom_position.Reset(x, y, z);
      }

      // Determine frame index
      int frame_index = 0;

      // Determine element index
      PDBElement *element = atom->Element();
      if (!element) continue;
      int element_index = element - PDBelements;
      assert((element_index >= 0) && (element_index < PDBnelements));
      if (element_index >= nelements) element_index = 0;

      // Determine value for splat
      RNScalar value = 1;
      if (normalize_by_ligand) {
        value /= ligand->NAtoms();
      }

      // Rasterize atom into map
      map->AddPoint(frame_index, element_index, atom_position, value);
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("  Added %s to map ...\n", file->Name());
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Frames = %d\n", 1);
    printf("    # Atoms = %d\n", atom_count);
    fflush(stdout);
  }

  // Return number of atoms
  return atom_count;
}



int
AddAtomsToMap(Map *map, PDBFile *file, PDBResidue *ligand)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int atom1_count = 0;
  int atom2_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Create KD tree of all protein atoms
  PDBAtom *tmp = model->Atom(0);
  int position_offset = (unsigned char *) &(tmp->position) - (unsigned char *) tmp;
  KDTree<PDBAtom *> *kdtree = new KDTree<PDBAtom *>(model->atoms, position_offset);
  if (!kdtree) { fprintf(stderr, "Unable to create KD tree for %s\n", file->Name()); return 0; }

  // Consider each protein atom
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (!atom1->AminoAcid()) continue;
    if ((asa_names) && (atom1->accessible_surface_area == 0)) continue;
    if (ligand && R3Distance(atom1->Position(), ligand->BBox()) > world_radius) continue;
    if (bonded_protein_atoms_only && ligand && (PDBDistance(atom1, ligand) > max_bond_length)) continue;
    int atom1_index = map->AtomType(atom1);
    assert(atom1_index < map->NAtomTypes());
    if (atom1_index < 0) continue;
    atom1_count++;

    // Find all nearby atoms 
    const R3Point& p1 = atom1->Position();
    RNArray<PDBAtom *> nearby_atoms;
    kdtree->FindAll(atom1, RN_EPSILON, 2 * nearby_radius, nearby_atoms);
    if (nearby_atoms.IsEmpty()) continue;

    // Compute first vector of atom coordinate system 
    R3Vector vector1 = R3zero_vector;
    for (int a = 0; a < nearby_atoms.NEntries(); a++) {
      PDBAtom *atom2 = nearby_atoms[a];
      if (atom2->IsHetAtom()) continue;
      if (!atom2->AminoAcid()) continue;
      R3Point p2 = atom2->Position();
      R3Vector v = p2 - p1;
      RNLength d = v.Length();
      if (RNIsZero(d)) continue;
      vector1 += v / d;
    }

    // Compute second vector of atom coordinate system 
    if (vector1.IsZero()) continue;
    R3Vector vector2 = R3zero_vector;
    R3Plane plane1(atom1->Position(), vector1);
    for (int a = 0; a < nearby_atoms.NEntries(); a++) {
      PDBAtom *atom2 = nearby_atoms[a];
      if (atom2->IsHetAtom()) continue;
      if (!atom2->AminoAcid()) continue;
      R3Point p2 = atom2->Position();
      p2.Project(plane1);
      R3Vector v = p2 - p1;
      RNLength d = v.Length();
      if (RNIsZero(d)) continue;
      vector2 += v / d;
    }

    // Construct coordinate system
    if (vector2.IsZero()) continue;
    R3Triad triad(vector1, vector2);
    R3CoordSystem cs(p1, triad);

    // Construct transformation from world to frame
    R3Affine world_to_frame(cs.InverseMatrix());

    // Rasterize atoms of residue2 into map
    for (int j = 0; j < model->NResidues(); j++) {
      PDBResidue *residue2 = model->Residue(j);
      if (ligand && (residue2 != ligand)) continue;
      for (int k = 0; k < residue2->NAtoms(); k++) {
        PDBAtom *atom2 = residue2->Atom(k);
        if (!atom2->IsHetAtom()) continue;
        RNLength distance = PDBDistance(atom1, atom2);
        if (distance > world_radius) continue;
        if (bonded_ligand_atoms_only && (distance > max_bond_length)) continue;
        atom2_count++;

        // Transform atom position from world coordinates to frame coordinates
        R3Point atom2_position = atom2->Position();
        atom2_position.Transform(world_to_frame);

        // Randomize atom position?
        if (randomize_atom_positions) {
          RNCoord x = 2 * world_radius * (RNRandomScalar() - 0.5);
          RNCoord y = 2 * world_radius * (RNRandomScalar() - 0.5);
          RNCoord z = 2 * world_radius * (RNRandomScalar() - 0.5);
          atom2_position.Reset(x, y, z);
        }

        // Determine element 
        PDBElement *element = atom2->Element();
        if (!element) continue;
        int element_index = element - PDBelements;
        assert((element_index >= 0) && (element_index < PDBnelements));
        if (element_index >= nelements) element_index = 0;

        // Determine value for splat
        RNScalar value = 1;
        if (normalize_by_ligand) {
          value /= residue2->NAtoms();
        }

        // Rasterize atom into map
        map->AddPoint(atom1_index, element_index, atom2_position, value);
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("  Added %s to map ...\n", file->Name());
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Frames = %d\n", atom1_count);
    printf("    # Atoms = %d\n", atom2_count);
    fflush(stdout);
  }

  // Return number of atoms
  return atom2_count;
}



int
AddTriplesToMap(Map *map, PDBFile *file, PDBResidue *ligand)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int atom1_count = 0;
  int atom2_count = 0;
  int atom3_count = 0;
  int atom4_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);
  if (model->NAtoms() == 0) return 0;

  // Get the number of possible atom types
  int natomtypes = map->NAtomTypes();

  // Create KD tree of all protein atoms
  PDBAtom *tmp = model->Atom(0);
  int position_offset = (unsigned char *) &(tmp->position) - (unsigned char *) tmp;
  KDTree<PDBAtom *> *kdtree = new KDTree<PDBAtom *>(model->atoms, position_offset);
  if (!kdtree) { fprintf(stderr, "Unable to create KD tree for %s\n", file->Name()); return 0; }

  // Consider every atom as potential origin of coordinate system
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (!atom1->AminoAcid()) continue;
    if ((asa_names) && (atom1->accessible_surface_area == 0)) continue;
    if (ligand && R3Distance(atom1->Position(), ligand->BBox()) > world_radius) continue;
    if (bonded_protein_atoms_only && ligand && (PDBDistance(atom1, ligand) > max_bond_length)) continue;
    int atom1_index = map->AtomType(atom1);
    assert(atom1_index < natomtypes);
    if (atom1_index < 0) continue;
    atom1_count++;

    // Find all other atoms nearby atom1 
    RNArray<PDBAtom *> nearby_atoms;
    kdtree->FindAll(atom1, RN_EPSILON, 2 * nearby_radius, nearby_atoms);

    // Consider all nearby protein atoms
    for (int j = 0; j < nearby_atoms.NEntries(); j++) {
      PDBAtom *atom2 = nearby_atoms[j];
      if (atom2 == atom1) continue;
      if (atom2->IsHetAtom()) continue;
      if (!atom2->AminoAcid()) continue;
      if ((asa_names) && (atom2->accessible_surface_area == 0)) continue;
      if (R3Distance(atom2->Position(), atom1->Position()) > nearby_radius) continue;
      int atom2_index = map->AtomType(atom2);
      assert(atom2_index < natomtypes);
      if (atom2_index < 0) continue;
      atom2_count++;

      // Consider all pairs of nearby protein atoms
      for (int k = j+1; k < nearby_atoms.NEntries(); k++) {
        PDBAtom *atom3 = nearby_atoms[k];
        if (atom3 == atom1) continue;
        if (atom3 == atom2) continue;
        if (atom3->IsHetAtom()) continue;
        if (!atom3->AminoAcid()) continue;
        if ((asa_names) && (atom3->accessible_surface_area == 0)) continue;
        if (R3Distance(atom3->Position(), atom2->Position()) > nearby_radius) continue;
        int atom3_index = map->AtomType(atom3);
        assert(atom3_index < natomtypes);
        if (atom3_index < 0) continue;
        atom3_count++;

        // Construct world to frame transformation
        R3Point origin = atom1->Position();
        R3Vector posx_vector = origin - atom2->Position();
        RNLength posx_length = posx_vector.Length();
        if (RNIsZero(posx_length)) continue;
        else posx_vector /= posx_length;
        R3Vector posy_vector = atom3->Position() - atom2->Position();
        R3Vector posz_vector = posx_vector % posy_vector;
        RNLength posz_length = posz_vector.Length();
        if (RNIsZero(posz_length)) return 0;
        else posz_vector /= posz_length;
        posy_vector = posz_vector % posx_vector;
        RNLength posy_length = posy_vector.Length();
        if (RNIsZero(posy_length)) return 0;
        else posy_vector /= posy_length;
        R3Triad triad(posx_vector, posy_vector, posz_vector);
        R3CoordSystem cs(origin, triad);
        R3Affine world_to_frame(cs.InverseMatrix());

        // Determine frame index 
        int frame_index = atom3_index;
        frame_index += atom2_index * natomtypes;
        frame_index += atom1_index * natomtypes * natomtypes;
        assert((frame_index >= 0) && (frame_index < map->NFrames()));

        // Splat residues of model into map
        for (int r = 0; r < model->NResidues(); r++) {
          PDBResidue *residue = model->Residue(r);
          if (ligand && (residue != ligand)) continue;
          if (R3Distance(residue->BBox(), origin) > world_radius) continue;
          if (bonded_protein_atoms_only && (PDBDistance(atom1, residue) > max_bond_length)) continue;

          // Splat atoms of residue into map
          for (int m = 0; m < residue->NAtoms(); m++) {
            PDBAtom *atom4 = residue->Atom(m);
            if (R3Distance(atom4->Position(), origin) > world_radius) continue;
            if (bonded_ligand_atoms_only && (PDBDistance(atom1, atom4) > max_bond_length)) continue;
            atom4_count++;

            // Transform atom position from world coordinates to frame coordinates
            R3Point atom4_position = atom4->Position();
            atom4_position.Transform(world_to_frame);

            // Randomize atom position?
            if (randomize_atom_positions) {
              RNCoord x = 2 * world_radius * (RNRandomScalar() - 0.5);
              RNCoord y = 2 * world_radius * (RNRandomScalar() - 0.5);
              RNCoord z = 2 * world_radius * (RNRandomScalar() - 0.5);
              atom4_position.Reset(x, y, z);
            }

            // Determine element type for atom4
            PDBElement *element4 = atom4->Element();
            int element4_index = 0;
            if (element4) element4_index = element4->ID();
            if (element4_index >= nelements) element4_index = 0;

            // Determine value for splat
            RNScalar value = 1;
            if (normalize_by_ligand) {
              value /= residue->NAtoms();
            }

            // Rasterize atom into map
            map->AddPoint(frame_index, element4_index, atom4_position, value);
          }
        }
      }
    }
  }

  // Delete the KD tree
  delete kdtree;

  // Print statistics
  if (print_verbose) {
    printf("  Added %s to map ...\n", file->Name());
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Origins = %d\n", atom1_count);
    printf("    # Pairs = %d\n", atom2_count);
    printf("    # Frames = %d\n", atom3_count);
    printf("    # Splats = %d\n", atom4_count);
    fflush(stdout);
  }

  // Return number of atoms
  return atom4_count;
}



int
AddResiduesToMap(Map *map, PDBFile *file, PDBResidue *ligand)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  int residue_count = 0;
  int atom_count = 0;

  // Get model from PDB file
  if (file->NModels() == 0) return 0;
  PDBModel *model = file->Model(0);

  // Consider each protein residue
  for (int i = 0; i < model->NResidues(); i++) {
    PDBResidue *residue1 = model->Residue(i);
    if (residue1->HasHetAtoms()) continue;
    if (bonded_protein_atoms_only && ligand && (PDBDistance(residue1, ligand) > max_bond_length)) continue;

    // Determine amino acid
    PDBAminoAcid *aminoacid1 = residue1->AminoAcid();
    if (!aminoacid1) continue;
    int aminoacid_index = aminoacid1 - PDBaminoacids;
    assert((aminoacid_index >= 0) && (aminoacid_index < PDBnaminoacids));
    assert(aminoacid_index < map->NFrames());

    // Check whether the residue is on the surface
    if (asa_names) {
      RNBoolean on_surface = FALSE;
      for (int j = 0; j < residue1->NAtoms(); j++) 
        if (residue1->Atom(j)->accessible_surface_area > 0) { on_surface = TRUE; break; }
      if (!on_surface) continue;
    }

    // Construct residue coordinate system 
    R3CoordSystem cs;
    if (!residue1->ConstructCoordinateSystem(cs, atom_type)) {
      // fprintf(stderr, "Unable to construct coordinate system for residue %s %d\n", residue1->Name(), residue1->Sequence());
      continue;
    }

    // Construct transformation from world to frame
    R3Affine world_to_frame(cs.InverseMatrix());

    // Splat residues of model into map
    for (int j = 0; j < model->NResidues(); j++) {
      PDBResidue *residue2 = model->Residue(j);
      if (ligand && (residue2 != ligand)) continue;
      if (R3Distance(residue1->BBox(), residue2->BBox()) > world_radius) continue;
      if (bonded_protein_atoms_only && (PDBDistance(residue1, residue2) > max_bond_length)) continue;

      // Rasterize atoms of residue2 into map
      for (int k = 0; k < residue2->NAtoms(); k++) {
        PDBAtom *atom = residue2->Atom(k);
        if (!atom->IsHetAtom()) continue;
        if (R3Distance(atom->Position(), cs.Origin()) > world_radius) continue;
        if (bonded_ligand_atoms_only && (PDBDistance(residue1, atom) > max_bond_length)) continue;

        // Transform atom position from world coordinates to frame coordinates
        R3Point atom_position = atom->Position();
        atom_position.Transform(world_to_frame);

        // Randomize atom position?
        if (randomize_atom_positions) {
          RNCoord x = 2 * world_radius * (RNRandomScalar() - 0.5);
          RNCoord y = 2 * world_radius * (RNRandomScalar() - 0.5);
          RNCoord z = 2 * world_radius * (RNRandomScalar() - 0.5);
          atom_position.Reset(x, y, z);
        }

        // Determine element 
        PDBElement *element = atom->Element();
        if (!element) continue;
        int element_index = element - PDBelements;
        assert((element_index >= 0) && (element_index < PDBnelements));
        if (element_index >= nelements) element_index = 0;

        // Determine value for splat
        RNScalar value = 1;
        if (normalize_by_ligand) {
          value /= residue2->NAtoms();
        }

        // Rasterize atom into map
        map->AddPoint(aminoacid_index, element_index, atom_position, value);

        // Update statistics
        atom_count++;
      }
    }

    // Update statistics
    residue_count++;
  }

  // Print statistics
  if (print_verbose) {
    printf("  Added %s to map ...\n", file->Name());
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Residues = %d\n", residue_count);
    printf("    # Atoms = %d\n", atom_count);
    fflush(stdout);
  }

  // Return number of atoms
  return atom_count;}



static Map *
AllocateMap(int frame_type, int atom_type, int storage_type, int nelements, RNLength world_radius, int resolution)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate map
  Map *map = new Map(frame_type, atom_type, storage_type, nelements, world_radius, resolution);
  if (!map) {
    fprintf(stderr, "Unable to allocate map.\n");
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Allocated map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Frame Type = %d\n", frame_type);
    printf("  Atom Type = %d\n", atom_type);
    printf("  Storage Type = %d\n", storage_type);
    printf("  # Frames = %d\n", map->NFrames());
    printf("  # Elements = %d\n", map->NElements());
    printf("  World Radius = %g\n", map->WorldRadius());
    printf("  World Spacing = %g\n", map->WorldSpacing());
    printf("  Resolution = %d\n", map->Resolution());
    fflush(stdout);
  }

  // Return map
  return map;
}



int
CreateMap(Map *map, char *pdb_names, char *ligand_names, char *asa_names)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Seed random number generator
  RNSeedRandomScalar();

  // Open the pdb file list
  FILE *pdb_list = fopen(pdb_names, "r");
  if (!pdb_list) {
    fprintf(stderr, "Unable to open list of pdb files: %s\n", pdb_names);
    return 0;
  }

  // Open the ligand list
  FILE *ligand_list = NULL;
  if (ligand_names) {
    ligand_list = fopen(ligand_names, "r");
    if (!ligand_list) {
      fprintf(stderr, "Unable to open list of ligands: %s\n", ligand_names);
      return 0;
    }
  }

  // Open the asa list
  FILE *asa_list = NULL;
  if (asa_names) {
    asa_list = fopen(asa_names, "r");
    if (!asa_list) {
      fprintf(stderr, "Unable to open list of asa files: %s\n", asa_names);
      return 0;
    }
  }

  // Read pdb files and add them to the map
  int file_count = 0;
  int atom_count = 0;
  char pdb_name[256];
  while (fscanf(pdb_list, "%s", pdb_name) == 1) {
    // Increment file counter
    file_count++;

    // Print message
    if (print_verbose) {
      printf("Processing %d ...\n", file_count);
    }

    // Read PDB file
    PDBFile *file = ReadPDB(pdb_name);
    if (!file) return 0;

    // Find ligand
    PDBResidue *ligand = NULL;
    if (ligand_names) {
      char ligand_name[128];
      if (fscanf(ligand_list, "%s", ligand_name) == 1) {
        ligand = FindLigand(file, ligand_name);
        if (!ligand) return 0;
      }
      else {
        fprintf(stderr, "Not enough entries in ligand list: %s\n", ligand_names);
        return 0;
      }
    }

    // Read accessible surface area file
    if (asa_names) {
      char asa_name[128];
      if (fscanf(asa_list, "%s", asa_name) == 1) {
        int status = ReadASAFile(file, asa_name);
        if (!status) return 0;
      }
      else {
        fprintf(stderr, "Not enough entries in asa file list: %s\n", asa_names);
        return 0;
      }
    }

    // Add PDB file to map
    if (frame_type == ATOM_MAP_FRAME) atom_count += AddAtomsToMap(map, file, ligand);
    else if (frame_type == TRIPLE_MAP_FRAME) atom_count += AddTriplesToMap(map, file, ligand);
    else if (frame_type == RESIDUE_MAP_FRAME) atom_count += AddResiduesToMap(map, file, ligand);
    else if (frame_type == SITE_MAP_FRAME) atom_count += AddSiteToMap(map, file, ligand);
    else RNAbort("Invalid frame type: %d", frame_type);

    // Delete the PDB file
    delete file;
  }

  // Close the list file
  fclose(pdb_list);

  // Print statistics
  if (print_verbose) {
    printf("Created map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Models = %d\n", file_count);
    printf("  # Atoms = %d\n", atom_count);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
BlurMap(Map *map, RNLength world_sigma)
{
  // Check if blurring map
  if (world_sigma == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Blur map
  map->Blur(world_sigma);

  // Print statistics
  if (print_verbose) {
    printf("Blurred map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  World Sigma = %g\n", world_sigma);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
BlendMap(Map *map)
{
  // Check if blending map
  if (!blend) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Apply gaussian falloff to map
  map->Blend();

  // Print statistics
  if (print_verbose) {
    printf("Blended map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
NormalizeMap(Map *map, int normalization_type)
{
  // Check if normalizing
  if (normalization_type == 0) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Check normalization type
  if (normalization_type == 1) {
    // Divide each distribution by L1Norm per element
    for (int j = 0; j < map->NElements(); j++) {
      // Compute sum 
      RNScalar sum = 0;
      for (int i = 0; i < map->NFrames(); i++) {
        sum += map->L1Norm(i, j);
      }

      // Divide by sum
      for (int i = 0; i < map->NFrames(); i++) {
        if (RNIsZero(sum)) continue;
        map->Scale(i, j, map->NFrames() / sum);
      }
    }
  }
  else if (normalization_type == 2) {
    // Divide each distribution by L1Norm per frame
    for (int i = 0; i < map->NFrames(); i++) {
      // Compute sum 
      RNScalar sum = 0;
      for (int j = 0; j < map->NElements(); j++) {
        sum += map->L1Norm(i, j);
      }

      // Divide by sum
      for (int j = 0; j < map->NElements(); j++) {
        if (RNIsZero(sum)) continue;
        map->Scale(i, j, map->NElements() / sum);
      }
    }
  }
  else if (normalization_type == 3) {
    // Allocate sums
    RNScalar *frame_sums = new RNScalar [ map->NFrames() ];
    RNScalar *element_sums = new RNScalar [ map->NElements() ];
    assert(frame_sums && element_sums);

    // Compute total sum
    RNScalar total_sum = 0;
    for (int i = 0; i < map->NFrames(); i++) {
      for (int j = 0; j < map->NElements(); j++) {
        total_sum += map->L1Norm(i, j);
      }
    }

    // Compute element sums
    for (int j = 0; j < map->NElements(); j++) {
      // Compute sum 
      element_sums[j] = 0;
      for (int i = 0; i < map->NFrames(); i++) {
        element_sums[j] += map->L1Norm(i, j);
      }
    }

    // Compute frame sums
    for (int i = 0; i < map->NFrames(); i++) {
      // Compute sum 
      frame_sums[i] = 0;
      for (int j = 0; j < map->NElements(); j++) {
        frame_sums[i] += map->L1Norm(i, j);
      }
    }

    // Divide by both element and frame sums
    for (int i = 0; i < map->NFrames(); i++) {
      for (int j = 0; j < map->NElements(); j++) {
        if (RNIsZero(frame_sums[i])) continue;
        if (RNIsZero(element_sums[j])) continue;
        map->Scale(i, j, total_sum / (frame_sums[i] * element_sums[j]));
      }
    }

    // Delete sums
    delete [] frame_sums;
    delete [] element_sums;
  }
  else if (normalization_type == 4) {
    // Divide each distribution by L1Norm per distribution
    for (int i = 0; i < map->NFrames(); i++) {
      for (int j = 0; j < map->NElements(); j++) {
        RNScalar sum = map->L1Norm(i, j);
        if (RNIsZero(sum)) continue;
        map->Scale(i, j, 1 / sum);
      }
    }
  }
  else if (normalization_type == 5) {
    // Compute sum 
    RNScalar sum = 0;
    for (int i = 0; i < map->NFrames(); i++) {
      for (int j = 0; j < map->NElements(); j++) {
        sum += map->L1Norm(i, j);
      }
    }

    // Divide by sum
    for (int i = 0; i < map->NFrames(); i++) {
      for (int j = 0; j < map->NElements(); j++) {
        if (RNIsZero(sum)) continue;
        map->Scale(i, j, map->NElements() *  map->NFrames() / sum);
      }
    }
  }
  else {
    RNAbort("Invalid normalization type");
  }

  // Print statistics
  if (print_verbose) {
    printf("Normalized map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
WriteMap(Map *map, char *map_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write map
  int status = map->WriteFile(map_name);
  if (!status) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Wrote map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdb2map listfile mapdir [-resolution x y z] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-ligands")) { 
        argc--; argv++; ligand_names = *argv; 
      }
      else if (!strcmp(*argv, "-asas")) { 
        argc--; argv++; asa_names = *argv; 
      }
      else if (!strcmp(*argv, "-frame_type")) { 
        argc--; argv++; frame_type = atoi(*argv);
      }
      else if (!strcmp(*argv, "-atom_type")) { 
        argc--; argv++; atom_type = atoi(*argv);
      }
      else if (!strcmp(*argv, "-storage_type")) { 
        argc--; argv++; storage_type = atoi(*argv);
      }
      else if (!strcmp(*argv, "-nelements")) { 
        argc--; argv++; nelements = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; resolution = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-bonded_ligand_atoms")) { 
        bonded_ligand_atoms_only = TRUE; 
      }
      else if (!strcmp(*argv, "-bonded_protein_atoms")) { 
        bonded_protein_atoms_only = TRUE; 
      }
      else if (!strcmp(*argv, "-nearby")) { 
        argc--; argv++; nearby_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-sigma")) { 
        argc--; argv++; world_sigma = atof(*argv); 
      }
      else if (!strcmp(*argv, "-blend")) { 
        blend = 1; 
      }
      else if (!strcmp(*argv, "-normalize")) { 
        argc--; argv++; normalization_type = atoi(*argv); 
        if (normalization_type & 0x8) {
          normalize_by_ligand = TRUE;
        }
      }
      else if (!strcmp(*argv, "-random")) { 
        randomize_atom_positions = TRUE; 
      }
      else if (!strcmp(*argv, "-debug")) { 
        print_debug = 1; 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_names) pdb_names = *argv;
      else if (!map_name) map_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check list filename
  if (!pdb_names) {
    fprintf(stderr, "You did not specify a list of pdb files.\n");
    return 0;
  }

  // Check map filename
  if (!map_name) {
    fprintf(stderr, "You did not specify a map filename.\n");
    return 0;
  }

  // Check frame type
  if ((frame_type < 0) || (frame_type >= NUM_MAP_FRAME_TYPES)) {
    fprintf(stderr, "Invalid map frame type: %d (%d)\n", frame_type, NUM_MAP_FRAME_TYPES);
    return 0;
  }

  // Check storage type
  if ((storage_type < 0) || (storage_type >= NUM_MAP_STORAGE_TYPES)) {
    fprintf(stderr, "Invalid map storage type: %d\n", storage_type);
    return 0;
  }

  // Check number of ligand elements
  if ((nelements <= 0) || (nelements > PDBnelements)) {
    fprintf(stderr, "Invalid number of elements: %d\n", nelements);
    return 0;
  }

  // Check resolution
  if (resolution <= 0) {
    fprintf(stderr, "Invalid map resolution: %d\n", resolution);
    return 0;
  }

  // Check world radius
  if (RNIsZero(world_radius)) {
    fprintf(stderr, "Invalid world radius: %g\n", world_radius);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Allocate map 
  Map *map = AllocateMap(frame_type, atom_type, storage_type, nelements, world_radius, resolution);
  if (!map) exit(-1);

  // Create map 
  int status1 = CreateMap(map, pdb_names, ligand_names, asa_names);
  if (!status1) exit(-1);

  // Blur map 
  int status2 = BlurMap(map, world_sigma);
  if (!status2) exit(-1);

  // Blend map 
  int status3 = BlendMap(map);
  if (!status3) exit(-1);

  // Normalize map 
  int status4 = NormalizeMap(map, normalization_type);
  if (!status4) exit(-1);

  // Write map
  int status5 = WriteMap(map, map_name);
  if (!status5) exit(-1);

  // Print more statistics
  if (print_debug) {
    printf("Debug Information ...\n");
    map->PrintStatistics();
    fflush(stdout);
  }

  // Return success
  return 0;
}
