// Source file for the pdb viewer program



// Include files 

#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"



// Program variables

static char *pdb_name = NULL;
static char *shd_name = NULL;
static char *ligand_name = NULL;
static char *grow_name = NULL;
static char *asa_name = NULL;
static int site_type = 0; // 0=protein, 1=ligand
static int descriptor_type = 0; // 0=shd, 1=sfd, 2=sgd
static int rasterization_type = 0;  // 0=point, 1=solid, 2=surface, 3=charge
static int grid_resolution[3] = { 64, 64, 64 };
static R3Point *world_center = NULL;
static RNScalar world_radius = 0;
static RNScalar max_distance = 0;
static RNScalar sigma = sqrt(8.0);
static RNBoolean elements = FALSE;
static RNScalar element_weights[5] = { 1, 1, 1, 1, 1 }; // C, N, O, P, S
static int normalization_type = 0;
static int print_verbose = 0;
static int debug = 0;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static PDBResidue *
FindLigand(PDBFile *file, const char *ligand_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Find ligand
  PDBResidue *ligand = file->FindResidue(ligand_name);
  if (!ligand) {
    fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, pdb_name); 
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Found ligand ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", ligand->NAtoms());
    fflush(stdout);
  }

  // Return ligand
  return ligand;
}



static int
ReadASAFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read ASA file
  int natoms = file->ReadASAFile(filename);
  if (natoms == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read accessible surface area file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Surface Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadGrowFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read grow file
  int nbonds = file->ReadGrowFile(filename);
  if (nbonds == 0) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read grow file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bonds = %d\n", nbonds);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
RasterizeAtoms(R3Grid *grid, const RNArray<PDBAtom *>& atoms, int rasterization_type, RNBoolean grow, RNBoolean asa, RNBoolean hetatoms, PDBElement *element)
{
  // Rasterize each atom according to the program arguments
  int atom_count = 0;
  for (int i = 0; i < atoms.NEntries(); i++) {
    PDBAtom *atom = atoms.Kth(i);
    if (grow && (atom->NBonds() == 0)) continue;
    if (asa && (atom->accessible_surface_area == 0)) continue;
    if (!hetatoms && atom->IsHetAtom()) continue;
    if (element && (element != atom->Element())) continue;
    if (rasterization_type == 0) grid->RasterizeWorldPoint(atom->Position(), 1.0);
    else if (rasterization_type == 3) grid->RasterizeWorldPoint(atom->Position(), atom->Charge());
    else grid->RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
    atom_count++;
  }

  // Blur grid, depending on rasterization type
  if (rasterization_type == 3) {
    grid->Blur(sigma * grid->WorldToGridScaleFactor());
  }

  // Return number of atoms
  return atom_count;
}



static R3Grid *
CreateGrid(PDBFile *file, PDBResidue *ligand, PDBElement *element)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get model from PDB file
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Compute ligand center
  R3Point world_centroid;
  if (world_center) world_centroid = *world_center;
  else if (ligand) world_centroid = ligand->Centroid();
  else world_centroid = model->Centroid();

  // Compute grid center
  int res = grid_resolution[0];
  if (res < grid_resolution[1]) res = grid_resolution[1];
  if (res < grid_resolution[2]) res = grid_resolution[2];
  R3Vector grid_diagonal(grid_resolution[0]-1, grid_resolution[1]-1, grid_resolution[2]-1);
  R3Vector grid_centroid = 0.5 * grid_diagonal;

  // Compute world radius
  if (world_radius == 0) {
    if (ligand) world_radius = 3.0 * PDBAverageDistance(ligand->atoms, world_centroid);
    else world_radius = 3.0 * PDBAverageDistance(model->atoms, world_centroid);
  }

  // Set affine transformation
  R3Affine world_to_grid = R3identity_affine;
  world_to_grid.Translate(grid_centroid);
  world_to_grid.Scale((0.5 * (res-1)) / world_radius);
  world_to_grid.Translate(-(world_centroid.Vector()));
  grid->SetWorldToGridTransformation(world_to_grid);

  // Rasterize atoms into grid
  int natoms = 0;
  RNBoolean grow = (grow_name) ? TRUE : FALSE;
  RNBoolean asa = (asa_name) ? TRUE : FALSE;
  if ((site_type == 1) && (ligand)) natoms = RasterizeAtoms(grid, ligand->atoms, rasterization_type, grow, asa, TRUE, element);
  else natoms = RasterizeAtoms(grid, model->atoms, rasterization_type, grow, asa, FALSE, element);

  // Compute surface, if needed
  if (rasterization_type == 2) {
    grid->Threshold(0.5, 0, 1);
    grid->DetectEdges(); 
    grid->Threshold(0.5, 0, 1);
  }

  // Mask to within max_distance of other structure (ligand or protein), if needed
  if (ligand && (max_distance > 0)) {
    R3Grid dgrid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    dgrid.SetWorldToGridTransformation(world_to_grid);
    if (site_type == 0) RasterizeAtoms(&dgrid, ligand->atoms, 1, FALSE, FALSE, TRUE, NULL);
    else RasterizeAtoms(&dgrid, model->atoms, 1, FALSE, FALSE, FALSE, NULL);
    dgrid.SquaredDistanceTransform();
    dgrid.Multiply(grid->GridToWorldScaleFactor() * grid->GridToWorldScaleFactor());
    dgrid.Threshold(max_distance * max_distance, 1, 0);
    grid->Mask(dgrid);
  }

  // Print statistics
  if (print_verbose) {
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Grid Resolution = %d %d% d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Grid Center = %g %g %g\n", grid_centroid[0], grid_centroid[1], grid_centroid[2]);
    printf("  World Center = %g %g %g\n", world_centroid[0], world_centroid[1], world_centroid[2]);
    printf("  World Radius = %g\n", world_radius);
    printf("  World Spacing = %g\n", grid->GridToWorldScaleFactor());
    RNInterval range = grid->Range();
    printf("  Minimum = %g\n", range.Min());
    printf("  Maximum = %g\n", range.Max());
    printf("  L2Norm = %g\n", grid->L2Norm());
    printf("  # Atoms = %d\n", natoms);
    fflush(stdout);
  }

  // Output debug information
  if (debug) {
    char grid_name[64];
    static int grid_count = 0;
    sprintf(grid_name, "debug%d.grd", ++grid_count);
    grid->WriteFile(grid_name);
  }

  // Return grid
  return grid;
}



static R3SphericalGrid *
CreateSphericalGrid(R3Grid *grid1)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical grid
  R3SphericalGrid *grid = new R3SphericalGrid(grid1->XResolution()/2, grid1->XResolution()/2, grid1->XResolution()/2);
  if (!grid) {
    fprintf(stderr, "Unable to allocate spherical grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(grid1->WorldBox());

  // Check sigma
  if ((rasterization_type < 3) && RNIsPositive(sigma)) {
    // Fill samples on spherical grid with Gaussian of DT
    const RNScalar sqrt_two_pi = sqrt(RN_TWO_PI);
    RNScalar a = sqrt_two_pi * sigma;
    RNScalar fac = 1.0 / (a * a * a);
    RNScalar denom = -2.0 * sigma * sigma;
    grid1->SquaredDistanceTransform();
    for (int k = 0; k < grid->RadiusResolution(); k++) {
      for (int j = 0; j < grid->ThetaResolution(); j++) {
        for (int i = 0; i < grid->PhiResolution(); i++) {
          R3Point world_position = grid->WorldPosition(i, j, k);
          RNScalar value1 = grid1->WorldValue(world_position);
          RNScalar value = (value1 > 0) ? fac * exp( value1 / denom ) : 0.0;
          grid->SetGridValue(i, j, k, value);
        }
      }
    }
  }
  else {
    // Handle case of sigma=0 separarately
    for (int k = 0; k < grid->RadiusResolution(); k++) {
      for (int j = 0; j < grid->ThetaResolution(); j++) {
        for (int i = 0; i < grid->PhiResolution(); i++) {
          R3Point world_position = grid->WorldPosition(i, j, k);
          RNScalar value = grid1->WorldValue(world_position);
          grid->SetGridValue(i, j, k, value);
        }
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created spherical grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->PhiResolution(), grid->ThetaResolution(), grid->RadiusResolution());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return spherical grid
  return grid;
}



static R3SphericalHarmonics *
CreateSphericalHarmonics(R3SphericalGrid *grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical harmonics
  R3SphericalHarmonics *harmonics = new R3SphericalHarmonics(grid->ThetaResolution()/2, grid->RadiusResolution());
  if (!harmonics) {
    fprintf(stderr, "Unable to allocate spherical harmonics\n");
    exit(-1);
  }

  // Compute forward harmonic transform
  grid->ForwardSphericalHarmonicTransform(*harmonics);

  // Print statistics
  if (print_verbose) {
    printf("Created spherical harmonics ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Coefficients = %d\n", harmonics->NCoefficients());
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    printf("  L2Norm = %g\n", harmonics->L2Norm());
    fflush(stdout);
  }

  // Return spherical harmonics
  return harmonics;
}



static int
CreateDescriptor(R3SphericalHarmonics *harmonics, int descriptor_type, float *values)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  float sum = 0.0;

  // Initialize counts
  int nvalues = 0;
  int count = 0;

  // Check descriptor type
  if (descriptor_type == 0) {
    // Compute number of values
    nvalues = harmonics->RadiusResolution() * harmonics->Bandwidth();

    // Compute L2Norm for each radius and frequency
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        float value_squared = harmonics->L2NormSquared(S, L);
        float value = sqrt(value_squared);
        assert(count <= nvalues);
        values[count++] = value;
        sum += value_squared;
      }
    }
  }
  else if (descriptor_type == 1) {
    // Compute number of values
    nvalues = harmonics->NCoefficients() / 2;

    // Compute L2Norm for each coefficient
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        for (int M = -L; M <= L; M++) {
          float value_squared = harmonics->L2NormSquared(S, L, M);
          float value = sqrt(value_squared);
          assert(count <= nvalues);
          values[count++] = value;
          sum += value_squared;
        }
      }
    }
  }
  else if (descriptor_type == 2) {
    // Compute number of values
    nvalues = harmonics->NCoefficients();

    // Compute coefficients for each radius and frequency
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        for (int M = -L; M <= L; M++) {
          for (int k = 0; k < 2; k++) {
            float value = harmonics->Shell(S).Coefficient(L, M, k);
            float value_squared = value * value;
            assert(count <= nvalues);
            values[count++] = value;
            sum += value_squared;
          }
        }
      }
    }
  }
  else {
    RNAbort("Invalid descriptor type");
  }

  // Compute l2norm
  assert(nvalues > 0);
  assert(count == nvalues);
  RNScalar l2norm = sqrt(sum);

  // Normalize descriptor
  if ((normalization_type % 2) == 1) {
    if (l2norm > 0) {
      for (int i = 0; i < nvalues; i++) {
        values[i] /= l2norm;
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Created spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", nvalues);
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    printf("  L2Norm = %g\n", l2norm);
    fflush(stdout);
  }

  // Return number of values
  return nvalues;
}



static int
NormalizeDescriptor(float *values, int nvalues, int normalization_type)
{
  // Check if normalizing
  if (normalization_type < 2) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute l2norm
  float l2norm = 0;
  for (int i = 0; i < nvalues; i++)
    l2norm += values[i] * values[i];
  l2norm = sqrt(l2norm);

  // Normalize descriptor
  if (l2norm > 0) {
    for (int i = 0; i < nvalues; i++)
      values[i] /= l2norm;
  }

  // Print statistics
  if (print_verbose) {
    printf("Normalized spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    printf("  L2Norm = %g\n", l2norm);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WriteDescriptor(float *values, int nvalues, const char *shd_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(shd_name, "wb");
  if (!fp) {
    fprintf(stderr, "Unable to open SHD file %s\n", shd_name);
    return 0;
  }

  // Write number of values
  if (fwrite(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to write number of values to SHD file %s\n", shd_name);
    return 0;
  }

  // Write values
  if (fwrite(values, sizeof(float), nvalues, fp) != (unsigned long) nvalues) {
    fprintf(stderr, "Unable to write values to SHD file %s\n", shd_name);
    return 0;
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", sizeof(int) + nvalues * sizeof(float));
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdb2grd pdbfile gridfile [-resolution x y z] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      else if (!strcmp(*argv, "-ligand")) { 
        argc--; argv++; ligand_name = *argv; 
      }
      else if (!strcmp(*argv, "-asa")) { 
        argc--; argv++; asa_name = *argv; 
      }
      else if (!strcmp(*argv, "-grow")) { 
        argc--; argv++; grow_name = *argv; 
      }
      else if (!strcmp(*argv, "-site")) { 
        argc--; argv++; site_type = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-rasterization")) { 
        argc--; argv++; rasterization_type = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-radius")) { 
        argc--; argv++; world_radius = atof(*argv); 
      }
      else if (!strcmp(*argv, "-center")) { 
        world_center = new R3Point();
        argc--; argv++; (*world_center)[0] = atof(*argv); 
        argc--; argv++; (*world_center)[1] = atof(*argv); 
        argc--; argv++; (*world_center)[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-max_distance")) { 
        argc--; argv++; max_distance = atof(*argv); 
      }
      else if (!strcmp(*argv, "-sigma")) { 
        argc--; argv++; sigma = atof(*argv); 
      }
      else if (!strcmp(*argv, "-elements")) {
        elements = TRUE; 
      }
      else if (!strcmp(*argv, "-weights")) { 
        argc--; argv++; element_weights[0] = atof(*argv); 
        argc--; argv++; element_weights[1] = atof(*argv); 
        argc--; argv++; element_weights[2] = atof(*argv); 
        argc--; argv++; element_weights[3] = atof(*argv); 
        argc--; argv++; element_weights[4] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-normalize")) {
        argc--; argv++; normalization_type = atoi(*argv);
      }
      else if (!strcmp(*argv, "-sig")) { 
        argc--; argv++; descriptor_type = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-debug")) { 
        debug = 1; 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else if (!shd_name) shd_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        return 0;
      }
    }
    argv++; argc--;
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check shd filename
  if (!shd_name) {
    fprintf(stderr, "You did not specify a signature file.\n");
    return 0;
  }

  // Check grid resolution
  if ((grid_resolution[0] <= 0) || (grid_resolution[1] <= 0) || (grid_resolution[2] <= 0)) {
    fprintf(stderr, "Invalid grid resolution: %d %d %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    return 0;
  }

  // Check rasterization type
  if ((rasterization_type < 0) || (rasterization_type > 3)) {
    fprintf(stderr, "Invalid rasterization type: %d\n", rasterization_type);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read PDB file
  PDBFile *file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Find ligand
  PDBResidue *ligand = NULL;
  if (ligand_name) {
    ligand = FindLigand(file, ligand_name);
    if (!ligand) exit(-1);
  }

  // Read accessible surface area file
  if (asa_name) {
    int status = ReadASAFile(file, asa_name);
    if (!status) exit(-1);
  }

  // Read grow file
  if (grow_name) {
    int status = ReadGrowFile(file, grow_name);
    if (!status) exit(-1);
  }

  // Allocate buffer for descriptor
  const int max_descriptor_size = 16 * 1024 * 1024;
  static float descriptor[max_descriptor_size];
  int descriptor_size = 0;

  // Compute/concatenate SHDs for all grids
  int ngrids = (elements) ? 5 : 1;
  for (int i = 0; i < ngrids; i++) {
    PDBElement *element = NULL;
    if (elements) {
      // Compute grid/descriptor for each element (C, N, O, P, S) separately (C=1, N=2, O=3, P=4, S=5)
      if (element_weights[i] == 0) continue;
      element = &PDBelements[i+1];
    }

    // Create grid 
    R3Grid *grid = CreateGrid(file, ligand, element);
    if (!grid) return 0;

    // Create spherical grid
    R3SphericalGrid *sgrid = CreateSphericalGrid(grid);
    if (!sgrid) exit(-1);

    // Create spherical harmonics
    R3SphericalHarmonics *harmonics = CreateSphericalHarmonics(sgrid);
    if (!harmonics) exit(-1);

    // Create spherical harmonic descriptor
    int nvalues = CreateDescriptor(harmonics, descriptor_type, &descriptor[descriptor_size]);
    if (nvalues == 0) exit(-1);

    // Update descriptor size
    descriptor_size += nvalues;
    assert(descriptor_size < max_descriptor_size);
  }

  // Normalize descriptor
  int status1 = NormalizeDescriptor(descriptor, descriptor_size, normalization_type);
  if (!status1) exit(-1);

  // Write descriptor
  int status2 = WriteDescriptor(descriptor, descriptor_size, shd_name);
  if (!status2) exit(-1);

  // Return success
  return 0;
}
