// Source file for the pdb viewer program



// Include files 

#include "PDB/PDB.h"
#include "R3Grids/R3Grids.h"



// Program variables

static char *input1_name = NULL;
static char *input2_name = NULL;
static char *ligand1_name = NULL;
static char *ligand2_name = NULL;
static char *atoms1_name = NULL;
static char *atoms2_name = NULL;
static char *output_name = NULL;
static char *wigner_grid_name = NULL;
static RNBoolean combine_files = FALSE;
static RNBoolean align_translation = TRUE;
static RNBoolean align_rotation = TRUE;
static RNBoolean align_scale = FALSE;
static RNBoolean align_elements = TRUE;
static RNBoolean align_aminoacid_types = FALSE;
static RNBoolean align_aminoacid_groups = FALSE;
static RNBoolean order_correspondence = FALSE;
static RNBoolean name_correspondence = FALSE;
static RNBoolean wigner = TRUE;
static RNBoolean icp = TRUE;
static RNScalar credit_for_each_matched_atom = 0;
static RNScalar penalty_for_each_unmatched_atom = 10;
static int max_icp_iterations = 100;
static RNScalar element_weights[5] = { 1, 1, 1, 1, 1 }; // C, N, O, P, S
static int rasterization_type = 0; // 0=point, 1=solid, 2=surface
static RNScalar distance_threshold = 0;
static int resolution = 64;
static int print_verbose = 0;
static int print_debug = 0;



// Statistics variables - hack

static RNScalar allocate_time = 0;
static RNScalar grid_time = 0;
static RNScalar sgrid_time = 0;
static RNScalar harmonics_time = 0;
static RNScalar wigner_setup_time = 0;
static RNScalar wigner_inverse_time = 0;
static RNScalar wigner_search_time = 0;



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read first file from PDB file
  if (!file->ReadFile(filename)) {
    RNFail("Unable to read PDB file for %s", filename);
    return NULL;
  }

  // Check file
  if (file->NModels() == 0) {
    RNFail("No models in PDB file %s", filename);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i = 0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static RNArray<PDBAtom *> *
FindAtoms(PDBFile *file, const char *ligand_name, const char *atoms_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get model
  if (file->NModels() == 0) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate array of atoms
  RNArray<PDBAtom *> *atoms = new RNArray<PDBAtom *>();
  if (!atoms) {
    fprintf(stderr, "Unable to allocate array of atoms.\n");
    return NULL;
  }

  // Find ligand
  PDBResidue *ligand = NULL;
  if (ligand_name) {
    ligand = file->FindResidue(ligand_name);
    if (!ligand) {
      fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_name, file->Name()); 
      return 0;
    }
  }

  // Find atoms
  if (atoms_name) {
    // Open atoms file
    FILE *fp = fopen(atoms_name, "r");
    if (!fp) {
      fprintf(stderr, "Unable to open atoms file: %s\n", atoms_name);
      return NULL;
    }

    // Read atoms file
    char atom_name[128];
    while (fgets(atom_name, 127, fp)) {
      // Remove white space from name1
      const char *name1 = atom_name;
      while (isspace(*name1)) { name1++; }
      const char *last1 = name1; while (*last1 && !isspace(*last1)) last1++;
      int length1 = last1 - name1;
        
      // Find atom
      RNBoolean found = FALSE;
      const RNArray<PDBAtom *>& tmp = (ligand) ? ligand->atoms : model->atoms;
      for (int i = 0; i < tmp.NEntries(); i++) {
        PDBAtom *atom = tmp.Kth(i);
        const char *name2 = atom->Name();
        while (isspace(*name2)) { name2++; }
        const char *last2 = name2; while (*last2 && !isspace(*last2)) last2++;
        int length2 = last2 - name2;

        // Compare names
        if (length1 == length2) {
          if (!strncmp(name1, name2, length1)) {
            atoms->Insert(atom);
            found = TRUE;
            break;
          }
        }
      }

      // Check if atom was found
      if (!found) {
        fprintf(stderr, "Unable to find atom %s in %s\n", name1, file->Name());
        return NULL;
      }
    }

    // Close atoms file
    fclose(fp);
  }
  else {
    // Copy all atoms from ligand/model
    *atoms = (ligand) ? ligand->atoms : model->atoms;
  }

  // Print statistics
  if (print_verbose) {
    printf("Found atoms ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atoms->NEntries());
    fflush(stdout);
  }

  // Return atoms
  return atoms;
}



static R4Matrix
WignerAlignmentRotation(const RNArray<PDBAtom *>& atoms1, const RNArray<PDBAtom *>& atoms2, 
  const R3Point& centroid1, const R3Point& centroid2, RNScalar radius1, RNScalar radius2, 
  RNScalar& error)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Determine maximum radius
  RNScalar radius = radius1 + radius2;

  // Allocate grids
  R3Grid raster_grid1(resolution, resolution, resolution);
  R3Grid raster_grid2(resolution, resolution, resolution);
  R3Grid distance_grid1(resolution, resolution, resolution);
  R3Grid distance_grid2(resolution, resolution, resolution);

  // Allocate spherical grids
  R3SphericalGrid raster_sgrid1(resolution, resolution, resolution/2);
  R3SphericalGrid raster_sgrid2(resolution, resolution, resolution/2);
  R3SphericalGrid distance_sgrid1(resolution, resolution, resolution/2);
  R3SphericalGrid distance_sgrid2(resolution, resolution, resolution/2);

  // Allocate spherical harmonicss
  R3SphericalHarmonics raster_harmonics1(resolution/2, resolution/2);
  R3SphericalHarmonics raster_harmonics2(resolution/2, resolution/2);
  R3SphericalHarmonics distance_harmonics1(resolution/2, resolution/2);
  R3SphericalHarmonics distance_harmonics2(resolution/2, resolution/2);

  // Set grid transformations
  raster_grid1.SetWorldToGridTransformation(centroid1, R3negz_vector, R3posy_vector, radius);
  raster_grid2.SetWorldToGridTransformation(centroid2, R3negz_vector, R3posy_vector, radius);
  distance_grid1.SetWorldToGridTransformation(centroid1, R3negz_vector, R3posy_vector, radius);
  distance_grid2.SetWorldToGridTransformation(centroid2, R3negz_vector, R3posy_vector, radius);

  // Set spherical grid transformations
  raster_sgrid1.SetWorldToGridTransformation(raster_grid1.WorldBox());
  raster_sgrid2.SetWorldToGridTransformation(raster_grid2.WorldBox());
  distance_sgrid1.SetWorldToGridTransformation(distance_grid1.WorldBox());
  distance_sgrid2.SetWorldToGridTransformation(distance_grid2.WorldBox());

  // Allocate WignerD structure
  R3WignerD wignerD(resolution/2);

  // Gather statistics
  allocate_time = start_time.Elapsed();
  start_time.Read();

  // Fill grids, sgrids, and harmonics for each element (C, N, O, P, S) and add them to WignerD structure
  int ngrids = (align_elements) ? 5 : 1;
  for (int i = 0; i < ngrids; i++) {
    // C=1, N=2, O=3, P=4, S=5
    PDBElement *element = &PDBelements[i+1];

    // Check weight
    if (align_elements && (element_weights[i] == 0)) continue;

    // Clear grids
    raster_grid1.Clear();
    raster_grid2.Clear();
    distance_grid1.Clear();
    distance_grid2.Clear();

    // Clear spherical grids
    raster_sgrid1.Clear();
    raster_sgrid2.Clear();
    distance_sgrid1.Clear();
    distance_sgrid2.Clear();

    // Clear spherical harmonics
    raster_harmonics1.Clear();
    raster_harmonics2.Clear();
    distance_harmonics1.Clear();
    distance_harmonics2.Clear();

    // Compute raster grid for atoms1
    for (int j = 0; j < atoms1.NEntries(); j++) {
      PDBAtom *atom = atoms1.Kth(j);
      RNScalar weight = (align_elements) ? element_weights[i]: 1;
      if (align_elements && (atom->Element() != element)) continue;
      if (rasterization_type == 0) raster_grid1.RasterizeWorldPoint(atom->Position(), weight);
      else raster_grid1.RasterizeWorldSphere(atom->Position(), atom->Radius(), weight);
    }

    // Compute raster grid for atoms2
    for (int j = 0; j < atoms2.NEntries(); j++) {
      PDBAtom *atom = atoms2.Kth(j);
      RNScalar weight = (align_elements) ? element_weights[i]: 1;
      if (align_elements && (atom->Element() != element)) continue;
      if (rasterization_type == 0) raster_grid2.RasterizeWorldPoint(atom->Position(), weight);
      else raster_grid2.RasterizeWorldSphere(atom->Position(), atom->Radius(), weight);
    }

    // Compute surfaces, if needed
    if (rasterization_type == 2) {
      raster_grid1.DetectEdges(); 
      raster_grid2.DetectEdges(); 
    }

    // Compute squared distance grid
    distance_grid1 = raster_grid1;
    distance_grid2 = raster_grid2;
    distance_grid1.SquaredDistanceTransform();
    distance_grid2.SquaredDistanceTransform();
    assert(distance_grid1.L1Norm() > 0);
    assert(distance_grid2.L1Norm() > 0);
    if (distance_threshold > 0) {
      RNScalar squared_distance_threshold = distance_threshold * distance_threshold;
      distance_grid1.Threshold(squared_distance_threshold, R3_GRID_KEEP_VALUE, squared_distance_threshold);
      distance_grid2.Threshold(squared_distance_threshold, R3_GRID_KEEP_VALUE, squared_distance_threshold);
    }

    // Gather statistics
    grid_time += start_time.Elapsed();
    start_time.Read();

    // Compute spherical grids
    raster_sgrid1.Add(raster_grid1);
    raster_sgrid2.Add(raster_grid2);
    distance_sgrid1.Add(distance_grid1);
    distance_sgrid2.Add(distance_grid2);

    // Gather statistics
    sgrid_time += start_time.Elapsed();
    start_time.Read();

    // Compute spherical harmonics
    raster_sgrid1.ForwardSphericalHarmonicTransform(raster_harmonics1);
    raster_sgrid2.ForwardSphericalHarmonicTransform(raster_harmonics2);
    distance_sgrid1.ForwardSphericalHarmonicTransform(distance_harmonics1);
    distance_sgrid2.ForwardSphericalHarmonicTransform(distance_harmonics2);

    // Gather statistics
    harmonics_time += start_time.Elapsed();
    start_time.Read();

    // Add spherical harmonics to WignerD structure
    for (int j = 0; j < resolution/2; j++) {
      wignerD.AddSphericalHarmonics(raster_harmonics1.Shell(j), distance_harmonics2.Shell(j));
      wignerD.AddSphericalHarmonics(distance_harmonics1.Shell(j), raster_harmonics2.Shell(j));
    }

    // Gather statistics
    wigner_setup_time += start_time.Elapsed();
    start_time.Read();
  }

  // Create an euler grid
  double *wigner_grid = new double [resolution * resolution * resolution]; 
  assert(wigner_grid); 

  // Compute the inverse WignerD transform
  wignerD.InverseWignerDTransform(wigner_grid);

  // Gather statistics
  wigner_inverse_time = start_time.Elapsed();
  start_time.Read();

  // Search for the rotation with minimum correlation
  int min_index = 0;
  RNScalar min_correlation = FLT_MAX;
  for (int i = 0; i < resolution * resolution * resolution; i++) {
    if (wigner_grid[i] < min_correlation) {
      min_correlation = wigner_grid[i];
      min_index = i;
    }
  }

  // Gather statistics
  wigner_search_time = start_time.Elapsed();
  start_time.Read();

  // Write the euler grid
  if (wigner_grid_name) {
    FILE *fp = fopen(wigner_grid_name, "wb");
    if (!fp) {
      fprintf(stderr, "Unable to open %s\n", wigner_grid_name);
    }
    else {
      // Write a .grd file
      fwrite(&resolution, sizeof(int), 1, fp);
      fwrite(&resolution, sizeof(int), 1, fp);
      fwrite(&resolution, sizeof(int), 1, fp);
      fwrite(&(R4identity_matrix[0][0]), sizeof(RNScalar), 16, fp);
      fwrite(wigner_grid, sizeof(double), resolution * resolution * resolution, fp);
      fclose(fp);
    }
  }

  // Delete euler grid
  delete [] wigner_grid;

  // Fill output variables
  R4Matrix rotation_matrix;
  R3EulerRotationMatrix(min_index, resolution, rotation_matrix);
  error = min_correlation;

  // Return the rotation matrix
  return rotation_matrix;
}



static int
PCAAlignmentTransformation(const RNArray<PDBAtom *>& atoms1, 
  RNBoolean align_translation, RNBoolean align_rotation, RNBoolean align_scale, 
  R3Affine& affine)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Align the models to canonical coordinate system
  affine = PDBAlignmentTransformation(atoms1, align_translation, align_rotation, align_scale);

  // Print statistics
  if (print_verbose) {
    const R4Matrix& m = affine.Matrix();
    printf("Computed alignment transformation with PCA ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Scale = %g\n", affine.ScaleFactor());
    printf("  Matrix[0][0-3] = %g %g %g %g\n", m[0][0], m[0][1], m[0][2], m[0][3]);
    printf("  Matrix[1][0-3] = %g %g %g %g\n", m[1][0], m[1][1], m[1][2], m[1][3]);
    printf("  Matrix[2][0-3] = %g %g %g %g\n", m[2][0], m[2][1], m[2][2], m[2][3]);
    printf("  Matrix[3][0-3] = %g %g %g %g\n", m[3][0], m[3][1], m[3][2], m[3][3]);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
OrderCorrespondenceAlignmentTransformation(const RNArray<PDBAtom *>& atoms1, const RNArray<PDBAtom *>& atoms2, 
  RNBoolean align_translation, RNBoolean align_rotation, RNBoolean align_scale,
  R3Affine& affine)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Align the two models assuming correspondence of atoms according to the order they appear in the lists
  affine = PDBAlignmentTransformation(atoms1, atoms2, NULL, align_translation, align_rotation, align_scale);

  // Compute error
  RNScalar error = PDBMeanAlignmentError(atoms1, atoms2, NULL, affine);

  // Print statistics
  if (print_verbose) {
    const R4Matrix& m = affine.Matrix();
    printf("Computed alignment transformation from order correspondences ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Matrix[0][0-3] = %g %g %g %g\n", m[0][0], m[0][1], m[0][2], m[0][3]);
    printf("  Matrix[1][0-3] = %g %g %g %g\n", m[1][0], m[1][1], m[1][2], m[1][3]);
    printf("  Matrix[2][0-3] = %g %g %g %g\n", m[2][0], m[2][1], m[2][2], m[2][3]);
    printf("  Matrix[3][0-3] = %g %g %g %g\n", m[3][0], m[3][1], m[3][2], m[3][3]);
    printf("  Scale = %g\n", affine.ScaleFactor());
    printf("  Error = %g\n", error);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
NameCorrespondenceAlignmentTransformation(const RNArray<PDBAtom *>& atoms1, const RNArray<PDBAtom *>& atoms2, 
  RNBoolean align_translation, RNBoolean align_rotation, RNBoolean align_scale,
  R3Affine& affine)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute atom correspondences
  RNArray<PDBAtom *> correspondences1;
  RNArray<PDBAtom *> correspondences2;
  for (int i = 0; i < atoms1.NEntries(); i++) {
    RNBoolean found = FALSE;
    PDBAtom *atom1 = atoms1.Kth(i);
    const char *name1 = atom1->Name();
    while (isspace(*name1)) { name1++; }
    const char *last1 = name1; while (*last1 && !isspace(*last1)) last1++;
    int length1 = last1 - name1;
    for (int j = 0; j < atoms2.NEntries(); j++) {
      PDBAtom *atom2 = atoms2.Kth(j);
      const char *name2 = atom2->Name();
      while (isspace(*name2)) { name2++; }
      const char *last2 = name2; while (*last2 && !isspace(*last2)) last2++;
      int length2 = last2 - name2;
      if (length1 == length2) {
        if (!strncmp(name1, name2, length1)) {
          correspondences1.Insert(atom1);
          correspondences2.Insert(atom2);
          found = TRUE;
          break;
        }
      }
    }
    if (!found) {
      fprintf(stderr, "No correspondence for %s\n", name1);
    }
  }

  // Align the corresponding atoms
  assert(correspondences1.NEntries() == correspondences2.NEntries());
  affine = PDBAlignmentTransformation(correspondences1, correspondences2, NULL, align_translation, align_rotation, align_scale);

  // Compute error
  int number_of_matched_atoms = correspondences1.NEntries() + correspondences2.NEntries();
  int number_of_unmatched_atoms = atoms1.NEntries() + atoms2.NEntries() - number_of_matched_atoms;
  RNScalar error = PDBMeanAlignmentError(correspondences1, correspondences2, NULL, affine);
  if ((penalty_for_each_unmatched_atom != 0) || (credit_for_each_matched_atom != 0)) {
    error *= number_of_matched_atoms;
    error += number_of_unmatched_atoms * penalty_for_each_unmatched_atom;
    error -= number_of_matched_atoms * credit_for_each_matched_atom;
    error /= number_of_matched_atoms + number_of_unmatched_atoms;
  }

  // Print statistics
  if (print_verbose) {
    const R4Matrix& m = affine.Matrix();
    printf("Computed alignment transformation from name correspondences ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d %d\n", atoms1.NEntries(), atoms2.NEntries());
    printf("  # Correspondences = %d\n", correspondences1.NEntries());
    printf("  # Unmatched = %d\n", number_of_unmatched_atoms);
    printf("  Matrix[0][0-3] = %g %g %g %g\n", m[0][0], m[0][1], m[0][2], m[0][3]);
    printf("  Matrix[1][0-3] = %g %g %g %g\n", m[1][0], m[1][1], m[1][2], m[1][3]);
    printf("  Matrix[2][0-3] = %g %g %g %g\n", m[2][0], m[2][1], m[2][2], m[2][3]);
    printf("  Matrix[3][0-3] = %g %g %g %g\n", m[3][0], m[3][1], m[3][2], m[3][3]);
    printf("  Scale = %g\n", affine.ScaleFactor());
    printf("  Error = %g\n", error);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WignerAlignmentTransformation(const RNArray<PDBAtom *>& atoms1, const RNArray<PDBAtom *>& atoms2, 
  RNBoolean align_translation, RNBoolean align_rotation, RNBoolean align_scale,
  R3Affine& affine)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  RNScalar error = 0;

  // Compute centroids to be used for translation, scale, and rotation alignment
  R3Point centroid1 = PDBCentroid(atoms1);
  R3Point centroid2 = PDBCentroid(atoms2);

  // Compute radii to be used for scale and rotation alignment
  RNScalar radius1 = 1.0;
  RNScalar radius2 = 1.0;
  if (align_scale || align_rotation) {
    radius1 = PDBAverageDistance(atoms1, centroid1);
    radius2 = PDBAverageDistance(atoms2, centroid2);
  }

  // Compute rotation matrix
  R4Matrix rotation_matrix = R4identity_matrix;
  if (align_rotation) {
    rotation_matrix = WignerAlignmentRotation(atoms1, atoms2, centroid1, centroid2, radius1, radius2, error);
  }

  // Compute complete transformation
  affine = R3identity_affine;
  if (align_translation) affine.Translate(centroid2.Vector());
  else affine.Translate(centroid1.Vector());
  if (align_scale) affine.Scale(radius2/radius1);
  if (align_rotation) affine.Transform(R3Affine(rotation_matrix));
  affine.Translate(-(centroid1.Vector()));

  // Print statistics
  if (print_verbose) {
    const R4Matrix& m = affine.Matrix();
    printf("Computed alignment transformation with inverse Wigner-D transform ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    if (allocate_time > 0) printf("    Grid Allocation = %.2f seconds\n", allocate_time);
    if (grid_time > 0) printf("    Grid Rasterization = %.2f seconds\n", grid_time);
    if (sgrid_time > 0) printf("    Spherical Grid = %.2f seconds\n", sgrid_time);
    if (harmonics_time > 0) printf("    Spherical Harmonics = %.2f seconds\n", harmonics_time);
    if (wigner_setup_time > 0) printf("    Wigner Setup = %.2f seconds\n", wigner_setup_time);
    if (wigner_inverse_time > 0) printf("    Wigner Inverse = %.2f seconds\n", wigner_inverse_time);
    if (wigner_search_time > 0) printf("    Wigner Search = %.2f seconds\n", wigner_search_time);
    printf("  Matrix[0][0-3] = %g %g %g %g\n", m[0][0], m[0][1], m[0][2], m[0][3]);
    printf("  Matrix[1][0-3] = %g %g %g %g\n", m[1][0], m[1][1], m[1][2], m[1][3]);
    printf("  Matrix[2][0-3] = %g %g %g %g\n", m[2][0], m[2][1], m[2][2], m[2][3]);
    printf("  Matrix[3][0-3] = %g %g %g %g\n", m[3][0], m[3][1], m[3][2], m[3][3]);
    printf("  Scale = %g\n", affine.ScaleFactor());
    printf("  Error = %g\n", error);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ICPAlignmentTransformation(const RNArray<PDBAtom *>& atoms1, const RNArray<PDBAtom *>& atoms2, 
  RNBoolean align_translation, RNBoolean align_rotation, RNBoolean align_scale, R3Affine& affine)
{
  // Check if should do anything
  if (max_icp_iterations <= 0) return 1;

  // Start statistics
  if (print_debug) printf("Aligning atoms ...\n");
  RNTime start_time;
  start_time.Read();

  // Get arrays of atoms
  if (atoms1.IsEmpty() || atoms2.IsEmpty()) return 1;
  static RNArray<PDBAtom *> correspondences1[2];
  static RNArray<PDBAtom *> correspondences2[2];
  int iteration_index = 0;

  // Get initial max distance for correspondence (has to start big to handle gross initial misalignments)
  RNLength max_distance_squared = RN_INFINITY;
  RNLength distance_threshold_squared = distance_threshold * distance_threshold;

  // Iterate solving transformation with computed correspondences
  for (int iteration = 0; iteration < max_icp_iterations; iteration++) {
    // Initialize at start of "swap buffers"
    iteration_index = iteration % 2;
    correspondences1[iteration_index].Empty();
    correspondences2[iteration_index].Empty();

    // Find closest atom in file2 for each atom in file1
    for (int i = 0; i < atoms1.NEntries(); i++) {
      PDBAtom *atom1 = atoms1[i];
      R3Point p1 = atom1->Position();
      p1.Transform(affine);

      // Find closest atom in file2 
      PDBAtom *closest_atom2 = NULL;
      RNLength closest_distance_squared = FLT_MAX;
      for (int j = 0; j < atoms2.NEntries(); j++) {
        PDBAtom *atom2 = atoms2[j];

        // Check element compatability
        if (align_elements) {
          if (atom1->Element() != atom2->Element()) continue;
        }

        // Check aminoacid compatability
        if ((align_aminoacid_types) || (align_aminoacid_groups)) {
          PDBAminoAcid *aminoacid1 = atom1->AminoAcid();
          PDBAminoAcid *aminoacid2 = atom2->AminoAcid();
          if (!aminoacid1) { if (aminoacid2) continue; }
          else if (!aminoacid2) { if (aminoacid1) continue; }
          else if (align_aminoacid_groups) { if (aminoacid1->Group() != aminoacid2->Group()) continue; }
          else { if (aminoacid1 != aminoacid2) continue; }
        }
      
        // Compute distance
        const R3Point& p2 = atom2->Position();
        RNLength dx = p1.X() - p2.X();
        RNLength dy = p1.Y() - p2.Y();
        RNLength dz = p1.Z() - p2.Z();
        RNLength distance_squared = dx * dx + dy * dy + dz * dz;
      
        // Check distance
        if (distance_squared < max_distance_squared) {
          if (distance_squared < closest_distance_squared) {
            closest_distance_squared = distance_squared;
            closest_atom2 = atom2;
          }
        }
      }

      // Set correspondence to closest atom in file2
      if (closest_atom2) {
        correspondences1[iteration_index].Insert(atom1);
        correspondences2[iteration_index].Insert(closest_atom2);
      }
      else {
        // printf("  Atom %s from %s unmatched\n", atom1->Name(), file1->Name());
      }
    }

    // Find closest atom in file1 for each atom in file2
    for (int i = 0; i < atoms2.NEntries(); i++) {
      PDBAtom *atom2 = atoms2[i];
      const R3Point& p2 = atom2->Position();

      // Find closest atom in file1 
      PDBAtom *closest_atom1 = NULL;
      RNLength closest_distance_squared = FLT_MAX;
      for (int j = 0; j < atoms1.NEntries(); j++) {
        PDBAtom *atom1 = atoms1[j];

        // Check element compatability
        if (align_elements) {
          if (atom1->Element() != atom2->Element()) continue;
        }

        // Check aminoacid compatability
        if ((align_aminoacid_types) || (align_aminoacid_groups)) {
          PDBAminoAcid *aminoacid1 = atom1->AminoAcid();
          PDBAminoAcid *aminoacid2 = atom2->AminoAcid();
          if (!aminoacid1) { if (aminoacid2) continue; }
          else if (!aminoacid2) { if (aminoacid1) continue; }
          else if (align_aminoacid_groups) { if (aminoacid1->Group() != aminoacid2->Group()) continue; }
          else { if (aminoacid1 != aminoacid2) continue; }
        }
      
        // Compute distance
        R3Point p1 = atom1->Position();
        p1.Transform(affine);
        RNLength dx = p2.X() - p1.X();
        RNLength dy = p2.Y() - p1.Y();
        RNLength dz = p2.Z() - p1.Z();
        RNLength distance_squared = dx * dx + dy * dy + dz * dz;
        if (distance_squared < max_distance_squared) {
          if (distance_squared < closest_distance_squared) {
            closest_distance_squared = distance_squared;
            closest_atom1 = atom1;
          }
        }
      }

      // Set correspondence to closest atom in file1
      if (closest_atom1) {
        correspondences1[iteration_index].Insert(closest_atom1);
        correspondences2[iteration_index].Insert(atom2);
      }
      else {
        // printf("  Atom %s from %s unmatched\n", atom2->Name(), file2->Name());
      }
    }
    
    // Check if correspondences are different than last iteration
    if ((iteration > 0) && ((distance_threshold == 0) || (max_distance_squared <= distance_threshold_squared))) {
      RNBoolean different = FALSE;
      if ((correspondences1[iteration_index].NEntries() == correspondences1[1 - iteration_index].NEntries())) {
        assert(correspondences2[iteration_index].NEntries() == correspondences2[1 - iteration_index].NEntries());
        for (int i = 0; i < correspondences1[iteration_index].NEntries(); i++) {
          if ((correspondences1[iteration_index][i] != correspondences1[1 - iteration_index][i]) ||
              (correspondences2[iteration_index][i] != correspondences2[1 - iteration_index][i])) {
            different = TRUE;
            break;
          }
        }
      }
      if (!different) break;
    }

    // Compute alignment with correspondences
    affine = PDBAlignmentTransformation(correspondences1[iteration_index], correspondences2[iteration_index], 
      NULL, align_translation, align_rotation, align_scale);

    // Update the maximum distance for correspondences
    if ((distance_threshold > 0) && (max_distance_squared > distance_threshold_squared)) {
      max_distance_squared *= 0.75;
      RNScalar error = PDBMeanAlignmentError(correspondences1[iteration_index], correspondences2[iteration_index], NULL, affine);
      RNScalar error_squared = error * error;
      if (max_distance_squared > error_squared) max_distance_squared = error_squared;
      if (max_distance_squared < distance_threshold_squared) max_distance_squared = distance_threshold_squared;
    }

    // Print debug message for each iteration
    if (print_debug) {
      int number_of_matched_atoms = correspondences1[iteration_index].NEntries();
      int number_of_unmatched_atoms = atoms1.NEntries() + atoms2.NEntries() - number_of_matched_atoms;
      RNScalar error = PDBMeanAlignmentError(correspondences1[iteration_index], correspondences2[iteration_index], NULL, affine);
      if ((penalty_for_each_unmatched_atom != 0) || (credit_for_each_matched_atom != 0)) {
        error *= number_of_matched_atoms;
        error += number_of_unmatched_atoms * penalty_for_each_unmatched_atom;
        error -= number_of_matched_atoms * credit_for_each_matched_atom;
        error /= number_of_matched_atoms + number_of_unmatched_atoms;
      }
      printf("    E %d = %g (%d %g)\n", iteration+1, error, number_of_unmatched_atoms, penalty_for_each_unmatched_atom);
    }
  }

  // Print statistics
  if (print_verbose) {
    const R4Matrix& m = affine.Matrix();
    int number_of_matched_atoms = correspondences1[iteration_index].NEntries();
    int number_of_unmatched_atoms = atoms1.NEntries() + atoms2.NEntries() - number_of_matched_atoms;
    RNScalar error = PDBMeanAlignmentError(correspondences1[iteration_index], correspondences2[iteration_index], NULL, affine);
    if ((penalty_for_each_unmatched_atom != 0) || (credit_for_each_matched_atom != 0)) {
      error *= number_of_matched_atoms;
      error += number_of_unmatched_atoms * penalty_for_each_unmatched_atom;
      error -= number_of_matched_atoms * credit_for_each_matched_atom;
      error /= number_of_matched_atoms + number_of_unmatched_atoms;
    }
    printf("Computed alignment transformation with ICP ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Unmatched = %d\n", number_of_unmatched_atoms);
    printf("  Matrix[0][0-3] = %g %g %g %g\n", m[0][0], m[0][1], m[0][2], m[0][3]);
    printf("  Matrix[1][0-3] = %g %g %g %g\n", m[1][0], m[1][1], m[1][2], m[1][3]);
    printf("  Matrix[2][0-3] = %g %g %g %g\n", m[2][0], m[2][1], m[2][2], m[2][3]);
    printf("  Matrix[3][0-3] = %g %g %g %g\n", m[3][0], m[3][1], m[3][2], m[3][3]);
    printf("  Scale = %g\n", affine.ScaleFactor());
    printf("  Error = %g\n", error);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
TransformFile(PDBFile *file, const R3Affine& affine)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Transform the file
  file->Transform(affine);

  // Print statistics
  if (print_verbose) {
    printf("Transformed file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
CombineFiles(PDBFile *file1, PDBFile *file2)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Add models from file2 into file1
  for (int i = 0; i < file2->NModels(); i++) 
    file1->InsertCopy(file2->Model(i));

  // Print statistics
  if (print_verbose) {
    printf("Combined files ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int 
WritePDB(PDBFile *file, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write PDB file
  int status = file->WriteFile(filename);
  if (!status) {
    fprintf(stderr, "Unable to write %s\n", filename);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Wrote PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i = 0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %d ...\n", i);
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: pdbalign input1 input2 output [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-debug")) print_debug = 1; 
      else if (!strcmp(*argv, "-order_correspondence")) { order_correspondence = TRUE; icp = FALSE;} 
      else if (!strcmp(*argv, "-name_correspondence")) { name_correspondence = TRUE; icp = FALSE;} 
      else if (!strcmp(*argv, "-wigner")) wigner = TRUE; 
      else if (!strcmp(*argv, "-nowigner")) wigner = FALSE; 
      else if (!strcmp(*argv, "-icp")) icp = TRUE;
      else if (!strcmp(*argv, "-noicp")) icp = FALSE;
      else if (!strcmp(*argv, "-translation")) align_translation = TRUE; 
      else if (!strcmp(*argv, "-notranslation")) align_translation = FALSE;
      else if (!strcmp(*argv, "-rotation")) align_rotation = TRUE; 
      else if (!strcmp(*argv, "-norotation")) align_rotation = FALSE;
      else if (!strcmp(*argv, "-scale")) align_scale = TRUE; 
      else if (!strcmp(*argv, "-noscale")) align_scale = FALSE;
      else if (!strcmp(*argv, "-elements")) align_elements = TRUE; 
      else if (!strcmp(*argv, "-noelements")) align_elements = FALSE; 
      else if (!strcmp(*argv, "-aminoacid_types")) align_aminoacid_types = TRUE; 
      else if (!strcmp(*argv, "-noaminoacid_types")) align_aminoacid_types = FALSE; 
      else if (!strcmp(*argv, "-aminoacid_groups")) align_aminoacid_groups = TRUE; 
      else if (!strcmp(*argv, "-noaminoacid_groups")) align_aminoacid_groups = FALSE; 
      else if (!strcmp(*argv, "-penalty_for_each_unmatched_atom")) { argc--; argv++; penalty_for_each_unmatched_atom = atof(*argv); }
      else if (!strcmp(*argv, "-credit_for_each_matched_atom")) { argc--; argv++; credit_for_each_matched_atom = atof(*argv); }
      else if (!strcmp(*argv, "-distance_threshold")) { argc--; argv++; distance_threshold = atof(*argv); }
      else if (!strcmp(*argv, "-rasterization")) { argc--; argv++; rasterization_type = atoi(*argv); }
      else if (!strcmp(*argv, "-resolution")) { argc--; argv++; resolution = atoi(*argv); }
      else if (!strcmp(*argv, "-icp_iterations")) { argc--; argv++; max_icp_iterations = atoi(*argv); }
      else if (!strcmp(*argv, "-wigner_grid")) { argc--; argv++; wigner_grid_name = *argv; }
      else if (!strcmp(*argv, "-combine")) combine_files = TRUE; 
      else if (!strcmp(*argv, "-ligand1")) { argc--; argv++; ligand1_name = *argv; }
      else if (!strcmp(*argv, "-ligand2")) { argc--; argv++; ligand2_name = *argv; }
      else if (!strcmp(*argv, "-atoms1")) { argc--; argv++; atoms1_name = *argv; }
      else if (!strcmp(*argv, "-atoms2")) { argc--; argv++; atoms2_name = *argv; }
      else if (!strcmp(*argv, "-weights")) { 
        argc--; argv++; element_weights[0] = atof(*argv); 
        argc--; argv++; element_weights[1] = atof(*argv); 
        argc--; argv++; element_weights[2] = atof(*argv); 
        argc--; argv++; element_weights[3] = atof(*argv); 
        argc--; argv++; element_weights[4] = atof(*argv); 
      }
      else { fprintf(stderr, "Invalid program argument: %s\n", *argv); exit(1); }
    }
    else {
      if (!input1_name) input1_name = *argv;
      else if (!input2_name) input2_name = *argv;
      else if (!output_name) output_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s\n", *argv); exit(1); }
    }
    argv++; argc--;
  }

  // Check input filename
  if (!input1_name) {
    fprintf(stderr, "You did not specify an input pdb file.\n");
    return 0;
  }

  // Check ligand name
  if (!ligand2_name && !input2_name) {
    fprintf(stderr, "You specified a -ligand2, but not a second pdb file.\n");
    return 0;
  }

  // Check output filename
  if (!output_name) {
    if (input2_name) {
      output_name = input2_name;
      input2_name = NULL;
    }
    else {
      fprintf(stderr, "You did not specify an output pdb file.\n");
      return 0;
    }
  }

  // Check rasterization type
  if ((rasterization_type < 0) || (rasterization_type > 2)) {
    fprintf(stderr, "Invalid rasterization type: %d\n", rasterization_type);
    return 0;
  }

  // Check resolution
  if (resolution <= 0) {
    fprintf(stderr, "Invalid grid resolution: %d\n", resolution);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Initialize affine transformation
  R3Affine affine = R3identity_affine;

  // Read first input file
  PDBFile *file1 = ReadPDB(input1_name);
  if (!file1) exit(-1);

  // Read second input file
  PDBFile *file2 = NULL;
  if (input2_name) {
    file2 = ReadPDB(input2_name);
    if (!file2) exit(-1);
  }

  // Find atoms in first input file
  RNArray<PDBAtom *> *atoms1 = FindAtoms(file1, ligand1_name, atoms1_name);
  if (!atoms1) exit(-1);

  // Compute alignment transformation
  if (!file2) {
    // Compute alignment transformation for file1 to canonical coordinate system
    int status = PCAAlignmentTransformation(*atoms1, align_translation, align_rotation, align_scale, affine);
    if (!status) exit(-1);
  }    
  else {
    // Find atoms in second input file
    RNArray<PDBAtom *> *atoms2 = FindAtoms(file2, ligand2_name, atoms2_name);
    if (!atoms2) exit(-1);

    // Compute alignment transformation for file1 to file2
    if (order_correspondence) {
      // Compute alignment transformation using correspondences infered from order of atoms in PDB files
      int status = OrderCorrespondenceAlignmentTransformation(*atoms1, *atoms2, align_translation, align_rotation, align_scale, affine);
      if (!status) exit(-1);
    }
    else if (name_correspondence) {
      // Compute alignment transformation using correspondences infered from atom names
      int status = NameCorrespondenceAlignmentTransformation(*atoms1, *atoms2, align_translation, align_rotation, align_scale, affine);
      if (!status) exit(-1);
    }
    else if (wigner) {
      // Compute alignment transformation using inverse Wigner-D transform
      int status = WignerAlignmentTransformation(*atoms1, *atoms2, align_translation, align_rotation, align_scale, affine);
      if (!status) exit(-1);
    }

    // Refine alignment transformation using ICP
    if (icp) {
      int status = ICPAlignmentTransformation(*atoms1, *atoms2, align_translation, align_rotation, align_scale, affine);
      if (!status) exit(-1);
    }
  }

  // Transform file1 to align with file2
  int status = TransformFile(file1, affine);
  if (!status) exit(-1);

  // Combine file2 into file1
  if (combine_files) {
    int status = CombineFiles(file1, file2);
    if (!status) exit(-1);
  }
  
  // Write PDB file
  status = WritePDB(file1, output_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}


