// Source file for KDTree class

#ifndef __KDTREE__I__
#define __KDTREE__I__





////////////////////////////////////////////////////////////////////////
// Include files
////////////////////////////////////////////////////////////////////////

#include "R3Shapes/R3Shapes.h"
#include "KDTree.h"





////////////////////////////////////////////////////////////////////////
// Constant definitions
////////////////////////////////////////////////////////////////////////

static const int max_points_per_kdtree_node = 32;





////////////////////////////////////////////////////////////////////////
// Node class definition
////////////////////////////////////////////////////////////////////////

// Node declaration

template <class PtrType>
class KDTreeNode {
public:
  KDTreeNode(KDTreeNode<PtrType> *parent);

public:
  class KDTreeNode<PtrType> *parent;
  class KDTreeNode<PtrType> *children[2];
  RNScalar split_coordinate;
  RNDimension split_dimension;
  PtrType points[max_points_per_kdtree_node];
  int npoints;
};



// Node constructor

template <class PtrType>
KDTreeNode<PtrType>::
KDTreeNode(KDTreeNode<PtrType> *parent)
  : parent(parent),
    npoints(0)
{
  // Initialize everything
  children[0] = NULL;
  children[1] = NULL;
  split_coordinate = 0;
  split_dimension = 0;
}





////////////////////////////////////////////////////////////////////////
// Public tree-level functions
////////////////////////////////////////////////////////////////////////

template <class PtrType>
KDTree<PtrType>::
KDTree(const R3Box& bbox, int position_offset)
  : bbox(bbox),
    position_offset(position_offset),
    nnodes(0)
{
  // Create root node
  root = new KDTreeNode<PtrType>(NULL);
  assert(root);
}



template <class PtrType>
KDTree<PtrType>::
KDTree(const RNArray<PtrType>& points, int position_offset)
  : position_offset(position_offset),
    nnodes(1)
{
  // Create root node
  root = new KDTreeNode<PtrType>(NULL);
  assert(root);

  // Determine bounding box 
  bbox = R3null_box;
  for (int i = 0; i < points.NEntries(); i++) {
    bbox.Union(Position(points[i]));
  }

  // Allocate copy of points array (so that it can be sorted)
  PtrType *copy = new PtrType [ points.NEntries() ];
  assert(copy);

  // Copy points
  for (int i = 0; i < points.NEntries(); i++) 
    copy[i] = points[i];

  // Insert points into root
  InsertPoints(root, bbox, copy, points.NEntries());

  // Delete copy of points array
  delete [] copy;
}



template <class PtrType>
KDTree<PtrType>::
~KDTree(void)
{
  // Check root
  if (!root) return;

  // Traverse tree deleting nodes
  RNArray<KDTreeNode<PtrType> *> stack;
  stack.InsertTail(root);
  while (!stack.IsEmpty()) {
    KDTreeNode<PtrType> *node = stack.Tail();
    stack.RemoveTail();
    if (node->children[0]) stack.Insert(node->children[0]);
    if (node->children[1]) stack.Insert(node->children[1]);
    delete node;
  }
}



template <class PtrType>
const R3Box& KDTree<PtrType>::
BBox(void) const
{
  // Return bounding box of the whole KD tree
  return bbox;
}



template <class PtrType>
int KDTree<PtrType>::
NNodes(void) const
{
  // Return number of nodes
  return nnodes;
}



#if 0

template <class PtrType>
int KDTree<PtrType>::
NEntries(void) const
{
  // Check root
  if (!root) return 0;

  // Traverse tree to count number of points
  int npoints = 0;
  RNArray<KDTreeNode<PtrType> *> stack;
  stack.InsertTail(root);
  while (!stack.IsEmpty()) {
    KDTreeNode<PtrType> *node = stack.Tail();
    stack.RemoveTail();
    npoints += node->npoints;
    if (node->children[0]) stack.Insert(node->children[0]);
    if (node->children[1]) stack.Insert(node->children[1]);
  }

  // Return total number of points
  return npoints;
}

#endif



template <class PtrType>
void KDTree<PtrType>::
FindClosest(KDTreeNode<PtrType> *node, const R3Box& node_box, const R3Point& position, RNScalar min_distance, RNScalar max_distance,
  PtrType& closest_point, RNScalar& closest_distance) const
{
  // Check if node is interior
  if (node->children[0]) {
    assert(node->children[1]);
    RNLength distance = R3Distance(position, node_box);
    if (distance >= closest_distance) return;
    RNLength side = position[node->split_dimension] - node->split_coordinate;
    if (side < closest_distance) {
      R3Box child_box(node_box);
      child_box[RN_HI][node->split_dimension] = node->split_coordinate;
      FindClosest(node->children[0], child_box, position, min_distance, max_distance, closest_point, closest_distance);
    }
    if (-side < closest_distance) {
      R3Box child_box(node_box);
      child_box[RN_LO][node->split_dimension] = node->split_coordinate;
      FindClosest(node->children[1], child_box, position, min_distance, max_distance, closest_point, closest_distance);
    }
  }
  else {
    for (int i = 0; i < node->npoints; i++) {
      PtrType point = node->points[i];
      RNLength distance = R3Distance(position, Position(point));
      if ((distance >= min_distance) && (distance <= closest_distance)) {
        closest_distance = distance;
        closest_point = point;
      }
    }
  }
}



template <class PtrType>
PtrType KDTree<PtrType>::
FindClosest(const R3Point& position, RNScalar min_distance, RNScalar max_distance, RNScalar *closest_distance) const
{
  // Check root
  if (!root) return NULL;

  // Initialize closest point 
  PtrType nearest_point = NULL;
  RNLength nearest_distance = max_distance;

  // Search nodes recursively
  FindClosest(root, bbox, position, min_distance, max_distance, nearest_point, nearest_distance);

  // Return closest distance
  if (closest_distance) *closest_distance = nearest_distance;

  // Return closest point
  return nearest_point;
}



template <class PtrType>
PtrType KDTree<PtrType>::
FindClosest(PtrType point, RNScalar min_distance, RNScalar max_distance, RNScalar *closest_distance) const
{
  // Find the closest point
  return FindClosest(Position(point), min_distance, max_distance, closest_distance);
}



template <class PtrType>
void KDTree<PtrType>::
FindAll(KDTreeNode<PtrType> *node, const R3Box& node_box, const R3Point& position, 
  RNScalar min_distance, RNScalar max_distance, RNArray<PtrType>& points) const
{
  // Check if node is interior
  if (node->children[0]) {
    assert(node->children[1]);
    RNLength distance = R3Distance(position, node_box);
    if (distance >= max_distance) return;
    RNLength side = position[node->split_dimension] - node->split_coordinate;
    if (side < max_distance) {
      R3Box child_box(node_box);
      child_box[RN_HI][node->split_dimension] = node->split_coordinate;
      FindAll(node->children[0], child_box, position, min_distance, max_distance, points);
    }
    if (-side < max_distance) {
      R3Box child_box(node_box);
      child_box[RN_LO][node->split_dimension] = node->split_coordinate;
      FindAll(node->children[1], child_box, position, min_distance, max_distance, points);
    }
  }
  else {
    for (int i = 0; i < node->npoints; i++) {
      PtrType point = node->points[i];
      RNLength distance = R3Distance(position, Position(point));
      if ((distance >= min_distance) && (distance <= max_distance)) points.Insert(point);
    }
  }
}



template <class PtrType>
int KDTree<PtrType>::
FindAll(const R3Point& position, RNScalar min_distance, RNScalar max_distance, RNArray<PtrType>& points) const
{
  // Check root
  if (!root) return 0;

  // Search nodes recursively
  FindAll(root, bbox, position, min_distance, max_distance, points);

  // Return number of points
  return points.NEntries();
}



template <class PtrType>
int KDTree<PtrType>::
FindAll(PtrType point, RNScalar min_distance, RNScalar max_distance, RNArray<PtrType>& points) const
{
  // Find all within some distance
  return FindAll(Position(point), min_distance, max_distance, points);
}




////////////////////////////////////////////////////////////////////////
// Internal tree creation functions
////////////////////////////////////////////////////////////////////////

template <class PtrType>
int KDTree<PtrType>::
PartitionPoints(PtrType *points, int npoints, RNDimension dim, int imin, int imax)
{
  // Check range
  assert(imin <= imax);
  assert(imin >= 0);
  assert(imin < npoints);
  assert(imax >= 0);
  assert(imax < npoints);
  if (imax == imin) return imin;

  // Choose a coordinate at random to split upon
  int irand = (int) (imin + RNRandomScalar() * (imax - imin));
  if (irand < imin) irand = imin;
  if (irand >= imax) irand = imax;
  RNCoord split_coord = Position(points[irand])[dim];

#if 0
  printf("%d %d %d %d : %g\n", dim, npoints, imin, imax, split_coord);
#endif

  // Partition points according to coordinate
  int left_index = imin;
  int right_index = imax;
  while (left_index < right_index) {
    while (left_index < right_index) {
      RNCoord left_coord = Position(points[left_index])[dim];
      if (left_coord >= split_coord) break;
      left_index++;
    }
    while (left_index < right_index) {
      RNCoord right_coord = Position(points[right_index])[dim];
      if (right_coord <= split_coord) break;
      right_index--;
    }
    if (left_index < right_index) {
      PtrType swap = points[right_index];
      points[right_index] = points[left_index];
      points[left_index] = swap;
      left_index++;
      right_index--;
    }
  }

#if 0
  if (imax-imin < 3) {
    printf("%d %d\n", left_index, right_index);
    for (int i = imin; i <= imax; i++) 
      printf(" %g", Position(points[i])[dim]);
    printf("\n");
  }
#endif

  // Handle boundary cases when less than 2 points
  if (imax-1 == imin) return imin;
  if ((left_index == right_index) && (left_index < imax-1)) {
    RNCoord left_coord = Position(points[left_index])[dim];
    if (left_coord <= split_coord) left_index++;
  }

  // Recurse until we find the median
  if (left_index > npoints/2) return PartitionPoints(points, npoints, dim, imin, left_index);
  else return PartitionPoints(points, npoints, dim, left_index, imax);
}



template <class PtrType>
void KDTree<PtrType>::
InsertPoints(KDTreeNode<PtrType> *node, const R3Box& node_box, PtrType *points, int npoints) 
{
  // Make sure node is an empty leaf
  assert(node);
  assert(node->children[0] == NULL);
  assert(node->children[1] == NULL);
  assert(node->npoints == 0);

  // Check number of points
  if (npoints <= max_points_per_kdtree_node) {
    // Insert new points into leaf node and return
    for (int i = 0; i < npoints; i++) {
      node->points[node->npoints++] = points[i];
    }
  }
  else {
    // Find dimension to split along
    node->split_dimension = node_box.LongestAxis();
    
    // Partition points according to coordinates in split_dimension
    int split_index = PartitionPoints(points, npoints, node->split_dimension, 0, npoints-1);
    assert((split_index >= 0) && (split_index < npoints));

    // Determine split coordinate
    node->split_coordinate = Position(points[split_index])[node->split_dimension];
    assert(node->split_coordinate >= node_box[RN_LO][node->split_dimension]);
    assert(node->split_coordinate <= node_box[RN_HI][node->split_dimension]);

    // Construct children node boxes
    R3Box node0_box(node_box);
    R3Box node1_box(node_box);
    node0_box[RN_HI][node->split_dimension] = node->split_coordinate;
    node1_box[RN_LO][node->split_dimension] = node->split_coordinate;

    // Create children
    node->children[0] = new KDTreeNode<PtrType>(node);
    node->children[1] = new KDTreeNode<PtrType>(node);

    // Insert points into children
    InsertPoints(node->children[0], node0_box, points, split_index);
    InsertPoints(node->children[1], node1_box, &points[split_index], npoints - split_index);

    // Increment number of nodes
    nnodes += 2;
  }
}



template <class PtrType>
void KDTree<PtrType>::
Outline(KDTreeNode<PtrType> *node, const R3Box& node_box) const
{
  // Draw kdtree nodes recursively
  if (node->children[0]) {
    assert(node->children[1]);
    assert(node->split_coordinate >= node_box[RN_LO][node->split_dimension]);
    assert(node->split_coordinate <= node_box[RN_HI][node->split_dimension]);
    R3Box child0_box(node_box);
    R3Box child1_box(node_box);
    child0_box[RN_HI][node->split_dimension] = node->split_coordinate;
    child1_box[RN_LO][node->split_dimension] = node->split_coordinate;
    Outline(node->children[0], child0_box);
    Outline(node->children[1], child1_box);
  }
  else {
    node_box.Outline();
  }
}



template <class PtrType>
void KDTree<PtrType>::
Outline(void) const
{
  // Draw kdtree nodes recursively
  if (!root) return;
  Outline(root, bbox);
}



#endif


