R3Grid *
CreateDistanceGrid(PDBModel *model, R3Grid *raster_grid, R3Grid *&voronoi_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate distance grid
  R3Grid *grid = new R3Grid(raster_grid->XResolution(), raster_grid->YResolution(), raster_grid->ZResolution());
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Allocate voronoi grid
  voronoi_grid = new R3Grid(*raster_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Compute voronoi regions and squared distance transform
  voronoi_grid->Voronoi(grid);

  // Convert squared distances
  grid->Sqrt();
  grid->Multiply(grid->GridToWorldScaleFactor());

  // Print statistics
  if (print_verbose) {
    printf("Created distance grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}




R3Grid *
CreateConservationGrid(PDBModel *model, R3Grid *raster_grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*raster_grid);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Create splat filter
  RNScalar filter_sigma = ligand_size * grid->WorldToGridScaleFactor();
  int filter_radius = 3 * filter_sigma;
  int filter_width = 2 * filter_radius + 1;
  R3Grid filter(filter_width, filter_width, filter_width);
  R3Point filter_origin(filter_radius, filter_radius, filter_radius);
  filter.SetGridValue(filter_radius, filter_radius, filter_radius, 1);
  filter.SquaredDistanceTransform();
  filter.Gauss(filter_sigma, FALSE);

  // Splat filter at every atom position
  for (int a = 0; a < model->NAtoms(); a++) {
    // Get atom
    PDBAtom *atom = model->Atom(a);

    // Check atom
    if (atom->IsHetAtom()) continue;
    if (atom->accessible_surface_area == 0) continue;

    // Compute atom position
    const R3Point& grid_position = grid->GridPosition(atom->Position());

    // Compute amplitude of splat
    RNScalar conservation = 0.5 * (2 - atom->conservation);
    if (atom->conservation == -999) conservation = 0;
    if (conservation < 0) conservation = 0;
    else if (conservation > 1) conservation = 1;

    // Splat filter into grid
    grid->Add(filter, grid_position, filter_origin, conservation);
  }

  // Print statistics
  if (print_verbose) {
    printf("Created conservation grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  Volume = %g\n", grid->Volume());
    RNInterval grid_range = grid->Range();
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  Filter radius = %d\n", filter_radius);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



////////////////////////////////////////////////////////////////////////
// DISTANCE GRID STUFF
////////////////////////////////////////////////////////////////////////

#if 0

R3Grid *
CreateDistanceGrid(PDBFile *file)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Create grid with 1/distance to a central point
  R3Grid distance_grid(2*grid_resolution[0], 2*grid_resolution[1], 2*grid_resolution[2]);
  distance_grid.SetWorldToGridTransformation(model->BBox());
  distance_grid.SetGridValue(grid_resolution[0], grid_resolution[1], grid_resolution[2], 1);
  distance_grid.SquaredDistanceTransform();
  distance_grid.Sqrt();
  distance_grid.Multiply(grid->GridToWorldScaleFactor());

  // Rasterize each atom into grid
  for (int a = 0; a < model->NAtoms(); a++) {
    int i, j, k;
    RNScalar x, y, z;
    PDBAtom *atom = model->Atom(a);
    if (atom->IsHetAtom()) continue;
    if (atom->accessible_surface_area == 0)) continue;
    RNScalar conservation =  2 - atom->conservation;
    if (conservation < 1.0E-6) conservation = 1.0E-6;
    if (atom->conservation == -999) conservation = 1.0E-6;
    const R3Point& atom_position = atom->Position();
    const R3Point& grid_position = grid->GridPosition(atom_position);
    RNScalar cell_radius = grid->GridToWorldScaleFactor();
    if (cell_radius > solvent_radius) cell_radius = solvent_radius;
    RNScalar radius = atom->Radius() + solvent_radius - cell_radius;
    RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
    for (k = 0, z = grid->ZResolution() - grid_position.Z(); k < grid->ZResolution(); k++, z++) {
      for (j = 0, y = grid->YResolution() - grid_position.Y(); j < grid->YResolution(); j++, y++) {
        for (i = 0, x = grid->XResolution() - grid_position.X(); i < grid->XResolution(); i++, x++) {
          RNScalar distance = distance_grid.GridValue(x, y, z);
          if (distance < radius) *grid_valuesp = -FLT_MAX;
          else *grid_valuesp += conservation / distance;
          grid_valuesp++;
        }
      }
    }
  }

  // Assign 0 everywhere inside an atom
  grid->Threshold(0, 0, R3_GRID_KEEP_VALUE);

  // Print statistics
  if (print_verbose) {
    RNInterval range = grid->Range();
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Volume = %g\n", grid->Volume());
    printf("  Minimum = %g\n", range.Min());
    printf("  Maximum = %g\n", range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}

#elif 0

R3Grid *
CreateDistanceGrid(PDBFile *file)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return NULL;
  PDBModel *model = file->Model(0);

  // Determine minimum and maximum distance from atom
  RNScalar min_distance = solvent_radius; // + 1.8;
  RNScalar max_distance = 10;

  // Determine grid resolution
  const R3Box& bbox = model->BBox();
  int grid_xresolution = 4 * bbox.XLength() / min_distance;
  int grid_yresolution = 4 * bbox.YLength() / min_distance;
  int grid_zresolution = 4 * bbox.ZLength() / min_distance;
  if (grid_xresolution > grid_resolution[0]) grid_xresolution = grid_resolution[0];
  if (grid_yresolution > grid_resolution[1]) grid_yresolution = grid_resolution[1];
  if (grid_zresolution > grid_resolution[2]) grid_zresolution = grid_resolution[2];

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_xresolution, grid_yresolution, grid_zresolution);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Create splat filter
  int filter_resolution = 2 * max_distance * grid->WorldToGridScaleFactor();
  if ((filter_resolution % 2) == 0) filter_resolution++;
  int filter_center = (filter_resolution-1) / 2;
  R3Point filter_origin(filter_center, filter_center, filter_center);
  R3Grid filter(filter_resolution, filter_resolution, filter_resolution);
#if 0
  RNScalar filter_sigma = max_distance * grid->WorldToGridScaleFactor() / 3;
  filter.SetGridValue(filter_center, filter_center, filter_center, 1);
  filter.SquaredDistanceTransform();
  filter.Multiply(grid->GridToWorldScaleFactor() * grid->GridToWorldScaleFactor());
  filter.Gauss(filter_sigma, FALSE);
  filter.Divide(filter.Maximum());
#else
  filter.RasterizeGridSphere(filter_origin, filter_center, 1);
#endif

  // Splat filter at every atom position
  for (int a = 0; a < model->NAtoms(); a++) {
    // Get atom
    PDBAtom *atom = model->Atom(a);

    // Check atom
    if (atom->IsHetAtom()) continue;
    if (atom->accessible_surface_area == 0) continue;

    // Compute atom position
    const R3Point& grid_position = grid->GridPosition(atom->Position());

    // Compute amplitude of splat
    RNScalar conservation =  2 - atom->conservation;
    if (conservation < 1.0E-6) conservation = 1.0E-6;
    if (atom->conservation == -999) conservation = 1.0E-6;
    RNScalar amplitude = conservation;

    // Splat filter into grid
    grid->Add(filter, grid_position, filter_origin, amplitude);
  }

  // Compute mask
  R3Grid mask(grid_xresolution, grid_yresolution, grid_zresolution);
  mask.SetWorldToGridTransformation(grid->WorldToGridTransformation());
  for (int a = 0; a < model->NAtoms(); a++) {
    PDBAtom *atom = model->Atom(a);
    if (atom->IsHetAtom()) continue;
    mask.RasterizeWorldSphere(atom->Position(), atom->Radius() + solvent_radius, 1.0);
  }
  mask.Threshold(0.5, 1, 0);

  // Mask grid
  grid->Mask(mask);

  // Print statistics
  if (print_verbose) {
    RNInterval grid_range = grid->Range();
    RNInterval filter_range = filter.Range();
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Grid Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Grid Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Grid Volume = %g\n", grid->Volume());
    printf("  Grid Minimum = %g\n", grid_range.Min());
    printf("  Grid Maximum = %g\n", grid_range.Max());
    printf("  Filter Resolution = %d %d %d\n", filter.XResolution(), filter.YResolution(), filter.ZResolution());
    printf("  Filter Minimum = %g\n", filter_range.Min());
    printf("  Filter Maximum = %g\n", filter_range.Max());
    fflush(stdout);
  }

  // Return grid
  return grid;
}

#else

R3Grid *
CreateDistanceGrid(PDBFile *file)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return NULL;
  PDBModel *model = file->Model(0);

  // Determine minimum and maximum distance from atom
  RNScalar min_distance = solvent_radius; // + 1.8;
  RNScalar max_distance = 10;

  // Determine grid resolution
  const R3Box& bbox = model->BBox();
  int grid_xresolution = 4 * bbox.XLength() / min_distance;
  int grid_yresolution = 4 * bbox.YLength() / min_distance;
  int grid_zresolution = 4 * bbox.ZLength() / min_distance;
  if (grid_xresolution > grid_resolution[0]) grid_xresolution = grid_resolution[0];
  if (grid_yresolution > grid_resolution[1]) grid_yresolution = grid_resolution[1];
  if (grid_zresolution > grid_resolution[2]) grid_zresolution = grid_resolution[2];

  // Compute squared distance transform
  R3Grid dt(grid_xresolution, grid_yresolution, grid_zresolution);
  dt.SetWorldToGridTransformation(model->BBox());
  for (int a = 0; a < model->NAtoms(); a++) {
    PDBAtom *atom = model->Atom(a);
    if (atom->IsHetAtom()) continue;
    dt.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
  }
  dt.SquaredDistanceTransform();
  dt.Multiply(dt.GridToWorldScaleFactor() * dt.GridToWorldScaleFactor());

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_xresolution, grid_yresolution, grid_zresolution);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize sphere at midpoint between every pair of atoms
  int pair_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom1 = model->Atom(i);
    if (atom1->IsHetAtom()) continue;
    if (atom1->accessible_surface_area == 0) continue;
    RNScalar radius1 = atom1->Radius();
    for (int j = i+1; j < model->NAtoms(); j++) {
      PDBAtom *atom2 = model->Atom(j);
      if (atom2->IsHetAtom()) continue;
      if (atom2->accessible_surface_area == 0) continue;
      RNScalar radius2 = atom2->Radius();
      R3Point center = 0.5 * (atom1->position + atom2->position);
      RNLength span_length = R3Distance(atom1->position, atom2->position) - radius1 - radius2;
      if (span_length  > 8) continue;
      RNScalar distance_squared = dt.WorldValue(center);
      if (distance_squared < 1.5 * 1.5) continue;
      if (distance_squared > 4 * 4) distance_squared = 4 * 4;
      grid->RasterizeWorldSphere(center, sqrt(distance_squared), 1.0);
      pair_count++;
    }
  }

  // Print statistics
  if (print_verbose) {
    RNInterval grid_range = grid->Range();
    int ncomponents = grid->ConnectedComponents();
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Resolution = %d %d %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Spacing = %g\n", grid->GridToWorldScaleFactor());
    printf("  Volume = %g\n", grid->Volume());
    printf("  Minimum = %g\n", grid_range.Min());
    printf("  Maximum = %g\n", grid_range.Max());
    printf("  Cardinality = %d\n", grid->Cardinality());
    printf("  # Connected Components = %d\n", ncomponents);
    printf("  # Pairs = %d\n", pair_count);
    fflush(stdout);
  }

  // Return grid
  return grid;
}

#endif



////////////////////////////////////////////////////////////////////////
// LIGAND ANALYSIS STUFF
////////////////////////////////////////////////////////////////////////

static int
AnalyzeLigand(R3Grid *grid, RNScalar *& ligand_values, int& ligand_nvalues, 
              PDBFile *file, PDBResidue *ligand, RNBoolean check_ligand_volume)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Initialize output variables
  ligand_nvalues = 0;
  ligand_values = NULL;

  // Check everything
  if (!grid || !file || !ligand) return 0;

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Make array of grid values at positions inside ligand
  if (check_ligand_volume) {
    // Get PDB model from file
    if (file->NModels() < 1) return NULL;
    PDBModel *model = file->Model(0);

    // Create grid with ligand rasterized
    R3Grid ligand_grid(grid->XResolution(), grid->YResolution(), grid->ZResolution());
    ligand_grid.SetWorldToGridTransformation(model->BBox());
    for (int i = 0; i < ligand->NAtoms(); i++) {
      PDBAtom *atom = ligand->Atom(i);
      ligand_grid.RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
    }

    // Determine the number of values to check in ligand
    ligand_nvalues = ligand_grid.Cardinality();

    // Allocate array of values found at positions of ligand atoms
    ligand_values = new RNScalar [ ligand_nvalues ];
    assert(ligand_values);

    // Fill array of ligand values 
    int count = 0;
    for (int i = 0; i < grid->NEntries(); i++) {
      if (ligand_grid.GridValue(i) == 0) continue;
      assert(count < ligand_nvalues);
      ligand_values[count++] = grid->GridValue(i);
    }
  }
  else {
    // Determine number of ligand values
    ligand_nvalues = ligand->NAtoms();

    // Allocate array of values found at positions of ligand atoms
    ligand_values = new RNScalar [ ligand_nvalues ];
    assert(ligand_values);

    // Fill array of values found at positions of ligand atoms
    for (int i = 0; i < ligand->NAtoms(); i++){
      PDBAtom *atom = ligand->Atom(i);
      const R3Point& position = atom->Position();
      ligand_values[i] = grid->WorldValue(position);
    }
  }

  // Just checking
  if (ligand_nvalues == 0) return 0;

  // Sort the values
  qsort(ligand_values, ligand_nvalues, sizeof(RNScalar), RNCompareScalars);

  // Set grid_threshold
  int index = 0.75 * ligand_nvalues;
  grid_threshold = ligand_values[index];

  // Print statistics
  if (print_verbose) {
    printf("Sorted ligand values ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", ligand_nvalues);
    printf("  1st percentile value = %g\n", ligand_values[ligand_nvalues-1]);
    printf("  25th percentile value = %g\n", ligand_values[(int) (0.75 * ligand_nvalues)]);
    printf("  50th percentile value = %g\n", ligand_values[(int) (0.50 * ligand_nvalues)]);
    printf("  75th percentile value = %g\n", ligand_values[(int) (0.25 * ligand_nvalues)]);
    printf("  100th percentile value = %g\n", ligand_values[0]);
    printf("  Total Ligand Volume = %g\n", ligand_nvalues * volume_scale_factor);
    fflush(stdout);
  }

  // Return the number of values
  return ligand_nvalues;
}



static int
AnalyzeGrid(R3Grid *grid, RNScalar *& grid_values, int& grid_nvalues)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Initialize output variables
  grid_nvalues = 0;
  grid_values = NULL;

  // Check everything
  if (!grid) return 0;

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Allocate array of grid values
  grid_nvalues = grid->NEntries();
  grid_values = new RNScalar [ grid_nvalues ];
  assert(grid_values);

  // Make array of grid values (copy of grid)
  RNScalar min_grid_value = FLT_MAX;
  RNScalar max_grid_value = -FLT_MAX;
  for (int i = 0; i < grid_nvalues; i++) {
    RNScalar value = grid->GridValue(i);
    if (value < min_grid_value) min_grid_value = value;
    if (value > max_grid_value) max_grid_value = value;
    grid_values[i] = value;
  }

  // Sort the array
  qsort(grid_values, grid_nvalues, sizeof(RNScalar), RNCompareScalars);

  // Print statistics
  if (print_verbose) {
    printf("Sorted grid values ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", grid_nvalues);
    printf("  Minimum Value = %g\n", min_grid_value);
    printf("  Maximum Value = %g\n", max_grid_value);
    printf("  Grid volume = %g\n", grid_nvalues * volume_scale_factor);
    fflush(stdout);
  }

  // Return the number of values
  return grid_nvalues;
}



static int 
PrintPlot(R3Grid *grid, RNScalar *grid_values, int grid_nvalues, RNScalar *ligand_values, int ligand_nvalues, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdout;
  if (filename) {
    fp = fopen(filename, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open site plot file: %s\n", filename);
      return 0;
    }
  }

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Print plot
  fprintf(fp, "0 0\n");
  int grid_index = grid_nvalues-1;
  for (int ligand_index = ligand_nvalues-1; ligand_index >= 0; ligand_index--) {
    // Check grid index
    if (grid_index <= 0) break;

    // Check ligand value
    if (ligand_values[ligand_index] == 0) break;

    // Find next grid value less or equal to ligand value
    for (; grid_index >= 0; grid_index--) {
      if (RNIsLess(grid_values[grid_index], ligand_values[ligand_index], 1.0E-8)) break;
    }

    // Determine fraction of ligand values considered
    RNScalar ligand_fraction = (RNScalar) (ligand_nvalues - ligand_index) / (RNScalar) ligand_nvalues;

    // Determine number of grid cells with value greater of equal to ligand value
    int grid_count = grid_nvalues - grid_index;

    // Convert number of grid cells to a volume
    RNScalar grid_volume = grid_count * volume_scale_factor;

    // Print plot numbers
    fprintf(fp, "%g %g\n", ligand_fraction, grid_volume);
  }

  // Close file
  if (filename) fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote plot ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", ligand_nvalues);
    fflush(stdout);
  }

  // Return success
  return 1;
}

////////////////////////////////////////////////////////////////////////

R3Grid *
CreateGrid(PDBFile *file)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize each atom into grid
  int atom_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (!atom->AminoAcid()) continue;
    grid->RasterizeWorldSphere(atom->Position(), atom->Radius() + atom_radius, 1.0);
    atom_count++;
  }

  // Assign -1 everywhere inside an atom, and 0 everywhere else
  grid->Threshold(0.5, 0, -1);

  // Rasterize each atom into grid
  R3Point lo = grid->WorldPosition(R3Point(0, 0, 0));
  R3Point hi = grid->WorldPosition(R3Point(grid->XResolution()-1, grid->YResolution()-1, grid->ZResolution()-1));
  RNScalar xstep = (hi.X() - lo.X()) / (grid->XResolution()-1);
  RNScalar ystep = (hi.Y() - lo.Y()) / (grid->YResolution()-1);
  RNScalar zstep = (hi.Z() - lo.Z()) / (grid->ZResolution()-1);
  for (int a = 0; a < model->NAtoms(); a++) {
    PDBAtom *atom = model->Atom(a);
    if (!atom->AminoAcid()) continue;
    RNScalar conservation =  1 - atom->tempFactor;
    if (conservation < 0.001) conservation = 0.001;
    if (atom->tempFactor == -999) conservation = 0.001;
    const R3Point& atom_position = atom->Position();
    for (int k = 0; k < grid->ZResolution(); k++) {
      RNScalar z = lo.Z() + k * zstep;
      RNScalar dz = z - atom_position.Z();
      RNScalar dz_squared = dz*dz;
      for (int j = 0; j < grid->YResolution(); j++) {
        RNScalar y = lo.Y() + j * ystep;
        RNScalar dy = y - atom_position.Y();
        RNScalar dy_squared = dy*dy;
        for (int i = 0; i < grid->XResolution(); i++) {
          if (grid->GridValue(i, j, k) < 0) continue;
          RNScalar x = lo.X() + i * xstep;
          RNScalar dx = x - atom_position.X();
          RNScalar dx_squared = dx*dx;
          RNScalar distance_squared = dx_squared + dy_squared + dz_squared;
          RNScalar distance = sqrt(distance_squared);
          if (RNIsPositive(distance)) grid->AddGridValue(i, j, k, conservation / distance);
          else grid->SetGridValue(i, j, k, 0.0); 
        }
      }
    }
  }

  // Assign 0 everywhere inside an atom
  grid->Threshold(-0.5, 0, R3_GRID_KEEP_VALUE);

  // Print statistics
  if (print_verbose) {
    RNInterval range = grid->Range();
    printf("  Created conservation grid ...\n");
    printf("    Time = %.2f seconds\n", start_time.Elapsed());
    printf("    # Atoms = %d\n", atom_count);
    printf("    Minimum = %g\n", range.Min());
    printf("    Maximum = %g\n", range.Max());
    printf("    L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}

////////////////////////////////////////////////////////////////////////

static int 
CreateGrids(PDBFile *file)
{
  // Check the file
  if (file->NModels() == 0) {
    fprintf(stderr, "PDB file %s has no models\n", file->Name());
    return 0;
  }

  // Get the first model
  PDBModel *model = file->Model(0);

  // Create grids
  ngrids = 2;
  grids = new R3Grid *[ngrids];
  grids[1] = CreateRasterGrid(model);
  grids[0] = CreateConservationGrid(model, grids[1]);

  // Set slice coords to middle of grid
  grid_slice_coords[RN_X] = grid_resolution[0]/2;
  grid_slice_coords[RN_Y] = grid_resolution[1]/2;
  grid_slice_coords[RN_Z] = grid_resolution[2]/2;

  // Return number of grids
  return ngrids;
}



R3Grid *
CreateRasterGrid(PDBModel *model)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize each atom into grid
  int atom_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (!atom->AminoAcid()) continue;
    grid->RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
    atom_count++;
  }

  // Take distance transform
  grid->Threshold(0.5, 0, 1);

  // Print statistics
  if (print_verbose) {
    printf("Created raster grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateDistanceTransformGrid(PDBModel *model)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize each atom into grid
  int atom_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (!atom->AminoAcid()) continue;
    grid->RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
    atom_count++;
  }

  // Take distance transform
  RNScalar s = grid->WorldToGridScaleFactor();
  grid->SquaredDistanceTransform();
  grid->Sqrt();
  grid->Divide(s);

  // Print statistics
  if (print_verbose) {
    printf("Created distance transform grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateConservedDistanceTransformGrid(PDBModel *model)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize each atom into grid
  int atom_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (!atom->AminoAcid()) continue;
    RNScalar c = atom->TempFactor();
    if (RNIsNotEqual(c, 8) && RNIsNotEqual(c, 9)) continue;
    grid->RasterizeWorldSphere(atom->Position(), atom->Radius(), 1.0);
    atom_count++;
  }

  // Take distance transform
  RNScalar s = grid->WorldToGridScaleFactor();
  grid->SquaredDistanceTransform();
  grid->Sqrt();
  grid->Divide(s);

  // Print statistics
  if (print_verbose) {
    printf("Created conserved distance transform grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateSPHDistanceGrid(PDBModel *model)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize each surfnet sphere into grid
  int atom_count = 0;
  if (surfnet_chain) {
    for (int i = 0; i < surfnet_chain->NAtoms(); i++) {
      PDBAtom *atom = surfnet_chain->Atom(i);
      grid->RasterizeWorldPoint(atom->Position(), 1.0);
      atom_count++;
    }

    // Take distance transform
    RNScalar s = grid->WorldToGridScaleFactor();
    grid->SquaredDistanceTransform();
    grid->Sqrt();
    grid->Divide(s);
  }

  // Print statistics
  if (print_verbose) {
    printf("Created SPH distance grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateCombinedGrid(PDBModel *model, R3Grid *grids[], int ngrids)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*(grids[0]));
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Combine grids
  RNScalar *grid_valuep = (RNScalar *) grid->GridValues();
  RNScalar *grid0_valuep = (RNScalar *) grids[0]->GridValues();
  RNScalar *grid1_valuep = (RNScalar *) grids[1]->GridValues();
  RNScalar *grid2_valuep = (RNScalar *) grids[2]->GridValues();
  for (int i = 0; i < grid->XResolution() * grid->YResolution() * grid->ZResolution(); i++) {
    if (*grid0_valuep == 0.0) *grid_valuep = 10.0;
    else *grid_valuep = (*grid1_valuep + *grid2_valuep);
    grid_valuep++;
    grid0_valuep++;
    grid1_valuep++;
    grid2_valuep++;
  }

  // Print statistics
  if (print_verbose) {
    const R3Box& b = model->BBox();
    printf("Created combined grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  L2Norm = %g\n", grid->L2Norm());
    printf("  Grid dimensions = %d x %d %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    printf("  World bounding box = ( %g %g %g ) ( %g %g %g )\n", b[0][0], b[0][1], b[0][2], b[1][0], b[1][1], b[1][2]);
    printf("  World resolution = %g\n", 1 / grid->WorldToGridScaleFactor());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateFabianGrid(PDBModel *model, R3Grid *dt)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize each atom into grid
  int atom_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom = model->Atom(i);
    if (!atom->AminoAcid()) continue;
    RNScalar distance = dt->WorldValue(atom->Position());
    if (distance > 3.5) continue;
    grid->RasterizeWorldSphere(atom->Position(), atom->Occupancy(), 1.0);
    atom_count++;
  }

  // Threshold grid (to mark anywhere within a sphere equally)
  grid->Threshold(0.5, 0.0, 1.0);

  // Print statistics
  if (print_verbose) {
    printf("Created Fabian grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateAbdullahGrid(PDBModel *model)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Just checking
  int atom_count = 0;
  if (surfnet_chain && ligand) {
    for (int i = 0; i < surfnet_chain->NAtoms(); i++) {
      PDBAtom *atom = surfnet_chain->Atom(i);
      for (int j = 0; j < ligand->NAtoms(); j++) {
        PDBAtom *ligand_atom = ligand->Atom(j);
        RNLength distance = R3Distance(atom->Position(), ligand_atom->Position());
        if (distance < ligand_atom->Radius() + atom->Occupancy() + 0.1) {
          grid->RasterizeWorldSphere(atom->Position(), atom->Occupancy(), 1.0);
          atom_count++;
        }
      }
    }

    // Threshold grid (to mark anywhere within a sphere equally)
    grid->Threshold(0.5, 0.0, 1.0);
  }

  // Print statistics
  if (print_verbose) {
    printf("Created Fabian grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateProductGrid(PDBModel *model, R3Grid *grids[], int ngrids)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*(grids[0]));
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Intersect other grids
  for (int i = 1; i < ngrids; i++) {
    grid->Multiply(*(grids[i]));
  }

  // Print statistics
  if (print_verbose) {
    const R3Box& b = model->BBox();
    printf("Created combined grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  L2Norm = %g\n", grid->L2Norm());
    printf("  Grid dimensions = %d x %d x %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    printf("  World bounding box = ( %g %g %g ) ( %g %g %g )\n", b[0][0], b[0][1], b[0][2], b[1][0], b[1][1], b[1][2]);
    printf("  World resolution = %g\n", 1 / grid->WorldToGridScaleFactor());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateHydrophobicityGrid(PDBModel *model)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize each atom into grid
  int atom_count = 0;
  R3Point lo = grid->WorldPosition(R3Point(0, 0, 0));
  R3Point hi = grid->WorldPosition(R3Point(grid->XResolution()-1, grid->YResolution()-1, grid->ZResolution()-1));
  RNScalar xstep = (hi.X() - lo.X()) / (grid->XResolution()-1);
  RNScalar ystep = (hi.Y() - lo.Y()) / (grid->YResolution()-1);
  RNScalar zstep = (hi.Z() - lo.Z()) / (grid->ZResolution()-1);
  for (int a = 0; a < model->NAtoms(); a++) {
    PDBAtom *atom = model->Atom(a);
    PDBAminoAcid *aminoacid = atom->AminoAcid();
    if (!aminoacid) continue;
    const R3Point& atom_position = atom->Position();
    for (int k = 0; k < grid->ZResolution(); k++) {
      RNScalar z = lo.Z() + k * zstep;
      RNScalar dz = z - atom_position.Z();
      RNScalar dz_squared = dz*dz;
      for (int j = 0; j < grid->YResolution(); j++) {
        RNScalar y = lo.Y() + j * ystep;
        RNScalar dy = y - atom_position.Y();
        RNScalar dy_squared = dy*dy;
        for (int i = 0; i < grid->XResolution(); i++) {
          RNScalar x = lo.X() + i * xstep;
          RNScalar dx = x - atom_position.X();
          RNScalar dx_squared = dx*dx;
          RNScalar distance_squared = dx_squared + dy_squared + dz_squared;
          grid->AddGridValue(i, j, k, aminoacid->Hydrophobicity() / distance_squared);
        }
      }
    }
    atom_count++;
  }

  // Print statistics
  if (print_verbose) {
    printf("Created hydrophobicity grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateMaskedGrid(PDBModel *model, R3Grid *src, R3Grid *mask)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(*src);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Mask grid
  grid->Mask(*mask);

  // Print statistics
  if (print_verbose) {
    const R3Box& b = model->BBox();
    printf("Created masked grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateSurfnetGrid(PDBModel *model, R3Grid *dt)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Rasterize sphere at midpoint between every pair of atoms
  int pair_count = 0;
  for (int i = 0; i < model->NAtoms(); i++) {
    PDBAtom *atom1 = model->Atom(i);
    if (!atom1->AminoAcid()) continue;
    RNScalar radius1 = atom1->Radius();
    for (int j = i+1; j < model->NAtoms(); j++) {
      PDBAtom *atom2 = model->Atom(j);
      if (!atom2->AminoAcid()) continue;
      RNScalar radius2 = atom2->Radius();
      R3Point center = 0.5 * (atom1->position + atom2->position);
      RNScalar distance = dt->WorldValue(center);
      RNLength span_length = R3Distance(atom1->position, atom2->position) - radius1 - radius2;
      if (span_length  > 8) continue;
      if (distance < 1.5) continue;
      if (distance > 4) distance = 4;
      grid->RasterizeWorldSphere(center, distance, 1.0);
      pair_count++;
    }

    // Print status message every ten seconds
    static RNScalar next_time = 10;
    if (start_time.Elapsed() > next_time) {
      printf("  Phase 1: %d %d%c", i, pair_count, 0xd);
      next_time += 10;
    }
  }

  // Threshold grid (to mark anywhere within any sphere equally)
  grid->Threshold(0.5, 0.0, 1.0);

  // Print statistics
  if (print_verbose) {
    printf("Created first surfnet grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  L2Norm = %g\n", grid->L2Norm());
    printf("  # Pairs = %d\n", pair_count);
    fflush(stdout);
  }

  // Return grid
  return grid;
}



R3Grid *
CreateDifferenceGrid(PDBModel *pdb, R3Grid *grid1, R3Grid *grid2) {
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Create a copy of input grid
  R3Grid *grid = new R3Grid(*grid1);
  if (!grid1) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Subtract grid2
  grid->Subtract(*grid2);
  grid->Abs();

  // Print statistics
  if (print_verbose) {
    printf("Created difference grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Maximum = %g\n", grid->Maximum());
    printf("  L2Norm = %g\n", grid->L2Norm());
    printf("  Grid dimensions = %d x %d x %d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  L1Distance: %g\n", grid1->L1Distance(*grid2));
    printf("  L2Distance: %g\n", grid1->L2Distance(*grid2));
    fflush(stdout);
  }

  // Return grid
  return grid;
}



