// Source file for the pdb viewer program



// Include files 

#include "R3Graphics/R3Graphics.h"
#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"
#include <GL/glut.h>
#include "glui.h"



// GLUI commands

enum {
  NULL_COMMAND,
  QUIT_COMMAND
};

typedef enum {
  FILE_COLOR_SCHEME,
  MODEL_COLOR_SCHEME,
  CHAIN_COLOR_SCHEME,
  RESIDUE_COLOR_SCHEME,
  ATOM_COLOR_SCHEME
} ColorScheme;



// Program variables

static char *pdb_name = NULL;
static char *surfnet_name = NULL;
static char *asa_name = NULL;
static char *conservation_name = NULL;
static char *ligand_string = NULL;
static char *image_name = NULL;
static char *grid_name = NULL;
static char *plot_name = NULL;
static RNBoolean hydrophobicity = FALSE;
static RNScalar background_color[3] = { 1, 1, 1 };
static int grid_resolution[3] = { 64, 64, 64 };
static R3Point initial_camera_origin = R3Point(0.0, 0.0, 0.0);
static R3Vector initial_camera_towards = R3Vector(0.0, 0.0, -1.0);
static R3Vector initial_camera_up = R3Vector(0.0, 1.0, 0.0);
static RNBoolean initial_camera = FALSE;
static RNLength solvent_radius = 1.4;
static int print_verbose = 0;
static int batch = 0;



// GLUI variables 

static GLUI *glui = NULL;
static int GLUIwindow = 0;
static int GLUIwindow_height = 1024;
static int GLUIwindow_width = 1024;
static int GLUImouse[2] = { 0, 0 };
static int GLUIbutton[3] = { 0, 0, 0 };
static int GLUIarrows[4] = { 0, 0, 0, 0 };
static int GLUImodifiers = 0;



// Application variables

static PDBFile *file = NULL;
static PDBFile *surfnet_file = NULL;
static PDBResidue *ligand = NULL;
static R3Viewer *viewer = NULL;
static R3Grid *grid = NULL;



// Display variables

static int show_grid = 1;
static int show_ligand = 1;
static int show_file = 1;
static int file_color_scheme = RESIDUE_COLOR_SCHEME;
static int ligand_color_scheme = RESIDUE_COLOR_SCHEME;
static int show_occupancy = 0;
static int show_conservation = 0;
static int show_accessible_surface_area = 0;
static int show_surfnet = 0;
static int show_grid_points = 0;
static int show_grid_text = 0;
static int show_grid_isosurface = 1;
static int show_grid_slices[3] = { 0, 0, 0 };
static int show_grid_box = 0;
static float grid_threshold = 1E-6;
static int grid_slice_coords[3] = { 0, 0, 0 };



void GLUIStop(void)
{
  // Destroy window 
  glutDestroyWindow(GLUIwindow);

  // Exit
  exit(0);
}



void GLUIRedraw(void)
{
  // Set viewing transformation
  viewer->Camera().Load();

  // Clear window 
  glClearColor(background_color[0], background_color[1], background_color[2], 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set lights
  static GLfloat light0_position[] = { 3.0, 4.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
  static GLfloat light1_position[] = { -3.0, -2.0, -3.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_POSITION, light1_position);

  // Show atoms
  if (file && show_file) {
    glEnable(GL_LIGHTING);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    RNRgb color(0.5, 0.25, 0.5);
    for (int m = 0; m < file->NModels(); m++) {
      PDBModel *model = file->Model(m);
      if (file_color_scheme == MODEL_COLOR_SCHEME) color = model->Color();
      for (int i = 0; i < model->NChains(); i++) {
        PDBChain *chain = model->Chain(i);
        if (file_color_scheme == CHAIN_COLOR_SCHEME) color = chain->Color();
        for (int j = 0; j < chain->NResidues(); j++) {
          PDBResidue *residue = chain->Residue(j);
          if (residue == ligand) continue;
          if (file_color_scheme == RESIDUE_COLOR_SCHEME) color = residue->Color();
          for (int k = 0; k < residue->NAtoms(); k++) {
            PDBAtom *atom = residue->Atom(k);
            if (show_conservation) {
              if (atom->conservation == -999) color = RNblue_rgb;
              else { color[0] = atom->conservation; color[1] = atom->conservation; color[2] = atom->conservation; }
            }
            else if (show_accessible_surface_area) {
              if (atom->accessible_surface_area == -999) color = RNblue_rgb;
              else { RNScalar asa = atom->accessible_surface_area / 50; color[0] = asa; color[1] = asa; color[2] = asa; }
            }
            else if (file_color_scheme == ATOM_COLOR_SCHEME) color = atom->Color();
            if (show_occupancy && (atom->Occupancy() < 1.0)) glEnable(GL_BLEND);
            static GLfloat material[4];
            material[0] = color[0]; material[1] = color[1]; material[2] = color[2]; material[3] = atom->Occupancy();
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
            R3Sphere(atom->Position(), atom->Radius()).Draw();
            if (show_occupancy && (atom->Occupancy() < 1.0)) glDisable(GL_BLEND);
          }
        }
      }
    }
  }
    
  // Show ligand
  if (ligand && show_ligand) {
    PDBChain *chain = ligand->Chain();
    glEnable(GL_LIGHTING);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    static RNRgb default_color(0, 1, 0);
    RNRgb color = default_color;
    if (ligand_color_scheme == CHAIN_COLOR_SCHEME) color = chain->Color();
    else if (ligand_color_scheme == RESIDUE_COLOR_SCHEME) color = ligand->Color();
    for (int i = 0; i < ligand->NAtoms(); i++) {
      PDBAtom *atom = ligand->Atom(i);
      if (ligand_color_scheme == ATOM_COLOR_SCHEME) color = atom->Color();
      if (atom->Occupancy() < 1.0) glEnable(GL_BLEND);
      static GLfloat material[4];
      material[0] = color[0]; material[1] = color[1]; material[2] = color[2]; material[3] = atom->Occupancy();
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
      R3Sphere(atom->Position(), atom->Radius()).Draw();
      if (atom->Occupancy() < 1.0) glDisable(GL_BLEND);
    }
#if 0
    glDisable(GL_LIGHTING);
    glColor3f(1, 1, 1);
    glLineWidth(5.0);
    for (int i = 0; i < ligand->NAtoms(); i++) {
      PDBAtom *atom = ligand->Atom(i);
      R3Sphere(atom->Position(), atom->Radius()).Outline();
    }
    glLineWidth(1.0);
#endif
  }
    
  // Show sites
  if (surfnet_file && show_surfnet) {
    glEnable(GL_LIGHTING);
    static GLfloat material[] = { 0, 0.5, 0, 1 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
    for (int i = 0; i < surfnet_file->NModels(); i++) {
      PDBModel *surfnet_model = surfnet_file->Model(i);
      for (int j = 0; j < surfnet_model->NAtoms(); j++) {
        PDBAtom *atom = surfnet_model->Atom(j);
        R3Sphere(atom->Position(), atom->Occupancy()).Draw();
      }
    }
  }

  // Draw grid
  if (grid && show_grid) {
    // Push grid transformation
    const R3Affine& grid_to_world = grid->GridToWorldTransformation();
    grid_to_world.Push();

    // Draw grid
    if (show_grid_points) {
      glPointSize(5.0);
      glDisable(GL_LIGHTING);
      glDepthMask(FALSE);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glEnable(GL_BLEND);
      glBegin(GL_POINTS);
      RNScalar maximum = grid->Maximum();
      for (int i = 0; i < grid->XResolution(); i++) {
        for (int j = 0; j < grid->YResolution(); j++) {
          for (int k = 0; k < grid->ZResolution(); k++) {
            float value = grid->GridValue(i, j, k);
            if (value <= grid_threshold) continue;
            RNLoadRgb(value * RNwhite_rgb / maximum);
            glVertex3i(i, j, k);
          }
        }
      }
      glEnd();
      glDisable(GL_BLEND);
      glBlendFunc(GL_ONE, GL_ZERO);
      glDepthMask(TRUE);
      glPointSize(1.0);
    }

    // Draw grid text
    if (show_grid_text) {
      char buffer[64];
      glDisable(GL_LIGHTING);
      RNLoadRgb(1 - background_color[0], 1 - background_color[1], 1 - background_color[2]);
      for (int i = 0; i < grid->XResolution(); i++) {
        for (int j = 0; j < grid->YResolution(); j++) {
          for (int k = 0; k < grid->ZResolution(); k++) {
            float value = grid->GridValue(i, j, k);
            if (value <= grid_threshold) continue;
            sprintf(buffer, "%.2g", value);
            R3DrawText((float) i, (float) j, (float) k, buffer);
          }
        }
      }
    }

    // Draw grid isosurface
    if (show_grid_isosurface) {
      glDisable(GL_LIGHTING);
      RNLoadRgb(RNred_rgb);
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      grid->DrawIsoSurface(grid_threshold);
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    // Draw grid X slices
    if (show_grid_slices[RN_X] || show_grid_slices[RN_Y] || show_grid_slices[RN_Z]) {
      glDisable(GL_LIGHTING);
      RNLoadRgb(RNwhite_rgb);
      if (show_grid_slices[RN_X]) grid->DrawSlice(RN_X, grid_slice_coords[RN_X]);
      if (show_grid_slices[RN_Y]) grid->DrawSlice(RN_Y, grid_slice_coords[RN_Y]);
      if (show_grid_slices[RN_Z]) grid->DrawSlice(RN_Z, grid_slice_coords[RN_Z]);
    }

    // Draw grid bounding box
    if (show_grid_box) {
      glDisable(GL_LIGHTING);
      glColor3f(0, 0, 0);
      R3Box(0, 0, 0, grid->XResolution()-1, grid->YResolution()-1, grid->ZResolution()-1).Outline();
      if (show_grid_slices[0]) R3Box(grid_slice_coords[0], 0, 0, grid_slice_coords[0], grid->YResolution()-1, grid->ZResolution()-1).Outline();
      if (show_grid_slices[1]) R3Box(0, grid_slice_coords[1], 0, grid->XResolution()-1, grid_slice_coords[1], grid->ZResolution()-1).Outline();
      if (show_grid_slices[2]) R3Box(0, 0, grid_slice_coords[2], grid->XResolution()-1, grid->YResolution()-1, grid_slice_coords[2]).Outline();
    }

    // Pop grid transformation
    grid_to_world.Pop();
  }

  // Capture image and exit
  if (image_name) {
    R2Image image(GLUIwindow_width, GLUIwindow_height, 3);
    image.Capture();
    image.Write(image_name);
    image_name = NULL;
    if (batch) GLUIStop();
  }

  // Swap buffers 
  glutSwapBuffers();
}    



void GLUIResize(int a, int b)
{
  // Resize window
  int x, y, w, h;
  GLUI_Master.get_viewport_area( &x, &y, &w, &h );
  glViewport( x, y, w, h );

  // Resize viewer viewport
  viewer->ResizeViewport(x, y, w, h);

  // Remember window size 
  GLUIwindow_width = w;
  GLUIwindow_height = h;

  // Redraw
  glutPostRedisplay();
}



void GLUIMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Compute mouse movement
  int dx = x - GLUImouse[0];
  int dy = y - GLUImouse[1];
  
  // World in hand navigation 
  R3Point origin = file->BBox().Centroid();
  if (GLUIbutton[0]) viewer->RotateWorld(1.0, origin, x, y, dx, dy);
  else if (GLUIbutton[1]) viewer->ScaleWorld(1.0, origin, x, y, dx, dy);
  else if (GLUIbutton[2]) viewer->TranslateWorld(1.0, origin, x, y, dx, dy);
  if (GLUIbutton[0] || GLUIbutton[1] || GLUIbutton[2]) glutPostRedisplay();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;
}



void GLUIMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;
  
  // Process mouse button event

  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUIbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Redraw
  glutPostRedisplay();
}



void GLUISpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Process keyboard button event 
  switch (key) {
  case GLUT_KEY_DOWN:
    grid_threshold *= 0.9;
    if (grid_threshold < 0.001) grid_threshold = 0.0;
    break;

  case GLUT_KEY_UP:
    grid_threshold *= 1.1;
    if (grid_threshold < 0.001) grid_threshold = 0.001;
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();
}



void GLUIKeyboard(unsigned char key, int x, int y)
{
  // Process keyboard button event 
  switch (key) {
  case '!':
    show_conservation = !show_conservation;
    break;

  case '@':
    show_accessible_surface_area = !show_accessible_surface_area;
    break;

  case 'A':
    ligand_color_scheme = ATOM_COLOR_SCHEME;
    break;

  case 'a':
    file_color_scheme = ATOM_COLOR_SCHEME;
    break;

  case 'B':
  case 'b':
    show_grid_box = !show_grid_box;
    break;

  case 'C':
    ligand_color_scheme = CHAIN_COLOR_SCHEME;
    break;

  case 'c':
    file_color_scheme = CHAIN_COLOR_SCHEME;
    break;

  case 'F':
  case 'f':
    show_file = !show_file;
    break;

  case 'G':
  case 'g':
    show_grid_points = !show_grid_points;
    break;

  case 'I':
  case 'i':
    show_grid_isosurface = !show_grid_isosurface;
    break;

  case 'L':
  case 'l':
    show_ligand = !show_ligand;
    break;

  case 'M':
    ligand_color_scheme = MODEL_COLOR_SCHEME;
    break;

  case 'm':
    file_color_scheme = MODEL_COLOR_SCHEME;
    break;

  case 'O':
  case 'o':
    show_occupancy = !show_occupancy;
    break;

  case 'R':
    ligand_color_scheme = RESIDUE_COLOR_SCHEME;
    break;

  case 'r':
    file_color_scheme = RESIDUE_COLOR_SCHEME;
    break;

  case 'S':
  case 's':
    show_surfnet = !show_surfnet;
    break;

  case 'T':
  case 't':
    show_grid_text = !show_grid_text;
    break;

  case 'W':
  case 'w':
    show_grid_slices[0] = !show_grid_slices[0];
    show_grid_slices[1] = show_grid_slices[0];
    show_grid_slices[2] = show_grid_slices[0];
    break;

  case 'X':
  case 'x':
    show_grid_slices[0] = !show_grid_slices[0];
    break;

  case 'Y':
  case 'y':
    show_grid_slices[1] = !show_grid_slices[1];
    break;

  case 'Z':
  case 'z':
    show_grid_slices[2] = !show_grid_slices[2];
    break;

  case ' ': {
    // Print camera
    const R3Camera& camera = viewer->Camera();
    printf("#camera  %g %g %g  %g %g %g  %g %g %g  %g \n",
           camera.Origin().X(), camera.Origin().Y(), camera.Origin().Z(),
           camera.Towards().X(), camera.Towards().Y(), camera.Towards().Z(),
           camera.Up().X(), camera.Up().Y(), camera.Up().Z(),
           camera.YFOV());
    break; }
      
  case 27: // ESCAPE
    GLUIStop();
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = GLUIwindow_height - y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();  
}




void GLUICallback(int control)
{
  switch (control) {
  case QUIT_COMMAND:
    GLUIStop();
    break;
  }
}



#if 0

void GLUIIdle(void)
{
  // Set current window
  if ( glutGetWindow() != GLUIwindow ) 
    glutSetWindow(GLUIwindow);  

  // Redraw
  glutPostRedisplay();
}

#endif



void GLUISetup(void)
{
  GLUI_Panel *panel;
  GLUI_Spinner *spinner;

  // Set the main window's title
  glutSetWindowTitle(pdb_name);

  // Initialize GLUI
  glui = GLUI_Master.create_glui("MBE Controls", 0, GLUIwindow_width + 120, 100);
  glui->set_main_gfx_window(GLUIwindow);

  // Add protein display variables
  panel = glui->add_panel("Protein Color Scheme");
  GLUI_RadioGroup *group = glui->add_radiogroup_to_panel(panel, &file_color_scheme );
  glui->add_radiobutton_to_group(group, "File");
  glui->add_radiobutton_to_group(group, "Model");
  glui->add_radiobutton_to_group(group, "Chain");
  glui->add_radiobutton_to_group(group, "Amino acid");
  glui->add_radiobutton_to_group(group, "Element");

  // Add ligand display variables
  if (ligand) {
    panel = glui->add_panel("Ligand Color Scheme");
    GLUI_RadioGroup *group = glui->add_radiogroup_to_panel(panel, &ligand_color_scheme );
    glui->add_radiobutton_to_group(group, "File");
    glui->add_radiobutton_to_group(group, "Model");
    glui->add_radiobutton_to_group(group, "Chain");
    glui->add_radiobutton_to_group(group, "Amino acid");
    glui->add_radiobutton_to_group(group, "Element");
  }

  panel = glui->add_panel("Binding Site Display");
  glui->add_checkbox_to_panel(panel, "Surfnet spheres", &show_surfnet );

  // Add grid display variables
  panel = glui->add_panel("Grid Display");
  glui->add_checkbox_to_panel(panel, "Points", &show_grid_points );
  glui->add_checkbox_to_panel(panel, "Labels", &show_grid_text );
  glui->add_checkbox_to_panel(panel, "Bounding box", &show_grid_text );

  // Add isosurface display variables
  panel = glui->add_panel("Isosurface Display");
  glui->add_checkbox_to_panel(panel, "Isosurface", &show_grid_isosurface );
  spinner = glui->add_spinner_to_panel(panel, "Threshold", GLUI_SPINNER_FLOAT, &grid_threshold);
  spinner->set_float_limits(0, grid->Maximum(), GLUI_LIMIT_CLAMP);
  spinner->set_speed(10);

  // Add slice display variables
  panel = glui->add_panel("Slice Display");
  glui->add_checkbox_to_panel(panel, "X Slice", &show_grid_slices[0] );
  spinner = glui->add_spinner_to_panel(panel, "X Slice", GLUI_SPINNER_INT, &grid_slice_coords[0]);
  spinner->set_int_limits(0, grid->XResolution()-1, GLUI_LIMIT_CLAMP);
  glui->add_checkbox_to_panel(panel, "Y Slice", &show_grid_slices[1] );
  spinner = glui->add_spinner_to_panel(panel, "Y Slice", GLUI_SPINNER_INT, &grid_slice_coords[1]);
  spinner->set_int_limits(0, grid->YResolution()-1, GLUI_LIMIT_CLAMP);
  glui->add_checkbox_to_panel(panel, "Z Slice", &show_grid_slices[2] );
  spinner = glui->add_spinner_to_panel(panel, "Z Slice", GLUI_SPINNER_INT, &grid_slice_coords[2]);
  spinner->set_int_limits(0, grid->ZResolution()-1, GLUI_LIMIT_CLAMP);

  // Add button
  glui->add_button("Quit", QUIT_COMMAND, GLUICallback);
}



void GLUIInit(int *argc, char **argv)
{
  // Open window 
  glutInit(argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUIwindow_width, GLUIwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUIwindow = glutCreateWindow("OpenGL Viewer");

  // Initialize background color 
  glClearColor(200.0/255.0, 200.0/255.0, 200.0/255.0, 1.0);

  // Initialize lights
  static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glEnable(GL_LIGHT0);
  static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);
  glEnable(GL_LIGHTING);

  // Initialize graphics modes 
  glEnable(GL_DEPTH_TEST);

  // Initialize GLUT callback functions 
  glutDisplayFunc(GLUIRedraw);
  GLUI_Master.set_glutReshapeFunc(GLUIResize);
  GLUI_Master.set_glutKeyboardFunc(GLUIKeyboard);
  GLUI_Master.set_glutSpecialFunc(GLUISpecial);
  GLUI_Master.set_glutMouseFunc(GLUIMouse);
  glutMotionFunc(GLUIMotion);

  // Initialize font
#if (RN_OS == RN_WINDOWSNT)
  int font = glGenLists(256);
  wglUseFontBitmaps(wglGetCurrentDC(), 0, 256, font); 
  glListBase(font);
#endif
}



void GLUIMainLoop(void)
{
  // Setup GLUI controls
  GLUISetup();

  // Run main loop -- never returns 
  glutMainLoop();
}



static PDBFile *
ReadPDB(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) {
    RNFail("Unable to read PDB file: %s", filename);
    return NULL;
  }

  // Find ligand
  if (ligand_string) {
    PDBStructureType foo =  file->FindAny(ligand_string);
    ligand = file->FindResidue(ligand_string);
    if (!ligand) {
      fprintf(stderr, "Unable to find ligand %s in %s\n", ligand_string, pdb_name); 
      return 0;
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Read PDB file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    for (int i =0; i < file->NModels(); i++) {
      PDBModel *model = file->Model(i);
      printf("  Model %s ...\n", model->Name());
      printf("  # Chains = %d\n", model->NChains());
      printf("  # Residues = %d\n", model->NResidues());
      printf("  # Atoms = %d\n", model->NAtoms());
    }
    fflush(stdout);
  }

  // Return success
  return file;
}



static int
ReadConservationFile(PDBFile *file, PDBModel *model, PDBChain *chain, char *filename)
{
  // Initialize all conservation scores
  for (int i = 0; i < chain->NAtoms(); i++) {
    PDBAtom *atom = chain->Atom(i);
    atom->conservation = -999;
    atom->conservation_std = -999;
  }

  // Open conservation file
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    // It is OK for the file to be missing
    // fprintf(stderr, "Unable to open conservation file %s\n", filename);
    return -1;
  }

  // Read lines from conservation file
  int residue_count = 0;
  int line_count = 0;
  char buffer[1024];
  while (fgets(buffer, 1024, fp)) {
    // Increment line count
    line_count++;

    // Check for comment or blank line
    char *bufferp = buffer;
    while (isspace(*bufferp)) bufferp++;
    if (*bufferp == '\0') continue;
    if (*bufferp == '#') continue;

    // Parse residue serial
    bufferp = strtok(bufferp, " ");
    if (!bufferp) { fprintf(stderr, "Unable to parse line %d of %s\n", line_count, filename); return 0; }
    int residue_serial = atoi(bufferp);

    // Parse residue letter
    bufferp = strtok(NULL, " ");
    if (!bufferp) { fprintf(stderr, "Unable to parse residue letter at line %d of %s\n", line_count, filename); return 0; }
    char residue_letter = *bufferp;

    // Parse residue conservation mean
    bufferp = strtok(NULL, "[");
    if (!bufferp) { fprintf(stderr, "Unable to parse mean at line %d of %s\n", line_count, filename); return 0; }
    RNScalar residue_conservation_mean = atof(bufferp);

    // Parse residue conservation minQQ
    bufferp = strtok(NULL, ",");
    if (!bufferp) { fprintf(stderr, "Unable to parse minQQ at line %d of %s\n", line_count, filename); return 0; }
    RNScalar residue_conservation_minQQ = atof(bufferp);

    // Parse residue conservation maxQQ
    bufferp = strtok(NULL, "]");
    if (!bufferp) { fprintf(stderr, "Unable to parse maxQQ at line %d of %s\n", line_count, filename); return 0; }
    RNScalar residue_conservation_maxQQ = atof(bufferp);

    // Parse residue conservation std
    bufferp = strtok(NULL, " ");
    if (!bufferp) { fprintf(stderr, "Unable to parse std at line %d of %s\n", line_count, filename); return 0; }
    RNScalar residue_conservation_std = atof(bufferp);

    // Parse residue conservation number of homologues
    bufferp = strtok(NULL, "/");
    if (!bufferp) { fprintf(stderr, "Unable to parse number of homologues at line %d of %s\n", line_count, filename); return 0; }
    int residue_conservation_num_hom = atoi(bufferp);

    // Parse residue conservation total number of homologues
    bufferp = strtok(NULL, "/");
    if (!bufferp) { fprintf(stderr, "Unable to parse total homologues at line %d of %s\n", line_count, filename); return 0; }
    int residue_conservation_total_hom = atoi(bufferp);

    // Get residue from PDB file
    PDBResidue *residue = NULL;
    PDBAminoAcid *aminoacid = NULL;
    while ((!residue || !aminoacid) && (residue_count < chain->NResidues())) {
      residue = chain->Residue(residue_count++);
      aminoacid = residue->AminoAcid();
    }

    // Check residue and aminoacid
    if (!residue || !aminoacid) {
      fprintf(stderr, "Residue at line %d of %s not found in PDB file\n", line_count, filename); 
      return 0; 
    }

    // Check residue letter
    if ((!aminoacid) || (aminoacid->Letter() != residue_letter)) {
      fprintf(stderr, "Mismatching residue letter (%c %c) at line %d of %s\n", aminoacid->Letter(), residue_letter, line_count, filename); 
      return 0; 
    }

    // Assign residue conservation
    // printf("%s %c %g %g\n", residue->Name(), aminoacid->Letter(), residue_conservation_mean, residue_conservation_std); 
    for (int i = 0; i < residue->NAtoms(); i++) {
      PDBAtom *atom = residue->Atom(i);
      atom->conservation = residue_conservation_mean;
      atom->conservation_std = residue_conservation_std;
    }
  }

  // Close conservation file
  fclose(fp);

  // Return success
  return 1;
}



static int
ReadConservationFiles(PDBFile *file, char *hssp_basename)
{
  // Check number of models
  if (file->NModels() < 1) {
    fprintf(stderr, "File must have at least one model to read conservation scores: %s.\n", hssp_basename);
    return 0;
  }

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read conservations for all chains
  RNBoolean found = 0;
  char chain_filename[256];
  PDBModel *model = file->Model(0);
  for (int i = 0; i < model->NChains(); i++) {
    PDBChain *chain = model->Chain(i);
    char chain_name[4] = { 0 };
    strcpy(chain_name, chain->Name());
    if (chain_name[0] == ' ') chain_name[0] = '_';
    strcpy(chain_filename, hssp_basename);
    strcat(chain_filename, chain_name);
    strcat(chain_filename, ".hssp.orig");
    int status = ReadConservationFile(file, model, chain, chain_filename);
    if (status == 0) return 0;
    else if (status > 0) found = 1;
  }

  // Print statistics
  if (print_verbose && found) {
    printf("Read conservation files ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
ReadASAFile(PDBFile *file, char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Read asa file
  PDBFile asa(filename);
  if (!asa.ReadFile(filename)) {
    RNFail("Unable to read solvent accessible surface area file: %s", filename);
    return NULL;
  }

  // Check number of models
  if (asa.NModels() != file->NModels()) {
    fprintf(stderr, "Number of models in ASA file does not match: %s\n", filename);
    return 0;
  }

  // Assign solvent accessible surface area for all atoms
  for (int i = 0; i < file->NModels(); i++) {
    int atom_count = 0;
    PDBModel *model = file->Model(i);
    PDBModel *asa_model = asa.Model(i);
    for (int j = 0; j < model->NAtoms(); j++) {
      PDBAtom *atom = model->Atom(j);
      if (atom->IsHetAtom()) continue; 
      if (atom_count < asa_model->NAtoms()) {
        PDBAtom *asa_atom = asa_model->Atom(atom_count);
        if (asa_atom->Serial() == atom->Serial()) {
          atom->accessible_surface_area = asa_atom->tempFactor;
          atom_count++;
        }
        else if (asa_atom->Serial() < atom->Serial()) {
          // My PDB parser skips some atoms (e.g., hydrogens), so be lenient to extra atoms in ASA file
          atom_count++;
        }
        else if (asa_atom->Serial() > atom->Serial()) {
          fprintf(stderr, "Serial number of atoms (%d %d) in ASA file does not match: %s\n", atom->Serial(), asa_atom->Serial(), filename);
          return 0;
        }
      }
      else {
        fprintf(stderr, "Number of atoms in %s is greater than in %s\n", file->Name(), filename);
        return 0;
      }
    }
  }

  // Print statistics
  if (print_verbose) {
    printf("Read accessible surface area file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    fflush(stdout);
  }

  // Return success
  return 1;
}



static PDBFile *
ReadSurfnetFile(char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate PDBFile
  PDBFile *file = new PDBFile(filename);
  if (!file) {
    RNFail("Unable to allocate PDB file for %s", filename);
    return NULL;
  }

  // Read PDB file
  if (!file->ReadFile(filename)) {
    RNFail("Unable to read PDB file: %s", filename);
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read surfnet file ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Spheres = %d\n", (file->NModels() > 0) ? file->Model(0)->NAtoms() : 0);
    fflush(stdout);
  }

  // Return success
  return file;
}



static R3Viewer *
CreateViewer(PDBFile *file, PDBResidue *ligand)
{
  // Get file bounding box
  R3Box bbox = file->BBox();
  assert(!bbox.IsEmpty());
  RNLength r = bbox.DiagonalRadius();
  assert((r > 0.0) && RNIsFinite(r));

  // Setup default camera view looking down the Z axis
  if (!initial_camera) initial_camera_origin = bbox.Centroid() - initial_camera_towards * (2 * r);;
  R3Camera camera(initial_camera_origin, initial_camera_towards, initial_camera_up, 0.4, 0.4, 0.1 * r, 1000.0 * r);
  R2Viewport viewport(0, 0, GLUIwindow_width, GLUIwindow_height);
  R3Viewer *viewer = new R3Viewer(camera, viewport);

  // Zoom in on ligand, if there is one
  if (!initial_camera && ligand) {
    // Reset camera
    R3Box ligand_bbox = ligand->BBox();
    RNLength ligand_radius = ligand_bbox.DiagonalRadius();
    assert((ligand_radius > 0.0) && RNIsFinite(ligand_radius));
    R3Point ligand_center = ligand_bbox.Centroid();
    R3Point file_center = file->BBox().Centroid();
    R3Vector towards = file_center - ligand_center;
    RNLength towards_length = towards.Length();
    if (RNIsZero(towards_length)) towards = R3negz_vector;
    else towards /= towards_length;
    R3Vector up = R3posy_vector;
    RNScalar dot = up.Dot(towards);
    if (RNIsEqual(fabs(dot), 1.0)) up = R3posx_vector;
    R3Point eye = ligand_center - towards * (2 * r);;
    viewer->ResetCamera(eye, towards, up);
  }

  // Return viewer
  return viewer;
}



R3Grid *
CreateGrid(PDBFile *file)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get PDB model from file
  if (file->NModels() < 1) return NULL;
  PDBModel *model = file->Model(0);

  // Allocate grid
  R3Grid *grid = new R3Grid(grid_resolution[0], grid_resolution[1], grid_resolution[2]);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(model->BBox());

  // Create grid with 1/distance to a central point
  R3Grid distance_grid(2*grid_resolution[0], 2*grid_resolution[1], 2*grid_resolution[2]);
  distance_grid.SetWorldToGridTransformation(model->BBox());
  distance_grid.SetGridValue(grid_resolution[0], grid_resolution[1], grid_resolution[2], 1);
  distance_grid.SquaredDistanceTransform();
  distance_grid.Sqrt();
  distance_grid.Multiply(grid->GridToWorldScaleFactor());

  // Rasterize each atom into grid
  int atom_count = 0;
  for (int a = 0; a < model->NAtoms(); a++) {
    int i, j, k;
    RNScalar x, y, z;
    PDBAtom *atom = model->Atom(a);
    if (!atom->AminoAcid()) continue;
    if ((atom->accessible_surface_area != -999) && (atom->accessible_surface_area < 10)) continue;
    RNScalar conservation =  1 - atom->conservation;
    if (conservation < 0.001) conservation = 0.001;
    if (atom->conservation == -999) conservation = 0.01;
    const R3Point& atom_position = atom->Position();
    const R3Point& grid_position = grid->GridPosition(atom_position);
    RNScalar cell_radius = grid->GridToWorldScaleFactor();
    if (cell_radius > solvent_radius) cell_radius = solvent_radius;
    RNScalar radius = atom->Radius() + solvent_radius - cell_radius;
    RNScalar *grid_valuesp = (RNScalar *) grid->GridValues();
    for (k = 0, z = grid->ZResolution() - grid_position.Z(); k < grid->ZResolution(); k++, z++) {
      for (j = 0, y = grid->YResolution() - grid_position.Y(); j < grid->YResolution(); j++, y++) {
        for (i = 0, x = grid->XResolution() - grid_position.X(); i < grid->XResolution(); i++, x++) {
          RNScalar distance = distance_grid.GridValue(x, y, z);
          if (distance < radius) *grid_valuesp = -FLT_MAX;
          else *grid_valuesp += conservation / distance;
          grid_valuesp++;
        }
      }
    }
    atom_count++;
  }

  // Assign 0 everywhere inside an atom
  grid->Threshold(0, 0, R3_GRID_KEEP_VALUE);

  // Print statistics
  if (print_verbose) {
    RNInterval range = grid->Range();
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Atoms = %d\n", atom_count);
    printf("  Minimum = %g\n", range.Min());
    printf("  Maximum = %g\n", range.Max());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Set slice coords to middle of grid
  grid_slice_coords[RN_X] = grid_resolution[0]/2;
  grid_slice_coords[RN_Y] = grid_resolution[1]/2;
  grid_slice_coords[RN_Z] = grid_resolution[2]/2;

  // Set grid threshold to maximum
  grid_threshold = grid->Maximum();

  // Return grid
  return grid;
}



static int 
WriteGrid(R3Grid *grid, const char *grid_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write grid
  int status = grid->WriteFile(grid_name);

  // Print statistics
  if (print_verbose) {
    printf("Wrote grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", status * sizeof(RNScalar));
    fflush(stdout);
  }

  // Return status
  return status;
}



static int 
PrintSitePlot(PDBFile *file, PDBResidue *ligand, R3Grid *grid, const char *filename)
{
  // Check everything
  if (!file || !ligand || !grid) return 0;

  // Open file
  FILE *fp = stdout;
  if (filename) {
    fp = fopen(filename, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open site plot file: %s\n", filename);
      return 0;
    }
  }

  // Allocate array of values found at positions of ligand atoms
  RNScalar *ligand_values = new RNScalar [ ligand->NAtoms() ];
  assert(ligand_values);

  // Fill array of values found at positions of ligand atoms
  for (int i = 0; i < ligand->NAtoms(); i++){
    PDBAtom *atom = ligand->Atom(i);
    const R3Point& position = atom->Position();
    ligand_values[i] = grid->WorldValue(position);
  }

  // Sort the array of values found at positions of ligand atoms
  for (int i = 0; i < ligand->NAtoms(); i++) {
    for (int j = i+1; j < ligand->NAtoms(); j++) {
      if (ligand_values[j] > ligand_values[i]) {
        RNScalar swap = ligand_values[i];
        ligand_values[i] = ligand_values[j];
        ligand_values[j] = swap;
      }
    }
  }

  // Determine scale factor to take number of grid cells to a volume
  RNScalar length_scale_factor = grid->GridToWorldScaleFactor();
  RNScalar volume_scale_factor = length_scale_factor * length_scale_factor * length_scale_factor;

  // Count grid values greater than grid values found at positions of ligand atoms
  for (int i = 0; i < ligand->NAtoms(); i++) {
    // Determine fraction of ligand atoms considered
    RNScalar ligand_fraction = (RNScalar) (i+1) / (RNScalar) ligand->NAtoms();

    // Count the number of grid cells with values greater than ligand_value
    int grid_count = 0;
    const RNScalar *grid_values = grid->GridValues();
    for (int j = 0; j < grid->NEntries(); j++) {
      if (grid_values[j] < 0) continue;
      if (grid_values[j] < ligand_values[i]) continue;
      grid_count++;
    }

    // Convert number of grid cells to a volume
    RNScalar grid_volume = grid_count * volume_scale_factor;

    // Print plot numbers
    fprintf(fp, "%g %g\n", ligand_fraction, grid_volume);
  }

  // Close file
  if (filename) fclose(fp);

  // Set grid_threshold
  int index = 0.75 * ligand->NAtoms();
  if (index == ligand->NAtoms()) index--;
  grid_threshold = ligand_values[index];

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: symsam [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-batch")) batch = 1; 
      else if (!strcmp(*argv, "-ligand")) { argc--; argv++; ligand_string = *argv; }
      else if (!strcmp(*argv, "-conservation")) { argc--; argv++; conservation_name = *argv; }
      else if (!strcmp(*argv, "-surfnet")) { argc--; argv++; surfnet_name = *argv; }
      else if (!strcmp(*argv, "-asa")) { argc--; argv++; asa_name = *argv; }
      else if (!strcmp(*argv, "-grid")) { argc--; argv++; grid_name = *argv; }
      else if (!strcmp(*argv, "-plot")) { argc--; argv++; plot_name = *argv; }
      else if (!strcmp(*argv, "-image")) { argc--; argv++; image_name = *argv; }
      else if (!strcmp(*argv, "-offset")) { argc--; argv++; solvent_radius = atof(*argv); }
      else if (!strcmp(*argv, "-show_conservation")) show_conservation = 1; 
      else if (!strcmp(*argv, "-colors")) {
        argv++; argc--; file_color_scheme = atoi(*argv) % 5; 
        argv++; argc--; ligand_color_scheme = atoi(*argv) % 5; 
      }
      else if (!strcmp(*argv, "-background")) { 
        argc--; argv++; background_color[0] = atof(*argv); 
        argc--; argv++; background_color[1] = atof(*argv); 
        argc--; argv++; background_color[2] = atof(*argv); 
      }
      else if (!strcmp(*argv, "-resolution")) { 
        argc--; argv++; grid_resolution[0] = atoi(*argv); 
        argc--; argv++; grid_resolution[1] = atoi(*argv); 
        argc--; argv++; grid_resolution[2] = atoi(*argv); 
      }
      else if (!strcmp(*argv, "-camera")) { 
        RNCoord x, y, z, tx, ty, tz, ux, uy, uz;
        argv++; argc--; x = atof(*argv); 
        argv++; argc--; y = atof(*argv); 
        argv++; argc--; z = atof(*argv); 
        argv++; argc--; tx = atof(*argv); 
        argv++; argc--; ty = atof(*argv); 
        argv++; argc--; tz = atof(*argv); 
        argv++; argc--; ux = atof(*argv); 
        argv++; argc--; uy = atof(*argv); 
        argv++; argc--; uz = atof(*argv); 
        initial_camera_origin = R3Point(x, y, z);
        initial_camera_towards.Reset(tx, ty, tz);
        initial_camera_up.Reset(ux, uy, uz);
        initial_camera = TRUE;
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!pdb_name) pdb_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s\n", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
  }

  // Check pdb filename
  if (!pdb_name) {
    fprintf(stderr, "You did not specify a pdb file.\n");
    return 0;
  }

  // Check grid resolution
  if ((grid_resolution[0] <= 0) || (grid_resolution[1] <= 0) || (grid_resolution[2] <= 0)) {
    fprintf(stderr, "Invalid grid resolution: %d %d %d\n", grid_resolution[0], grid_resolution[1], grid_resolution[2]);
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Initialize GLUI
  GLUIInit(&argc, argv);

  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read pdb
  file = ReadPDB(pdb_name);
  if (!file) exit(-1);

  // Read conservation file
  if (conservation_name) {
    int status = ReadConservationFiles(file, conservation_name);
    if (!status) exit(-1);
  }

  // Read accessible surface area file
  if (asa_name) {
    int status = ReadASAFile(file, asa_name);
    if (!status) exit(-1);
  }

  // Read surfnet file
  if (surfnet_name) {
    surfnet_file = ReadSurfnetFile(surfnet_name);
    if (!surfnet_file) exit(-1);
  }

  // Create viewer
  viewer = CreateViewer(file, ligand);
  if (!viewer) exit(-1);

  // Create grid
  grid = CreateGrid(file);
  if (!grid) exit(-1);

  // Write grid
  if (grid_name) {
    int status = WriteGrid(grid, grid_name);
    if (!status) exit(-1);
  }

  // Print binding site plot
  if (plot_name) {
    int status = PrintSitePlot(file, ligand, grid, plot_name);
    if (!status) exit(-1);
  }

  // Run GUI
  if (image_name || !batch) {
    // Run GLUT interface
    GLUIMainLoop();
  }

  // Return success 
  return 0;
}



